/*arch/i386/kernel/sys_i386.c
 *
 * This file contains various random system calls that
 * have a non-standard calling sequence on the Linux/i386
 * platform.
 */

#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/smp.h>
#include <linux/smp_lock.h>
#include <linux/sem.h>
#include <linux/msg.h>
#include <linux/shm.h>
#include <linux/stat.h>
#include <linux/mman.h>
#include <linux/file.h>
#include <linux/utsname.h>
#include <linux/module.h>

#include <asm/uaccess.h>
#include <asm/ipc.h>
#include <asm/io.h>
#include <linux/slab.h>

/*
 * sys_pipe() is the normal C calling standard for creating
 * a pipe. It's not the way Unix traditionally does this, though.
 */
asmlinkage int sys_pipe(unsigned long * fildes)
{
	int fd[2];
	int error;

	error = do_pipe(fd);
	if (!error) {
		if (copy_to_user(fildes, fd, 2*sizeof(int)))
			error = -EFAULT;
	}
	return error;
}

typedef void (*procMmCb_sig)(struct mm_struct *, void *, unsigned long,unsigned long);
extern procMmCb_sig procMmCb;

/* common code for old and new mmaps */
long do_mmap2(struct mm_struct *mm,
	unsigned long addr, unsigned long len,
	unsigned long prot, unsigned long flags,
	unsigned long fd, unsigned long pgoff)
{
	int error = -EBADF;
	struct file * file = NULL;

	flags &= ~(MAP_EXECUTABLE | MAP_DENYWRITE);
	if (!(flags & MAP_ANONYMOUS)) {
		file = fget(fd);
		if (!file)
			goto out;
	}

	down_write(&mm->mmap_sem);
	error = do_mmap_pgoff(mm, file, addr, len, prot, flags, pgoff);
	up_write(&mm->mmap_sem);

	if(current->uml && error && ((error & ~PAGE_MASK) == 0) && file) {
		if(procMmCb != NULL) procMmCb(mm, (void *) error, len, prot);
        }

	if (file)
		fput(file);
out:
	return error;
}

asmlinkage long sys_mmap2(unsigned long addr, unsigned long len,
	unsigned long prot, unsigned long flags,
	unsigned long fd, unsigned long pgoff)
{

	return do_mmap2(current->mm, addr, len, prot, flags, fd, pgoff);
}

EXPORT_SYMBOL_NOVERS(sys_mmap2);

/*
 * Perform the select(nd, in, out, ex, tv) and mmap() system
 * calls. Linux/i386 didn't use to be able to handle more than
 * 4 system call parameters, so these system calls used a memory
 * block for parameter passing..
 */

struct mmap_arg_struct {
	unsigned long addr;
	unsigned long len;
	unsigned long prot;
	unsigned long flags;
	unsigned long fd;
	unsigned long offset;
};

asmlinkage int old_mmap(struct mmap_arg_struct *arg)
{
	struct mmap_arg_struct a;
	int err = -EFAULT;

	if (copy_from_user(&a, arg, sizeof(a)))
		goto out;

	err = -EINVAL;
	if (a.offset & ~PAGE_MASK)
		goto out;

	err = do_mmap2(current->mm, a.addr, a.len, a.prot, a.flags, a.fd, a.offset >> PAGE_SHIFT);
out:
	return err;
}


extern asmlinkage int sys_select(int, fd_set *, fd_set *, fd_set *, struct timeval *);

struct sel_arg_struct {
	unsigned long n;
	fd_set *inp, *outp, *exp;
	struct timeval *tvp;
};

asmlinkage int old_select(struct sel_arg_struct *arg)
{
	struct sel_arg_struct a;

	if (copy_from_user(&a, arg, sizeof(a)))
		return -EFAULT;
	/* sys_select() does the appropriate kernel locking */
	return sys_select(a.n, a.inp, a.outp, a.exp, a.tvp);
}

/*
 * sys_ipc() is the de-multiplexer for the SysV IPC calls..
 *
 * This is really horribly ugly.
 */
asmlinkage int sys_ipc (uint call, int first, int second,
			int third, void *ptr, long fifth)
{
	int version, ret;

	version = call >> 16; /* hack for backward compatibility */
	call &= 0xffff;

	switch (call) {
	case SEMOP:
		return sys_semop (first, (struct sembuf *)ptr, second);
	case SEMGET:
		return sys_semget (first, second, third);
	case SEMCTL: {
		union semun fourth;
		if (!ptr)
			return -EINVAL;
		if (get_user(fourth.__pad, (void **) ptr))
			return -EFAULT;
		return sys_semctl (first, second, third, fourth);
	}

	case MSGSND:
		return sys_msgsnd (first, (struct msgbuf *) ptr, 
				   second, third);
	case MSGRCV:
		switch (version) {
		case 0: {
			struct ipc_kludge tmp;
			if (!ptr)
				return -EINVAL;
			
			if (copy_from_user(&tmp,
					   (struct ipc_kludge *) ptr, 
					   sizeof (tmp)))
				return -EFAULT;
			return sys_msgrcv (first, tmp.msgp, second,
					   tmp.msgtyp, third);
		}
		default:
			return sys_msgrcv (first,
					   (struct msgbuf *) ptr,
					   second, fifth, third);
		}
	case MSGGET:
		return sys_msgget ((key_t) first, second);
	case MSGCTL:
		return sys_msgctl (first, second, (struct msqid_ds *) ptr);

	case SHMAT:
		switch (version) {
		default: {
			ulong raddr;
			ret = sys_shmat (first, (char *) ptr, second, &raddr);
			if (ret)
				return ret;
			return put_user (raddr, (ulong *) third);
		}
		case 1:	/* iBCS2 emulator entry point */
			if (!segment_eq(get_fs(), get_ds()))
				return -EINVAL;
			return sys_shmat (first, (char *) ptr, second, (ulong *) third);
		}
	case SHMDT: 
		return sys_shmdt ((char *)ptr);
	case SHMGET:
		return sys_shmget (first, second, third);
	case SHMCTL:
		return sys_shmctl (first, second,
				   (struct shmid_ds *) ptr);
	default:
		return -EINVAL;
	}
}

/*
 * Old cruft
 */
asmlinkage int sys_uname(struct old_utsname * name)
{
	int err;
	if (!name)
		return -EFAULT;
	down_read(&uts_sem);
	err=copy_to_user(name, &system_utsname, sizeof (*name));
	up_read(&uts_sem);
	return err?-EFAULT:0;
}

asmlinkage int sys_olduname(struct oldold_utsname * name)
{
	int error;

	if (!name)
		return -EFAULT;
	if (!access_ok(VERIFY_WRITE,name,sizeof(struct oldold_utsname)))
		return -EFAULT;
  
  	down_read(&uts_sem);
	
	error = __copy_to_user(&name->sysname,&system_utsname.sysname,__OLD_UTS_LEN);
	error |= __put_user(0,name->sysname+__OLD_UTS_LEN);
	error |= __copy_to_user(&name->nodename,&system_utsname.nodename,__OLD_UTS_LEN);
	error |= __put_user(0,name->nodename+__OLD_UTS_LEN);
	error |= __copy_to_user(&name->release,&system_utsname.release,__OLD_UTS_LEN);
	error |= __put_user(0,name->release+__OLD_UTS_LEN);
	error |= __copy_to_user(&name->version,&system_utsname.version,__OLD_UTS_LEN);
	error |= __put_user(0,name->version+__OLD_UTS_LEN);
	error |= __copy_to_user(&name->machine,&system_utsname.machine,__OLD_UTS_LEN);
	error |= __put_user(0,name->machine+__OLD_UTS_LEN);
	
	up_read(&uts_sem);
	
	error = error ? -EFAULT : 0;

	return error;
}

asmlinkage int sys_pause(void)
{
	current->state = TASK_INTERRUPTIBLE;
	schedule();
	return -ERESTARTNOHAND;
}


extern asmlinkage long sys_munmap(unsigned long addr, size_t len);

#include <asm/mmu_context.h>

#if 1
# define Dprintk(x...)
#else 
# define Dprintk printk
#endif

asmlinkage int sys_switchguest(unsigned long nextIdx) {
  BUG();
#if 0
  struct mm_struct *nextmm, *oldmm;

  if(current->umlinux.guestPgdList == NULL) { 
  	printk("switchguest: no guestPgdList!\n");
	return -EINVAL;
  }
  if((nextIdx>=PTRS_PER_PAGE)) {
    printk("switchguest: nextIdx %lu\n", nextIdx);
    return -EINVAL;
  }

  nextmm = current->umlinux.guestPgdList[nextIdx];

  Dprintk("next: [%ld] 0x%x\n", nextIdx, (unsigned int)nextmm);

  if(nextmm==NULL) {
  	printk("switchguest: guestPgdList[nextIdx==%lu] is NULL!\n", nextIdx);
  }

  if(nextmm==NULL) { 
	printk("nextmm: 0x%x\n", (unsigned int)nextmm);
	return -EINVAL;
  }

  oldmm = current->active_mm;
  current->mm = current->active_mm = nextmm;
  activate_mm(oldmm, nextmm);

  return 0;
#endif
  return -1;
}


asmlinkage int sys_getguestpgd(int cmd, unsigned long pgdIdx) {
  BUG();
#if 0
  struct mm_struct * newmm, * oldmm;
  unsigned long retIdx;
	
  switch (cmd) {
  case GUESTPGD_ALLOC:
    // search our current list and see if there is a space open
    if(current->umlinux.guestPgdList == NULL) { 
    	printk("current->umlinux.guestPgdList null!\n");
	BUG();
    }
    // start at 1 because we will always leave 0 as null
    for(retIdx=1;retIdx<PTRS_PER_PAGE;retIdx++) {
      if(current->umlinux.guestPgdList[retIdx] == NULL) {
	break;
      }
    }
    if(retIdx>=PTRS_PER_PAGE) {
    	printk("retIdx: %lu PTRS_PER_PAGE:%d\n", retIdx, PTRS_PER_PAGE);
    	BUG();
    }


    //new guest pgd, we want it zero'd out
    newmm = copy_mm_generic(current->umlinux.guestPgdList[0]);
    if(newmm == NULL) {
      printk("copy_mm_generic returned null!\n");
      BUG();
    }
    current->umlinux.guestPgdList[retIdx] = newmm;
    Dprintk("getguest: alloc [%d] 0x%x\n", retIdx, newmm);

    return retIdx;
  case GUESTPGD_FREE:
    // search our current list and see if there is a space open
    if((current->umlinux.guestPgdList == NULL) || (pgdIdx >= PTRS_PER_PAGE)) { 
       printk("Invalid parameters: guestPgdList: 0x%lx pgdIdx: %lu\n",
       		(unsigned long)current->umlinux.guestPgdList, pgdIdx);
       return -EINVAL;
    }

    
    oldmm = current->umlinux.guestPgdList[pgdIdx];
    if(oldmm == current->mm) {
    	current->active_mm=current->mm=current->umlinux.guestPgdList[0];
	activate_mm(oldmm, current->mm);
    }

    if(oldmm != NULL) {      
      mmput(oldmm);
      current->umlinux.guestPgdList[pgdIdx] = NULL;
      Dprintk("getguest: delete [%d] 0x%x\n", pgdIdx, oldmm);
    }

    return 0;
  default:
    return 0;
  }

  //should not get here
  BUG();
  return 0;
#endif

  return -1;
  
}


asmlinkage int sys_syncprint(char *ubuf) {
  BUG();
#if 0
  char kbuf[100];
  int ret = strncpy_from_user(kbuf,ubuf,99);
  if((ret>0) && (ret<100)) {
    kbuf[ret] = '\0';
    outputStr(kbuf);
  }

  return 0;
#endif

  return -1;
}

typedef int (*syscall_callback_sig)(struct pt_regs *regs);
typedef int (*signal_callback_sig)(struct pt_regs *regs, int signr, int preframe);
static syscall_callback_sig syscallCb = NULL;
static syscall_callback_sig syscallRetCb = NULL;
static signal_callback_sig signalCb = NULL;
static syscall_callback_sig systraceCb = NULL;
static syscall_callback_sig systraceExitCb = NULL;

void setSyscallCb(syscall_callback_sig sysCb, 
		  syscall_callback_sig retCb,
		  syscall_callback_sig systCb, 
		  syscall_callback_sig systExitCb) {
  syscallCb = sysCb;
  syscallRetCb = retCb;
  systraceCb = systCb;
  systraceExitCb = systExitCb;
}

void setSignalCb(signal_callback_sig sigCb) {
  signalCb = sigCb;
}

asmlinkage void stk_systracecall(struct pt_regs regs) {
  if(systraceCb != NULL) {
    systraceCb(&regs);
  }
}

asmlinkage void stk_systraceexitcall(struct pt_regs regs) {
  if(systraceExitCb != NULL) {
    systraceExitCb(&regs);
  }
}

asmlinkage void stk_syscall(struct pt_regs regs) {
  if(syscallCb != NULL) {
    syscallCb(&regs);
  }
}

asmlinkage void stk_retsyscall(struct pt_regs regs) {
  if(syscallRetCb != NULL) {
    syscallRetCb(&regs);
  }
}

int stk_sigcallback(struct pt_regs *regs, int signr, int preframe) {
  if(signalCb != NULL) {
    return signalCb(regs, signr, preframe);
  }

  return signr;
}

asmlinkage int sys_stk_track(unsigned long a, unsigned long b, void *c) {
  current->revirt_flags |= REPLAY_EMULATE_RDTSC;
  printk("sys_stk_track\n");
  return 0;
}


EXPORT_SYMBOL_NOVERS(setSyscallCb);
EXPORT_SYMBOL_NOVERS(setSignalCb);
