/*
 * lexecve.c - Log executable using the ReVirt system
 *
 * Copyright (C) 2002 George Dunlap
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/config.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/smp_lock.h>
#include <linux/file.h>
#include <asm/uaccess.h>
#include <asm/pgtable.h>
#include <asm/perfctr.h>
#include <linux/slab.h>
#include <linux/sched.h>
#include <revirt/rlog.h>
#include <revirt/ztrace.h>

#define MAX_ARG_SPACE 2*4096

static int log_exec_params(struct task_struct* task, char *fname,char** argv,
								char **env);
extern inline void clear_branchdata(struct task_struct *tsk);

/* FIXME -- dynamic allocation is a pain... we'll work on it later. */
struct revirt_group_struct revirt_group;

int create_new_rtsk(struct task_struct * tsk, struct task_struct * parent, int syscall) {
	int error;
	struct revirt_tsk_s * rtsk;
	struct revirt_group_struct * rgroup;

	if(parent && !parent->revirt_tsk) {
		printk("create_new_rtsk: parent pid %d has no rtsk!\n", parent->pid);
		error=-EINVAL;
		goto out;
	}

	if(!(rtsk=kmalloc(sizeof(*rtsk),GFP_KERNEL))) {
		error=-ENOMEM;
		goto out;
	}
	/* next, group, and sched handled below */

	rtsk->current_syscall=syscall;
	rtsk->syscall_id=1;

	rtsk->sysret_cache.rh.type=-1;
	rtsk->sysret_cache.syscall_id=0;

	rtsk->syserror_replay_unlogged=0;
	rtsk->syserror_write_cleanup=0;
	rtsk->syserror_save=0;

	if(parent) {
	    rtsk->sched = parent->revirt_tsk->sched;
	    rgroup=parent->revirt_tsk->group;
	} else {
	    rtsk->sched=NULL;
#if 0
	    if(!(rgroup = kmalloc(sizeof(*rgroup),GFP_KERNEL))) {
		error=-ENOMEM;
		goto free_rtsk;
	    }
#else
	    rgroup = &revirt_group;
#endif
	    rgroup->ref=0;	/* incremented below */
	    rgroup->sched = NULL;
	    rgroup->procs=tsk;

	    rgroup->zt_ctl.flags=0;
	    rgroup->zt_ctl.yield=0;
	    rgroup->zt_ctl.pid=0;
	    rgroup->zt_ctl.rhid=0;
	    rgroup->zt_ctl.last_sched_id=0;
	    rgroup->zt_ctl.perferr=0;
	    rgroup->zt_ctl.can_checkpoint=0;

	    rgroup->active.i = 0;
	    rgroup->active.pid[0] = 0;
	    rgroup->active.epochs_skipped = 0;
	    rgroup->active.runnable=1;

	    rgroup->replay_pos=0;
	    rgroup->log_id=0;
	    rgroup->log_pos=0;
	    rgroup->scallf=NULL;
	    rgroup->cache.valid=0;

	    rgroup->filename[0]=0;
	}

	rtsk->group = rgroup;
	rtsk->sched = rgroup->sched;

	/* This is the point-of-no-return... */
	rgroup->ref++;
	tsk->revirt_tsk=rtsk;
	if(parent) {
	    rtsk->next=parent->revirt_tsk->next;
	    parent->revirt_tsk->next=tsk;
	} else {
	    rtsk->next = tsk;
	}

	return 0;

    free_rgroup:
    	
    free_rtsk:
	kfree(rtsk);
    out:
    	return error;
}

int put_rtsk(struct task_struct * tsk) {
    struct revirt_tsk_s * rtsk;
    struct revirt_group_struct * rgroup;
    struct task_struct * p;

    if(!tsk)
    	return 0;

    if(!(rtsk=tsk->revirt_tsk)) 
    	return 0;
    

    /* FIXME for SMP, we really need some kind of locking here... */

    /* Find the guy behind us in the circular queue, and point him to the guy
       ahead of us... */
    /* 
     * This has to be after log_exit(), because it wants to walk the circle
     * and clear branch data.
     */
    for(p=rtsk->next; 
    		p->revirt_tsk && p->revirt_tsk->next != tsk; 
		p=p->revirt_tsk->next);

    if(p->revirt_tsk) {
	p->revirt_tsk->next = rtsk->next;
	/* in the case that p == tsk, don't set it to null quite yet... */
	Rprintk("put_rtsk: pid %d revirt_tsk->next moved from pid %d to %d\n",
			p->pid, tsk->pid, p->revirt_tsk->next->pid, 0, 0);
	rtsk->next = tsk;
    } else {
	printk("put_rtsk: WARNING pid %d has no revirt_tsk, ring broken!\n",
		p->pid);
    }

    rgroup = rtsk->group;
    rtsk->group = NULL;

    if(rgroup) {
    	rgroup->ref--;

	if(rgroup->zt_ctl.tsk == tsk) {
	    Rprintk("put_rtsk: setting ztc.tsk to NULL\n", 0,0,0,0,0);
	    rgroup->zt_ctl.tsk=NULL;
	}

	Rprintk("put_rtsk: rgroup->ref %d\n", rgroup->ref,0,0,0,0);

	/* FIXME -- add in "scheduler" deletion logic... */
	if(!rgroup->ref) {
	    Rprintk("put_rtsk: rgroup->ref null, freeing.\n",0,0,0,0,0);
	    /* revirt_read_close checks for non-nullity of scallf. */
	    revirt_read_close(rgroup);
	    if(tsk->revirt_flags & REPLAY_LOG && rgroup->sched) {
	    	printk("Last log proc exiting, scheduler %d still running! Waking it up...\n",
				rgroup->sched->pid);
		rgroup->zt_ctl.yield = ZTRACE_WAITEXIT;
		rgroup->sched->revirt_flags &= ~REPLAY_WAIT;
	    }
#if 0
	    kfree(rgroup);
#endif
	    rgroup=NULL;
	} else {
	    if(rgroup->sched == tsk) {
		Rprintk("put_rtsk: settng rgroup->sched to NULL due to proc %d\n",
			tsk->pid,0,0,0,0);
	    	rgroup->sched = NULL;
	    } else if(rgroup->procs == tsk) {
	    	Rprintk("put_rtsk: moving rgroup->procs from pid %d to %d\n",
			tsk->pid, rtsk->next->pid,0,0,0);
	    	rgroup->procs = rtsk->next;
	    }
	}
    }

    kfree(rtsk);
    tsk->revirt_tsk=NULL;
    return 0;
}

static int copy_filename(struct task_struct * tsk, char * f) {
    struct revirt_group_struct * rgroup;
    char * p;
    int i;

    rgroup=tsk->revirt_tsk->group;
    p=getname(f);

    for(i=0; i<REVIRT_MAX_PATHNAME && p[i]; i++) {
    	rgroup->filename[i]=p[i];
    }
    rgroup->filename[i]=0;

    printk(" copy_filename copied %d bytes (%s)\n", i, rgroup->filename);
    putname(p);

    return 0;
}

/*
 * sys_lexecve() sets the logging flag of the process and
 * executes a new program. 
 * Synopsis:
 * lexecve( char *filename, char **argp, char **envp, 
 *                                               char *syscall_log_filename )
 */
asmlinkage int sys_lexecve(struct pt_regs regs)
{
	char * filename, *logfilename;
	int error;
        struct rlog_schedule_entry le;
	struct task_struct * tsk=current;

        /* perhaps we should do some checking, for the newer libc. */
	printk("sys_lexecve is called\n");

	if(tsk->revirt_tsk!=NULL) {
		error=-EBUSY;
		goto out;
	}

	if((error=create_new_rtsk(tsk, NULL, regs.orig_eax))<0) {
		goto out;
	}

	if((logfilename=(char *)(regs.esi))) {
	    copy_filename(tsk, logfilename);
	}

	/*
	 * Tasks frequently get scheduled out several times before they
	 * ever make it out to userland, so we want to suppress "schedule" 
	 * entries.  Set the trap flag...
	 *
	 * FIXME -- maybe we should set this in the exec code somewhere?
	 * maybe on an iret or something?
	 */
	tsk->revirt_flags= REPLAY_LOG;
	tsk->revirt_flags|= REPLAY_SUPPRESS_SCHED;
	zreg(tsk,eflags) |= TRAP_FLAG;

        le= (struct rlog_schedule_entry) { 
                //.type= ENTRY_TOP,
                .pid= tsk->pid,
        	.data= 0
              };

	revirt_write_log((char *)&le, ENTRY_TOP, sizeof(le), tsk->revirt_tsk->group);

	perfctr_enable_counters();
	perfctr_clear_counters();

	clear_branchdata(tsk);

	error= log_exec_params(tsk,(char*) regs.ebx,(char**) regs.ecx,
	                 (char**) regs.edx);
	if (error)
		goto out_free;

	/* rest is from the original execve code */
	filename = getname((char *) regs.ebx);
	error = PTR_ERR(filename);
	if (IS_ERR(filename))
		goto out_fname;
	error = do_execve(filename, (char **) regs.ecx, (char **) regs.edx, 
							&regs);
	return error;

out_fname:
	putname(filename);
out_free:
	put_rtsk(tsk);
out:
	tsk->revirt_flags= 0;
	return error;
}

/*
 * We need to know the size of the buffer so that we can ensure we have no
 * buffer overruns -- i.e., we need strnlen_user to make sense.
 */
static int copy_argv(char *as, char **argv, int size) {
	char * arg;
	int argsize, i;

	for(i=0; ; i++) {
		arg = 0;
		copy_from_user(&arg, argv+i, 4);
		if(arg == 0) {
			return size;
		}
		argsize=strnlen_user(arg, MAX_ARG_SPACE-size-1);
		if(arg[argsize-1]!=0) 
			return -E2BIG;
		copy_from_user(as, arg, argsize);
		size += argsize;
		as += argsize;
	}
}


static int log_exec_params(struct task_struct *task, char *fname,
				char** argv, char **env) {
	int size, fsize;
	char *buf, *a, *e, *f;
	int error;
	struct rlog_entry_scall_lexecve * se;

	/* FIXME: do PTR_ERR checks here */
	fsize=strnlen_user(fname, MAX_ARG_SPACE-1);
	error= -E2BIG;
	if (fname[fsize-1]!=0)
		goto out;

	error= -ENOMEM;
	buf= kmalloc(MAX_ARG_SPACE, GFP_KERNEL);
	if (!buf)
		goto out;
	
	se = (struct rlog_entry_scall_lexecve *)buf;
	//se->type= ENTRY_EXECPARM;
	se->pid = current->pid;
	size= sizeof(struct rlog_entry_scall_lexecve);

	f= buf + size;
	copy_from_user(f, fname, fsize);
	size += fsize;

	a = buf + size;
	se->argv_index = size;
	size = copy_argv(a, argv, size);
	if (size<0) {
		error= size;
		goto out_free_arg_space;
	}


	e = buf+size;
	se->env_index = size;
	size = copy_argv(e, env, size);
	if (size<0) {
		error= size;
		goto out_free_arg_space;
	}

	revirt_write_log(buf, ENTRY_EXECPARM, size, task->revirt_tsk->group);
	error=0;

out_free_arg_space:
	kfree(buf);

out:
	return error;
}
