/*
 * logfile.c - Functions associated with reading a ReVirt logfile during replay
 *
 * Copyright (C) 2002 George Dunlap
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/sched.h>
#include <asm/processor.h>
#include <linux/fs.h>
#include <linux/module.h>
#include <asm/uaccess.h>
#include <revirt/rlog.h>

#define REPLAY_INPUT_BUF 4096*32

#define SEEK_CUR 1 /* see fs/read_write.c:39 */

//static struct file * revirt_scallf = NULL;

/*
 * Yes, this is wasteful. We can optimize it later.
 *
 * Do we need some kind of locks around this?
 */
char revirt_scall_buf[REPLAY_INPUT_BUF];

EXPORT_SYMBOL_NOVERS(revirt_scall_buf);

int revirt_read_open(char * fname, struct revirt_group_struct * rgroup) {
	int error;

	if(rgroup->scallf != NULL) {
		printk("revirt_read_open: scallf not NULL!\n");
		return -EBUSY;
	}

	error=PTR_ERR(fname);
	if(IS_ERR(fname))
		goto out;

	rgroup->scallf = filp_open(fname, O_RDONLY, 0644);
	rgroup->replay_pos=0;
	error=PTR_ERR(rgroup->scallf);
	if(IS_ERR(rgroup->scallf))
		goto out;
	error=0;
   out:
	return error;
}

int revirt_read_close(struct revirt_group_struct * rgroup) {
	int ret=0;

	if(rgroup->scallf) {
	    printk("revirt_read_close: closing scallf\n");
	    ret=filp_close(rgroup->scallf, NULL);
	    rgroup->scallf=NULL;
	    rgroup->cache.valid=0;
	} 
	return ret;
}

/*
 * So, there are several people who need to read the log.  The "main" readers
 * have their in-order ppos, but there are others who need to scan forward
 * for a bit without perturbing that.  The __ functions take a pointer to
 * a file offset.  The "main" readers will use the pointer in the rgroup
 * struct, while the scanners should copy that pointer into a local buffer
 * and use that.
 */

static inline void check_filepos(struct rlog_header * rh, 
	struct revirt_group_struct * rgroup, loff_t * ppos) {
    if(*ppos - sizeof(*rh) != rh->pos) {
    	printk("check_fpos: id %lu wanted %lld, got %lld\n", rh->id, rh->pos,
		*ppos);
    }

    if(rgroup->log_pos > rh->pos) {	
    	Rprintk("revirt_read_log: rgroup->log_pos %d > r.pos %d (id %lu) *ppos %d\n",
		rgroup->log_pos, rh->pos, rh->id, rgroup->replay_pos, 0);
    }

    rgroup->log_pos = rh->pos;
    rgroup->log_id = rh->id;
}


/* 
 * Reads the header, using the given offset.  
 */
static ssize_t __revirt_read_header(struct rlog_header *rh, loff_t * ppos,
	struct revirt_group_struct * rgroup) {
	ssize_t (*read) (struct file *, char *, size_t, loff_t *);
	ssize_t ret;
	read= rgroup->scallf->f_op->read;
	ret= (*read)(rgroup->scallf, (char *)rh, sizeof(*rh), ppos);
	Rprintk("__revirt_read_header returned %d (t %x id %d)\n", 
		ret, rh->type, rh->id, 0, 0);
	return ret;
}

static ssize_t __revirt_read_next_sys(struct rlog_header *rh, loff_t * ppos,
	struct revirt_group_struct * rgroup) {
    int l;
    while(1) {
	    l=__revirt_read_header(rh, ppos, rgroup);
	    if(l<0) {
	    	printk("revirt_read_log: Error reading log: %d\n", l);
		goto out;
	    }

	    check_filepos(rh, rgroup, ppos);
	    /*
	     * fork entries need to be read both by zr and the kernel revirt
	     * system.
	     *
	     * Actually, I suppose fork need not be read by zr, it could infer
	     * one by seeing a new pid, but that sounds sketchy...
	     */
	    if(IS_SCHED(rh->type) && rh->type!=ENTRY_FORK) {
		(*ppos) += rh->len;
	    } else
	    	break;

	    /* play nice */	    
	    if(current->need_resched)
	    	schedule();
    }
    out:

    return l;
}

/*
 *  Given the header read already, reads data into the target location.
 *
 * WARNING: it's up to the caller to make sure there's enough space...
 */
static ssize_t __revirt_read_data(struct rlog_header *rh, char * d, 
	loff_t * ppos, struct revirt_group_struct * rgroup) {
    ssize_t (*read) (struct file *, char *, size_t, loff_t *);
    ssize_t ret;
    read= rgroup->scallf->f_op->read;

    if((d == revirt_scall_buf) && (rh->len > REPLAY_INPUT_BUF)) {
    	loff_t tpos;
    	printk("WARNING: trying to read %d into revirt_scall_buf, max %d\n",
		rh->len, REPLAY_INPUT_BUF);
	tpos = *ppos + rh->len;
	ret=(*read)(rgroup->scallf, d, REPLAY_INPUT_BUF, ppos);
	*ppos = tpos;
    } else {
	ret=(*read)(rgroup->scallf, d, rh->len, ppos);
    }

    Rprintk("__revirt_read_data returned %d\n", ret, 0, 0, 0, 0);
    if(ret < 0) {
    	printk("revirt_read_log: Error reading log: %d\n", ret);
    }
    return ret;
}



/*
 * this function will read the next log, but put it in a cache so that the next
 * revirt_read_log() will get the same thing.
 *
 * FIXME: right now, this will read any log, not just a 'schedule' log...
 */
int revirt_read_ahead(struct rlog_header * rh, char * data) {
	struct revirt_group_struct * rgroup;
	loff_t pos;
	ssize_t r;
	mm_segment_t oldfs;

	if(!(rgroup=get_rgroup(current))) {
		printk("revirt_read_ahead: no rgroup!\n");
		return -EBADF;
	}
	
	pos=rgroup->replay_pos;

	oldfs=get_fs();
	set_fs(KERNEL_DS);
	current->revirt_flags |= REPLAY_NO_YIELD_ON_BLOCK;
	/* 
	 * FIXME These need to be atomic, so there should probably be a lock...
	 */
#if 1
	if((r=__revirt_read_next_sys(rh, &pos, rgroup))<0) {
	    goto out_setfs;
	}
#else
	if((r=__revirt_read_header(rh, &pos, rgroup))<0) {
	    goto out_setfs;
	}
	check_filepos(rh, rgroup, &pos);
#endif
	
	r=__revirt_read_data(rh, revirt_scall_buf, &pos, rgroup);

	out_setfs:
	    current->revirt_flags &= ~REPLAY_NO_YIELD_ON_BLOCK;
	    set_fs(oldfs);
	
	return r;
}

EXPORT_SYMBOL_NOVERS(revirt_read_ahead);

/*
 * Search for the next sysreturn log for a given pid; or, if an exit log is found for
 * that pid instead, stop and return it.
 */
int revirt_read_next_sysret(struct rlog_header *rh, 
		struct rlog_entry_sysret *le, 
		s64 * cache_syscall_id, 
		int pid, s64 current_syscall_id){
	int r;
	mm_segment_t oldfs;
	//struct rlog_entry_sysret * tmp=NULL;
	struct rlog_exit_entry ee;
	struct rlog_schedule_entry se;
	loff_t pos;
	struct revirt_group_struct * rgroup;
	//ssize_t (*read) (struct file *, char *, size_t, loff_t *);
	//int i=0;
	//int try_cache=0;


	if(!(rgroup=get_rgroup(current))) {
	    printk("revirt_read_next_sysret: no rgroup!\n");
	    return -EBADF;
	}

	oldfs=get_fs();
	set_fs(KERNEL_DS);

	current->revirt_flags |= REPLAY_NO_YIELD_ON_BLOCK;
	
	/* Use a local copy of pos, because we don't want to disturb the "global" copy */
	pos = rgroup->replay_pos;
	//read= rgroup->scallf->f_op->read;

	for(;;) {
	    r=__revirt_read_header(rh, &pos, rgroup);
	    Rprintk("__revirt_read_header_stateless: pid %d pos %lu type %x id %lu\n",
	    		current->pid, (unsigned long)pos, rh->type, rh->id, 0);
	    if(r < 0) 
	    	goto out_setfs;
	
	    if(rh->type == ENTRY_END) {
	    	printk("revirt_read_next_sysret: found end entry (#%lu)\n",
			rh->id);
		do_exit(1);
		//ztrace_debug_pause();
		goto out_setfs;
	    } else if(rh->type == ENTRY_SYSRET) {
		if(rh->len != sizeof(*le)) {
		    printk("revirt_read_ahead: unexpected len %d\n",
		    	rh->len);
		    goto out_setfs;
		}
		r=__revirt_read_data(rh, (char *)le, 
				&pos, rgroup);
		if(r<0) {
	    	    printk("__revirt_read_data returned %d\n", r);
		    rh->type = -1;
		    goto out_setfs;
		}

		Rprintk("next_sysret: le pid %d, sid %d\n",
			le->pid, le->syscall_id, 0, 0, 0);

		if(le->pid == pid)  {
		    *cache_syscall_id = le->syscall_id;
		    goto out_setfs;
		}
	    } else if(rh->type == ENTRY_EXIT) {
		if(rh->len != sizeof(ee)) {
		    printk("revirt_read_ahead: unexpected len %d\n",
		    	rh->len);
		    goto out_setfs;
		}
		r=__revirt_read_data(rh, (char *)&ee, &pos, rgroup);
		if(r<0)
			goto out_setfs;

		if(ee.pid == pid) {
		    Rprintk("revirt_read_next_sysret: found exit entry (#%lu) for pid %d, ending search.\n",
		    		rh->id, pid, 0, 0, 0);
		    goto out_setfs;
		}
	    } else if(rh->type == ENTRY_SCHED) {
		if(rh->len != sizeof(se)) {
		    printk("revirt_read_ahead: unexpected len %d\n",
		    	rh->len);
		    goto out_setfs;
		}

		r=__revirt_read_data(rh, (char *)&se, &pos, rgroup);
		if(r<0)
			goto out_setfs;

		if(se.pid == pid && se.syscall_id > current_syscall_id) {
		    Rprintk("revirt_read_next_sysret: found sched (#%lu) for pid %d, w/ sid %ld > %ld\n",
		    		rh->id, pid, se.syscall_id, current_syscall_id,0);
		    *cache_syscall_id = se.syscall_id;
		    goto out_setfs;
		}
	    } else {
	    	/* Don't bother to read the data, just skip over it... */
		pos+= rh->len;
	    }

	    /* play nice */	    
	    if(current->need_resched)
	    	schedule();
	}


       out_setfs:
	current->revirt_flags &= ~REPLAY_NO_YIELD_ON_BLOCK;
	set_fs(oldfs);

       //out:
       	return r;
}

int revirt_read_log(struct rlog_header * rh) {
	mm_segment_t oldfs;
	//struct rlog_header r;
	int l;
	struct revirt_group_struct * rgroup;
	//ssize_t (*read) (struct file *, char *, size_t, loff_t *);
	//loff_t (*llseek) (struct file *, loff_t, int);
	loff_t * ppos;
	//int i=0;

	if(!(rgroup=get_rgroup(current))) {
	    printk("revirt_read_log: no rgroup!\n");
	    return -EBADF;
	}

#if 0
	if(rgroup->cache.valid) {
		*rh=rgroup->cache.rh;
		rgroup->cache.valid=0;
		return rgroup->cache.l;
	}
#endif

	oldfs= get_fs();
	set_fs(KERNEL_DS);

	current->revirt_flags |= REPLAY_NO_YIELD_ON_BLOCK;

	//read= rgroup->scallf->f_op->read;
	//llseek= rgroup->scallf->f_op->llseek;
	/* This is a pointer so that read() can modify it in-place */
	ppos=&(rgroup->replay_pos);
	/*
	 * If we didn't split the schedule and syscall log entries
	 * into different files, skip scheduling entries.
	 */
#if 0
	while(1) {
	    l=__revirt_read_header(rh, ppos, rgroup);
	    if(l<0) {
	    	printk("revirt_read_log: Error reading log: %d\n", l);
		goto out;
	    }

	    check_filepos(rh, rgroup);
	    /*
	     * fork entries need to be read both by zr and the kernel revirt
	     * system.
	     *
	     * Actually, I suppose fork need not be read by zr, it could infer
	     * one by seeing a new pid, but that sounds sketchy...
	     */
	    if(IS_SCHED(rh->type) && rh->type!=ENTRY_FORK) {
		(*ppos) += rh->len;
	    } else
	    	break;

	    /* play nice */	    
	    if(current->need_resched)
	    	schedule();
	}
#endif
	if((l=__revirt_read_next_sys(rh, ppos, rgroup))<0) {
		goto  out_setfs;
	}
	l=__revirt_read_data(rh, revirt_scall_buf, ppos, rgroup);
	//l=(*read)(rgroup->scallf, revirt_scall_buf, r.len, ppos);

    out_setfs:	    
	current->revirt_flags &= ~REPLAY_NO_YIELD_ON_BLOCK;
	set_fs(oldfs);	

	if(rh->type == ENTRY_END) {
		printk("revirt_read_log: reached ENTRY_END, halting...\n");
		ztrace_debug_pause();
	}

      out:
	return l;
}
