/*
 * revirt/wraw.c
 *
 * Hacked from linux/drivers/char/raw.c
 *
 * Front-end raw character devices.  These can be bound to any block
 * devices to provide genuine Unix raw character device semantics.
 *
 * We reserve minor number 0 for a control interface.  ioctl()s on this
 * device are used to bind the other minor numbers to block devices.
 */

#include <linux/fs.h>
#include <linux/iobuf.h>
#include <linux/major.h>
#include <linux/blkdev.h>
#include <linux/raw.h>
#include <linux/capability.h>
#include <linux/smp_lock.h>
#include <asm/uaccess.h>
#include <linux/slab.h>

#include <revirt/wraw.h>

#define dprintk(x...) 

typedef struct wraw_device_data_s {
	struct block_device *binding;
	int inuse, sector_size, sector_bits;
	struct semaphore mutex;
} wraw_device_data_t;

static struct wraw_watch_s watch;
//static struct wraw_write_record wr_buf[WBUFSIZE];
//static char __data_buf[DBUFSIZE+4096];
//static char *__data_buf=NULL;
//static char * data_buf=NULL;

static wraw_device_data_t wraw_devices[256];

static ssize_t rw_wraw_dev(int rw, struct file *, char *, size_t, loff_t *);

ssize_t	wraw_read(struct file *, char *, size_t, loff_t *);
ssize_t	wraw_write(struct file *, const char *, size_t, loff_t *);
int	wraw_open(struct inode *, struct file *);
int	wraw_release(struct inode *, struct file *);
int	wraw_ctl_ioctl(struct inode *, struct file *, unsigned int, unsigned long);
int	wraw_ioctl(struct inode *, struct file *, unsigned int, unsigned long);


static struct file_operations wraw_fops = {
	read:		wraw_read,
	write:		wraw_write,
	open:		wraw_open,
	release:	wraw_release,
	ioctl:		wraw_ioctl,
};

static struct file_operations wraw_ctl_fops = {
	ioctl:		wraw_ctl_ioctl,
	open:		wraw_open,
};

static int __init wraw_init(void)
{
	int i;
	register_chrdev(WRAW_MAJOR, "wraw", &wraw_fops);

	for (i = 0; i < 256; i++) 
		init_MUTEX(&wraw_devices[i].mutex);


	/* FIXME -- Ideally we'd like one per device, with buffers 
	 * allocated when 'grabbed', and released when released.
	 */
	init_MUTEX(&watch.mutex);
	init_waitqueue_head(&watch.wait_read);
	init_waitqueue_head(&watch.wait_write);
	watch.wr_buf_u=NULL;
	watch.data_buf_u=NULL;
	watch.maxwr=watch.maxdata=watch.curwr=watch.curdata=0;
	if((watch.__data_buf_k=kmalloc(DBUFSIZE+4096, GFP_KERNEL))) 
	    watch.data_buf_k=
		((char *)(((unsigned long)watch.__data_buf_k + 4096) 
			    & ~(4096-1)));
	else {
	    printk("wraw: kmalloc failed, no data buffer allocated.\n");
	    watch.data_buf_k=NULL;
	}
	return 0;
}

__initcall(wraw_init);

int wraw_ioctl(struct inode *inode, 
		struct file *filp, 
		unsigned int command, 
		unsigned long argp) {
    //struct wraw_watch_s * wp;
    struct wraw_watch_arg_s arg;
    int minor, err;
    minor = MINOR(filp->f_dentry->d_inode->i_rdev);
    err=-EINVAL;
    //wp=&wraw_devices[minor].watch;

    switch(command) {
    	case WRAW_GRAB:
	    down(&watch.mutex);
	    if(watch.watched) {
	    	err=-EBUSY;
		up(&watch.mutex);
		goto out;
	    } else {
	    	watch.watched=1;
		watch.tsk=current;
	    }
	    err=0;
	    up(&watch.mutex);
	    break;	    
    	case WRAW_LISTEN:
	    down(&watch.mutex);
	    if(!watch.watched || watch.tsk!=current) {
		err=-EINVAL;
		goto out_up;
	    }
	    if(copy_from_user(&arg, (void *)argp, sizeof(arg))) {
	    	err=-EFAULT;
		goto out_up;
	    }
	    if(arg.wraw_data_buf!=NULL) {
		if(watch.data_buf_k==NULL) {
			err=-ENODATA;
			goto out_up;
		}
		watch.data_buf_u=arg.wraw_data_buf;
		watch.maxdata=(arg.maxdata<DBUFSIZE) ? arg.maxdata : DBUFSIZE;
	    }

	    watch.wr_buf_u=arg.wraw_wr_buf;
	    watch.maxwr=(arg.maxwr < WBUFSIZE) ? arg.maxwr : WBUFSIZE;
	    watch.curwr=watch.curdata=0;

	    err=0;

	    /*
	     * FIXME -- this is a race condition for SMP or pre-emptible
	     */
	    wake_up_interruptible(&watch.wait_write);
	    while((watch.curwr<watch.maxwr) && (err==0)) {
                up(&watch.mutex);
		/* FIXME -- SMP / preemptible kernel race here */
                err=wait_event_interruptible(watch.wait_read, 
			watch.curwr >= watch.maxwr);
                down(&watch.mutex); 
	    }
	    dprintk("copying %d records to 0x%x, err=%d\n", watch.curwr, watch.wr_buf_u, err);
	    if(err==-ERESTARTSYS)
		    err=-EINTR;
	    if(copy_to_user(watch.wr_buf_u, watch.wr_buf_k,
		sizeof(struct wraw_write_record) * watch.curwr)) {
		    err=-EFAULT;
		    goto done;
	    }
	    if(watch.data_buf_u) {
	     if(copy_to_user(watch.data_buf_u, watch.data_buf_k, 
				       watch.curdata)){
		    err=-EFAULT;
		    goto done;
	     }
	    }
	    arg.maxwr=watch.curwr;
	    arg.maxdata=watch.curdata;
	    if(copy_to_user((void *)argp, &arg, sizeof(arg))) {
		    err=-EFAULT;
		    goto done;
	    }
	   done:

	    watch.wr_buf_u=NULL;
	    watch.data_buf_u=NULL;
	    watch.maxwr=watch.curwr=watch.maxdata=watch.curdata=0;
	    break;
	case WRAW_RELEASE:
	    down(&watch.mutex);
	    watch.watched=0;
	    watch.wr_buf_u=NULL;
	    watch.data_buf_u=NULL;
	    watch.maxwr=0;
	    watch.maxdata=0;
	    wake_up_interruptible(&watch.wait_read);
	    wake_up_interruptible(&watch.wait_write);
	    err=0;
	    break;
	default:
	    err=-EINVAL;
	    goto out;
    }

    out_up:
    	up(&watch.mutex);

    out:
       return err;
}

/* 
 * Open/close code for raw IO.
 */

int wraw_open(struct inode *inode, struct file *filp)
{
	int minor;
	struct block_device * bdev;
	kdev_t rdev;	/* it should eventually go away */
	int err;
	int sector_size;
	int sector_bits;

	minor = MINOR(inode->i_rdev);
	
	/* 
	 * Is it the control device? 
	 */
	
	if (minor == 0) {
		filp->f_op = &wraw_ctl_fops;
		return 0;
	}
	
	if (!filp->f_iobuf) {
		err = alloc_kiovec(1, &filp->f_iobuf);
		if (err)
			return err;
	}

	down(&wraw_devices[minor].mutex);
	/*
	 * No, it is a normal wraw device.  All we need to do on open is
	 * to check that the device is bound, and force the underlying
	 * block device to a sector-size blocksize. 
	 */

	bdev = wraw_devices[minor].binding;
	err = -ENODEV;
	if (!bdev)
		goto out;

	atomic_inc(&bdev->bd_count);
	rdev = to_kdev_t(bdev->bd_dev);
	err = blkdev_get(bdev, filp->f_mode, 0, BDEV_RAW);
	if (err)
		goto out;
	
	/*
	 * Don't change the blocksize if we already have users using
	 * this device 
	 */

	if (wraw_devices[minor].inuse++)
		goto out;

	/* 
	 * Don't interfere with mounted devices: we cannot safely set
	 * the blocksize on a device which is already mounted.  
	 */
	
	sector_size = 512;
	if (is_mounted(rdev)) {
		if (blksize_size[MAJOR(rdev)])
			sector_size = blksize_size[MAJOR(rdev)][MINOR(rdev)];
	} else {
		if (hardsect_size[MAJOR(rdev)])
			sector_size = hardsect_size[MAJOR(rdev)][MINOR(rdev)];
	}

	set_blocksize(rdev, sector_size);
	wraw_devices[minor].sector_size = sector_size;

	for (sector_bits = 0; !(sector_size & 1); )
		sector_size>>=1, sector_bits++;
	wraw_devices[minor].sector_bits = sector_bits;

 out:
	up(&wraw_devices[minor].mutex);
	
	return err;
}

int wraw_release(struct inode *inode, struct file *filp)
{
	int minor;
	struct block_device *bdev;
	
	minor = MINOR(inode->i_rdev);
	down(&wraw_devices[minor].mutex);
	bdev = wraw_devices[minor].binding;
	wraw_devices[minor].inuse--;
	up(&wraw_devices[minor].mutex);
	blkdev_put(bdev, BDEV_RAW);
	return 0;
}



/*
 * Deal with ioctls against the raw-device control interface, to bind
 * and unbind other raw devices.  
 */

int wraw_ctl_ioctl(struct inode *inode, 
		  struct file *flip,
		  unsigned int command, 
		  unsigned long arg)
{
	struct raw_config_request rq;
	int err = 0;
	int minor;
	
	switch (command) {
	case RAW_SETBIND:
	case RAW_GETBIND:

		/* First, find out which raw minor we want */

		err = copy_from_user(&rq, (void *) arg, sizeof(rq));
		if (err)
			break;
		
		minor = rq.raw_minor;
		if (minor <= 0 || minor > MINORMASK) {
			err = -EINVAL;
			break;
		}

		if (command == RAW_SETBIND) {
			/*
			 * This is like making block devices, so demand the
			 * same capability
			 */
			if (!capable(CAP_SYS_ADMIN)) {
				err = -EPERM;
				break;
			}

			/* 
			 * For now, we don't need to check that the underlying
			 * block device is present or not: we can do that when
			 * the raw device is opened.  Just check that the
			 * major/minor numbers make sense. 
			 */

			if ((rq.block_major == NODEV && 
			     rq.block_minor != NODEV) ||
			    rq.block_major > MAX_BLKDEV ||
			    rq.block_minor > MINORMASK) {
				err = -EINVAL;
				break;
			}
			
			down(&wraw_devices[minor].mutex);
			if (wraw_devices[minor].inuse) {
				up(&wraw_devices[minor].mutex);
				err = -EBUSY;
				break;
			}
			if (wraw_devices[minor].binding)
				bdput(wraw_devices[minor].binding);
			wraw_devices[minor].binding = 
				bdget(kdev_t_to_nr(MKDEV(rq.block_major, rq.block_minor)));
			up(&wraw_devices[minor].mutex);
		} else {
			struct block_device *bdev;
			kdev_t dev;

			bdev = wraw_devices[minor].binding;
			if (bdev) {
				dev = to_kdev_t(bdev->bd_dev);
				rq.block_major = MAJOR(dev);
				rq.block_minor = MINOR(dev);
			} else {
				rq.block_major = rq.block_minor = 0;
			}
			err = copy_to_user((void *) arg, &rq, sizeof(rq));
		}
		break;
		
	default:
		err = -EINVAL;
	}
	
	return err;
}



ssize_t	wraw_read(struct file *filp, char * buf, 
		 size_t size, loff_t *offp)
{
	return rw_wraw_dev(READ, filp, buf, size, offp);
}

ssize_t	wraw_write(struct file *filp, const char *buf, 
		  size_t size, loff_t *offp)
{
	return rw_wraw_dev(WRITE, filp, (char *) buf, size, offp);
}

#define SECTOR_BITS 9
#define SECTOR_SIZE (1U << SECTOR_BITS)
#define SECTOR_MASK (SECTOR_SIZE - 1)

ssize_t	rw_wraw_dev(int rw, struct file *filp, char *buf, 
		   size_t size, loff_t *offp)
{
	struct kiobuf * iobuf;
	int		new_iobuf;
	int		err = 0;
	unsigned long	blocknr, blocks;
	size_t		transferred=0;
	int		iosize;
	int		i;
	int		minor;
	kdev_t		dev;
	unsigned long	limit;

	int		sector_size, sector_bits, sector_mask;
	int		max_sectors;
	loff_t		old_off = 0xDEADBEEF;
	
	/*
	 * First, a few checks on device size limits 
	 */

	minor = MINOR(filp->f_dentry->d_inode->i_rdev);

	new_iobuf = 0;
	iobuf = filp->f_iobuf;
	if (test_and_set_bit(0, &filp->f_iobuf_lock)) {
		/*
		 * A parallel read/write is using the preallocated iobuf
		 * so just run slow and allocate a new one.
		 */
		err = alloc_kiovec(1, &iobuf);
		if (err)
			goto out;
		new_iobuf = 1;
	}

	dev = to_kdev_t(wraw_devices[minor].binding->bd_dev);
	sector_size = wraw_devices[minor].sector_size;
	sector_bits = wraw_devices[minor].sector_bits;
	sector_mask = sector_size- 1;
	max_sectors = KIO_MAX_SECTORS >> (sector_bits - 9);
	
	if (blk_size[MAJOR(dev)])
		limit = (((loff_t) blk_size[MAJOR(dev)][MINOR(dev)]) << BLOCK_SIZE_BITS) >> sector_bits;
	else
		limit = INT_MAX;
	dprintk ("rw_wraw_dev: dev %d:%d (+%d)\n",
		 MAJOR(dev), MINOR(dev), limit);
	
	err = -EINVAL;
	if ((*offp & sector_mask) || (size & sector_mask))
		goto out_free;
	err = 0;
	if (size)
		err = -ENXIO;
	if ((*offp >> sector_bits) >= limit)
		goto out_free;

	/*
	 * Split the IO into KIO_MAX_SECTORS chunks, mapping and
	 * unmapping the single kiobuf as we go to perform each chunk of
	 * IO.  
	 */

	transferred = 0;
	old_off=*offp;
	blocknr = *offp >> sector_bits;
	while (size > 0) {
		blocks = size >> sector_bits;
		if (blocks > max_sectors)
			blocks = max_sectors;
		if (blocks > limit - blocknr)
			blocks = limit - blocknr;
		if (!blocks)
			break;

		iosize = blocks << sector_bits;

		err = map_user_kiobuf(rw, iobuf, (unsigned long) buf, iosize);
		if (err)
			break;

		for (i=0; i < blocks; i++) 
			iobuf->blocks[i] = blocknr++;
		
		err = brw_kiovec(rw, 1, &iobuf, dev, iobuf->blocks, sector_size);

		if (rw == READ && err > 0)
			mark_dirty_kiobuf(iobuf, err);
		
		if (err >= 0) {
			transferred += err;
			size -= err;
			buf += err;
		}

		unmap_kiobuf(iobuf);

		if (err != iosize)
			break;
	}
	
	if (transferred) {
		*offp += transferred;
		err = transferred;
	}

 out_free:
	if (!new_iobuf)
		clear_bit(0, &filp->f_iobuf_lock);
	else
		free_kiovec(1, &iobuf);

	/*
	 *  We don't want to block with filp->f_iobuf_lock locked...
	 */
	if(transferred) {
		if(watch.watched) {
		    int intr=0;
		    down(&watch.mutex);
		    while(watch.curwr >= watch.maxwr 
				    && watch.watched) {
		    	/*
			 * FIXME -- this is a race in SMP or pre-emptible kern
			 */
		    	up(&watch.mutex);
			intr=wait_event_interruptible(watch.wait_write,
				(watch.curwr < watch.maxwr) || !watch.watched);
			if(intr || watch.watched==0)
			    goto out; /* mutex is not held, we're OK */
			down(&watch.mutex);
		    }
		   watch.wr_buf_k[watch.curwr].offset=old_off;
		   watch.wr_buf_k[watch.curwr].len=transferred;
		    watch.curwr++;
		    if(watch.curwr >= watch.maxwr)
		    	wake_up_interruptible(&watch.wait_read);
		    up(&watch.mutex);
		}
	}

 out:	
	return err;
}
