/*
 * perfctr.c - Provide abstract access to arch-dependent hardware counters
 *
 * Copyright (C) 2002 George Dunlap
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/types.h>
#include <linux/config.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/smp_lock.h>
#include <linux/file.h>
#include <asm/uaccess.h>
#include <asm/pgtable.h>
#include <asm/perfctr.h>
#include <linux/slab.h>
#include <linux/sched.h>
#include <revirt/rlog.h>
#include <revirt/perfctr.h>

/*
 * Software counters
 *
 * revirt_exceptions1: those without an error code
 * revirt_exceptions2: with an error code
 *    (This distinction is useful for some alternate P4 counters)
 * revirt_edi: where a copy_to_user was interrupted
 */
unsigned int revirt_interrupts;
unsigned int revirt_exceptions1;
unsigned int revirt_exceptions2;
#ifdef REPLAY_RECORD_ALL_EXCEPTIONS
unsigned int revirt_ecount[REVIRT_NR_EX];
#endif
unsigned int revirt_edi=0;
unsigned int revirt_current_syscall;

void perfctr_set_msr(unsigned int addr, s64 x) { 
	u32 hi, lo;
	lo = (u32) x;
	hi = (u32) ((x-lo) >> 32);
        wrmsr(addr, lo, hi);
}

void perfctr_get_msr(unsigned int addr, s64 * x) {
	u32 hi, lo;
        rdmsr(addr, lo, hi);
	*x = ((s64)(hi) << 32) + lo;
}

void perfctr_set_ctr(unsigned int addr, s64 x) { 
	u32 hi, lo;
	lo = (u32) x;
	hi = (u32) ((x-lo) >> 32);
        wrmsr(addr, lo, hi);
}
/*
 * Like get_msr, but sign extended according to the platform
 */
void perfctr_get_ctr(unsigned int addr, s64 * x) {
	u32 hi, lo;
        rdmsr(addr, lo, hi);
#if defined(CONFIG_MPENTIUM4) || \
	defined(CONFIG_M686) || defined(CONFIG_MPENTIUMIII)
	/*
	 * P4 and P6 perfctrs are only 40 bits, so we need to 
	 * extend hi from 8 to 32 bits if negative
	 */
	if(hi & 0x80) {
	    hi |= 0xFFFFFF00L;
	}
#elif defined(CONFIG_MK6) || defined(CONFIG_MK7)
	/*
	 * Athalon perfctrs, OTOH, are 48 bits...
	 */
	if(hi & 0x8000) {
	    hi |= 0xFFFF0000L;
	}
#else
# warning "No sign extension for performance counters!"
#endif
	*x = ((s64)(hi) << 32) + lo;
}


void save_perfctr_regs(struct task_struct * prev) {
	int i;
	struct revirt_tsk_s * rtsk = prev->revirt_tsk;
	if(rtsk==NULL) 
		return;
	for(i=0; i<revirt_pfctr.count; i++) {
		perfctr_get_ctr(revirt_pfctr.list[i].counter_msr,
				&rtsk->perfctr[i].ctr);
	}
	if(revirt_pfctr.subtract_exceptions) {
	    rtsk->swctr[0]=revirt_interrupts;
	    rtsk->swctr[1]=revirt_exceptions1;
	    rtsk->swctr[2]=revirt_exceptions2;
	}
	rtsk->current_syscall=revirt_current_syscall;
#ifdef REPLAY_RECORD_ALL_EXCEPTIONS
	for(i=0; i<REVIRT_NR_EX; i++)
	    rtsk->ecount[i]=revirt_ecount[i];
#endif
}

void restore_perfctr_regs(struct task_struct * next, struct task_struct * prev) {
	int i ;
	struct revirt_tsk_s * rtsk = next->revirt_tsk;

	/*
	 * The problem with just not touching counters is that if we swap out a task
	 * waiting for an overflow, the interrupt may happen in the next process, even
	 * though the current process is still mostly at a negative.  Setting it to
	 * zero will prevent this.
	 */
	if(rtsk==NULL) {
	    if(prev && (prev->revirt_flags & (REPLAY_LOG|REPLAY_REPLAY))) {
		for(i=0; i<revirt_pfctr.count; i++)
			perfctr_set_ctr(revirt_pfctr.list[i].counter_msr, 0);
	    }
	    return;
	}
	/*
	 * Hardware counters.
	 */
	for(i=0; i<revirt_pfctr.count; i++) {
		perfctr_set_ctr(revirt_pfctr.list[i].counter_msr,
				rtsk->perfctr[i].ctr);
	}
	/*
	 * Software counters
	 */
	if(revirt_pfctr.subtract_exceptions) {
	    revirt_interrupts=rtsk->swctr[0];
	    revirt_exceptions1=rtsk->swctr[1];
	    revirt_exceptions2=rtsk->swctr[2];
	}
	    revirt_current_syscall=rtsk->current_syscall;

#ifdef REPLAY_RECORD_ALL_EXCEPTIONS
	for(i=0; i<REVIRT_NR_EX; i++)
	    revirt_ecount[i]=rtsk->ecount[i];
#endif
}

void perfctr_enable_counters(void) {
	int i;
	unsigned int lo, hi;
	for(i=0; i<revirt_pfctr.count; i++) {
#if defined(CONFIG_MPENTIUM4)
		perfctr_set_msr(revirt_pfctr.list[i].arch.p4.cccr_msr,
				revirt_pfctr.list[i].arch.p4.cccr_val);
		perfctr_set_msr(revirt_pfctr.list[i].arch.p4.escr_msr,
				revirt_pfctr.list[i].arch.p4.escr_val);
#elif defined(CONFIG_MK6) || defined (CONFIG_MK7)
		perfctr_set_msr(revirt_pfctr.list[i].arch.athlon.evtsel_msr,
				revirt_pfctr.list[i].arch.athlon.evtsel_val);
#elif defined(CONFIG_M686) || defined (CONFIG_MPENTIUMIII)
		perfctr_set_msr(revirt_pfctr.list[i].arch.p6.evtsel_msr,
				revirt_pfctr.list[i].arch.p6.evtsel_val);
#else
# error "Don't know how to deal with this architecture's performance counters!"
#endif
	}
#if defined(CONFIG_MPENTIUM4)
# warning "Turning on P4 FOPCODE Compatability mode"
	/* 
	 * Not perfctr, but important: make sure FOP code is the same
	 * See ia32-1:8.1.8.1 "FOPCODE Compatability Mode"
	 * and ia32-3:B-7, IA32_MISC_ENABLE
	 */
#define IA32_MISC_ENABLE	0x1a0
#define IA32_FOP_COMPAT		(1 << 2)
	rdmsr(IA32_MISC_ENABLE, lo, hi);
	lo |= IA32_FOP_COMPAT;
	wrmsr(IA32_MISC_ENABLE, lo, hi);
#endif
}

void perfctr_clear_counters(void) {
	int i;

	/*
	 * Clear all of them, only count some...
	 */
	for(i=0; i<revirt_pfctr.count; i++) {
		perfctr_set_ctr(revirt_pfctr.list[i].counter_msr, 0);
	}

	revirt_interrupts=0;
	revirt_exceptions1=0;
	revirt_exceptions2=0;

#ifdef REPLAY_RECORD_ALL_EXCEPTIONS
	for(i=0; i<REVIRT_NR_EX; i++)
	    revirt_ecount[i]=0;
#endif
}

void perfctr_clear_ovf(void) {
	int l, h;
#if defined(CONFIG_MPENTIUM4)
	rdmsr(revirt_pfctr.overflow.control_msr, l, h);
	l &= revirt_pfctr.overflow.clear_mask;
	wrmsr(revirt_pfctr.overflow.control_msr, l, h);
#endif
	l=apic_read(APIC_LVTPC);
	l &= ~(APIC_LVT_MASKED);
	apic_write_around(APIC_LVTPC, l);
}

void perfcntr_enable_overflow(void) {
	int l, h;

	/*
	 * Shouldn't be strictly necessary, but couldn't hurt...
	 */
#if defined(CONFIG_MPENTIUM4)
	rdmsr(revirt_pfctr.overflow.control_msr, l, h);
	l &= revirt_pfctr.overflow.clear_mask;
	wrmsr(revirt_pfctr.overflow.control_msr, l, h);
#endif
	rdmsr(revirt_pfctr.overflow.control_msr, l, h);
	l |= revirt_pfctr.overflow.enable_bit;
	wrmsr(revirt_pfctr.overflow.control_msr, l, h);
}

void perfcntr_disable_overflow(void) {
	int l, h;

	rdmsr(revirt_pfctr.overflow.control_msr, l, h);
	l &= ~(revirt_pfctr.overflow.enable_bit);
	wrmsr(revirt_pfctr.overflow.control_msr, l, h);
}
