#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/unistd.h>
#include <linux/string.h>
#include <linux/types.h>
#include <linux/param.h>
#include <linux/fs.h>
#include <linux/slab.h>

#include <asm/uaccess.h>

#include <asm/signal.h>
#include <linux/mman.h>

#include <asm/system.h>
#include <asm/segment.h>

#include <asm/semaphore.h>

#include <linux/file.h>

#include "myutil.h"
#include "checkpoint.h"
#include "cow_disk.h"

#define COW_OFFSET_TO_IDX(offset) ((offset) >> COW_BLOCK_SHIFT)

extern asmlinkage long sys_unlink(const char * pathname);
void writeDiskFile(struct file *file, loff_t offset, void *data);

typedef enum {
  NAME_UNDO,
  NAME_REDO,
  NAME_DATA
} undo_redo_name_t;

struct checkpoint_struct {
  undo_redo_t undo;
  undo_redo_t redo;
  unsigned long long id;
  struct checkpoint_struct *next;
  struct checkpoint_struct *prev;
  unsigned long refCount;
};


#define NUM_CHARS_LONG_LONG 64
#define UNDO_STR "undo"
#define REDO_STR "redo"
#define DATA_STR "data"
#define MAX_UNDO_REDO_STRLEN 4

static char *createBasename(char *baseName, unsigned long long id, 
			    undo_redo_name_t urn) {
  unsigned long len;
  char *name;

  len = strlen(baseName); 
  len += NUM_CHARS_LONG_LONG + 2;
  len += MAX_UNDO_REDO_STRLEN + 2;
  len += 1;

  name = util_alloc_small(len);
  if(urn == NAME_UNDO) {
    sprintf(name,"%s.%s.%llu", baseName, UNDO_STR,id);
  } else if(urn == NAME_REDO) {
    sprintf(name,"%s.%s.%llu", baseName, REDO_STR,id);
  } else if(urn == NAME_DATA) {
    sprintf(name,"%s.%s.%llu", baseName, DATA_STR,id);
  } else {
    assert(0);
  }

  return name;
}

static checkpoint_t cp_new_open(char *baseName, unsigned long long id, int new) {
  char *undoName, *redoName, *dataName;
  checkpoint_t cp = util_alloc_small(sizeof(struct checkpoint_struct));
  
  // init the ref count
  cp->refCount = 1;
  cp->id = id;

  // now open up undo and redo files
  undoName = createBasename(baseName, id, NAME_UNDO);
  redoName = createBasename(baseName, id, NAME_REDO);
  dataName = createBasename(baseName, id, NAME_DATA);

  if(new) {
    ur_new(undoName, redoName, dataName, &cp->undo, &cp->redo);
  } else {
    ur_open(undoName, redoName, dataName, &cp->undo, &cp->redo);
  }
  assert((cp->undo != NULL) && (cp->redo != NULL));

  util_free_small(undoName);
  util_free_small(redoName);
  util_free_small(dataName);

  // nullify the links for the linked list
  cp->next = NULL;
  cp->prev = NULL;
  
  return cp;  
}

/**
 * these are the checkpoint object functions
 */
// these are just to make zeroing out the arguments and such a little easier
checkpoint_t cp_new(char *baseName, unsigned long long id) {
  return cp_new_open(baseName, id, 1);
}
checkpoint_t cp_open(char *baseName, unsigned long long id) {
  return cp_new_open(baseName, id, 0);
}

unsigned long cp_get_ref_count(checkpoint_t cp) {
  return cp->refCount;
}

void cp_increment_ref_count(checkpoint_t cp) {
  cp->refCount++;
}

unsigned long cp_delete(checkpoint_t cp, int unlinkUndoRedoFiles) {
  unsigned long finalRefCount;
  assert(cp->refCount > 0);
  cp->refCount--;
  finalRefCount = cp->refCount;
  if(cp->refCount == 0) {
    if(unlinkUndoRedoFiles) {
      ur_close_unlink(cp->undo);
      ur_close_unlink(cp->redo);
    } else {
      ur_close(cp->undo);
      ur_close(cp->redo);
    }
    util_free_small(cp);
  }
  return finalRefCount;
}

undo_redo_t cp_get_undo(checkpoint_t cp) {
  return cp->undo;
}

undo_redo_t cp_get_redo(checkpoint_t cp) {
  return cp->redo;
}
checkpoint_t cp_get_next(checkpoint_t cp) {
  return cp->next;
}
checkpoint_t cp_get_prev(checkpoint_t cp) {
  return cp->prev;
}
void cp_set_next(checkpoint_t cp, checkpoint_t next) {
  cp->next = next;
}
void cp_set_prev(checkpoint_t cp, checkpoint_t prev) {
  cp->prev = prev;
}
unsigned long long cp_get_id(checkpoint_t cp) {
  return cp->id;
}

void cp_insert(checkpoint_t prevCp, checkpoint_t newCp) {
  assert(prevCp != NULL);
  assert(newCp != NULL);

  newCp->next = prevCp->next;
  newCp->prev = prevCp;
  if(prevCp->next != NULL) {
    prevCp->next->prev = newCp;
  }
  prevCp->next = newCp;
}

void cp_remove(checkpoint_t cp) {
  checkpoint_t prevCp = cp->prev;
  checkpoint_t nextCp = cp->next;

  if(prevCp != NULL) {
    prevCp->next = nextCp;
  }

  if(nextCp != NULL) {
    nextCp->prev = prevCp;
  }

  cp->next = NULL;
  cp->prev = NULL;
}

/**
 * these are for the undo redo file objects
 */

#define UNDO_REDO_TIME_EXT       ".urt"

//#define UNDO_RECORD_SIZE (sizeof(char) + sizeof(loff_t) + sizeof(unsigned long long) + COW_BLOCK_SIZE)
// this tells the offset within a record of the valid field
//#define VALID_RECORD_OFFSET 0

struct undo_redo_struct {
  data_file_t blockData;
  data_file_t metaData;
  fileUtil_t timeFile;
  char *timeFileName;
  struct sort_element *findData;
  char *baseName;
  unsigned long baseNameLen;
};

struct ur_meta_data_struct {
  loff_t offset;
  unsigned long long id;
};


static undo_redo_t undo_redo_file(char *fileName, int new) {
  undo_redo_t ur = (undo_redo_t) util_alloc_small(sizeof(struct undo_redo_struct));
  memset(ur, 0, sizeof(*ur));

  // set up the meta data
  if(new) {
    ur->metaData = df_new(fileName, sizeof(struct ur_meta_data_struct));
  } else {
    ur->metaData = df_open(fileName, sizeof(struct ur_meta_data_struct));
    if(ur->metaData == NULL) {
      util_free_small(ur);
      ur = NULL;
    }
  }

  // now open up our time file so we can keep track of the latest 
  // writes to blocks.  This is going to be a sparse file and only
  // data in the bitmap will be here
  if(ur != NULL) {
    ur->timeFileName = util_alloc_small(strlen(fileName) + 
					strlen(UNDO_REDO_TIME_EXT) + 1);
    sprintf(ur->timeFileName, "%s%s", fileName, UNDO_REDO_TIME_EXT);
    if(new) {
      ur->timeFile = util_open(ur->timeFileName, O_LARGEFILE | O_RDWR | 
			       O_CREAT | O_TRUNC, 0600);
      assert(ur->timeFile != NULL);
    } else {
      ur->timeFile = util_open(ur->timeFileName, O_LARGEFILE | O_RDWR,
			       0600);
      // data file already open, this should be there
      assert(ur->timeFile != NULL);
    }
  }

  if(ur != NULL) {
    ur->baseNameLen = strnlen(fileName, UTIL_MAX_PATH) + 1;
    ur->baseName = util_alloc_small(ur->baseNameLen);
    sprintf(ur->baseName,"%s",fileName);
  }
      
  return ur;
}

void ur_new(char *undoFileName, char *redoFileName, char *blockDataFileName,
	    undo_redo_t *undo, undo_redo_t *redo) {
  data_file_t df;

  *undo = undo_redo_file(undoFileName, 1);
  df = df_new(blockDataFileName, COW_BLOCK_SIZE);
  (*undo)->blockData = df;

  if(redo != NULL) {
    *redo = undo_redo_file(redoFileName,1);
    df_inc_object_ref_count(df);
    (*redo)->blockData = df;
  }
}

void ur_open(char *undoFileName, char *redoFileName, char *blockDataFileName,
	     undo_redo_t *undo, undo_redo_t *redo) {
  data_file_t df;
    
  *undo = undo_redo_file(undoFileName, 0);
  df = df_open(blockDataFileName, COW_BLOCK_SIZE);
  (*undo)->blockData = df;

  if(redo != NULL) {
    *redo = undo_redo_file(redoFileName,0);
    df_inc_object_ref_count(df);
    (*redo)->blockData = df;
  }
}

// XXX FIXME this is bad if the blockData is shared
void ur_reset(undo_redo_t *ur, char *dataFileName) {
  char *fileName = util_alloc_small((*ur)->baseNameLen);
  sprintf(fileName,"%s",(*ur)->baseName);
  ur_close_unlink(*ur);
  *ur = undo_redo_file(fileName,1);
  (*ur)->blockData = df_new(dataFileName, COW_BLOCK_SIZE);
  util_free_small(fileName);
}

void ur_reset_time(undo_redo_t ur) {
  util_truncate(ur->timeFile, 0);
}

static void ur_local_close(undo_redo_t ur, int unlink) {
  if(unlink) {
    df_close_unlink(ur->blockData);
    df_close_unlink(ur->metaData);
    util_unlink(ur->timeFileName);
  } else {
    df_close(ur->blockData);
    df_close(ur->metaData);
  }

  util_close(ur->timeFile);
  util_free_small(ur->timeFileName);
  util_free_small(ur->baseName);
  ur_free_find_data(ur);
  util_free_small(ur);
}  

void ur_close_unlink(undo_redo_t ur) {
  ur_local_close(ur, 1);
}

void ur_close(undo_redo_t ur) {
  ur_local_close(ur, 0);
}

unsigned long ur_get_num_rec(undo_redo_t ur) {
  return df_get_num_rec(ur->metaData);
}

static void ur_ll_get_rec(undo_redo_t ur, unsigned long idx, loff_t *offset, 
			  unsigned long long *id, void *data) {
  struct ur_meta_data_struct meta;

  assert(idx < ur_get_num_rec(ur));

  // first read in the meta data
  df_get_rec_by_idx(ur->metaData, idx, &meta, NULL);
  *offset = meta.offset;
  *id = meta.id;

  // read in the data
  if(data != NULL) {
    df_get_rec(ur->blockData, COW_OFFSET_TO_IDX(*offset), data);
  }
}


void ur_get_rec(undo_redo_t ur, unsigned long idx, loff_t *offset,
		unsigned long long *id, void *data) {
  ur_ll_get_rec(ur, idx, offset, id, data);
}
void ur_get_rec_meta(undo_redo_t ur,unsigned long idx, loff_t *offset,
		     unsigned long long *id) {
  ur_ll_get_rec(ur, idx, offset, id, NULL);
}

int ur_is_block_present(undo_redo_t ur, loff_t offset) {
  return df_is_rec_present(ur->metaData, COW_OFFSET_TO_IDX(offset));
}

void ur_add_block(undo_redo_t ur, loff_t offset, unsigned long long id, void *data) {
  struct ur_meta_data_struct meta;

  assert(offset >= 0);
  assert((((unsigned long) offset) % COW_BLOCK_SIZE) == 0);
  assert(!ur_is_block_present(ur, offset));

  meta.offset = offset;
  meta.id = id;

  df_add_rec(ur->metaData, COW_OFFSET_TO_IDX(offset), &meta);
  if(df_is_rec_present(ur->blockData, COW_OFFSET_TO_IDX(offset))) {
    df_inc_rec_ref_count(ur->blockData, COW_OFFSET_TO_IDX(offset));
  } else {
    df_add_rec(ur->blockData, COW_OFFSET_TO_IDX(offset), data);
  }
  

  ur_free_find_data(ur);
}

static void init_sort_data(undo_redo_t ur) {
  unsigned long idx, numRec;
  loff_t tmpOffset;
  unsigned long long tmpId;

  // first, check if we need to create the find data
  if(ur->findData == NULL) {
    ur->findData = util_alloc_large(sizeof(struct sort_element) * ur_get_num_rec(ur));
    numRec = ur_get_num_rec(ur);
    for(idx = 0; idx < numRec; idx++) {
      ur_get_rec_meta(ur, idx, &tmpOffset, &tmpId);
      ur->findData[idx].idx = idx;
      ur->findData[idx].key = tmpOffset;
    }
    util_heap_sort(ur->findData, (int) numRec);
  }
}
unsigned long ur_find_block(undo_redo_t ur, loff_t offset) {
  assert(ur_is_block_present(ur, offset));

  return df_find_rec(ur->metaData, COW_OFFSET_TO_IDX(offset));
}
void ur_get_rec_meta_sorted(undo_redo_t ur,unsigned long idx, loff_t *offset,
			    unsigned long long *id) {
  assert(ur_get_num_rec(ur) > 0);
  assert(idx < ur_get_num_rec(ur));
  init_sort_data(ur);

  ur_get_rec_meta(ur, ur->findData[idx].idx, offset, id);
}
void ur_get_rec_sorted(undo_redo_t ur, unsigned long idx, loff_t *offset, 
		       unsigned long long *id, void *data) {
  assert(ur_get_num_rec(ur) > 0);
  assert(idx < ur_get_num_rec(ur));
  init_sort_data(ur);

  ur_get_rec(ur, ur->findData[idx].idx, offset, id, data);
}

void ur_free_find_data(undo_redo_t ur) {
  if(ur->findData != NULL) {
    util_free_large(ur->findData);
    ur->findData = NULL;
  }
}

/*
static int ur_is_valid(undo_redo_t ur, unsigned long idx) {
  assert(0);
  return 0;
}
static void ur_set_valid(undo_redo_t ur, unsigned long idx, char valid) {
  assert(0);
}
*/

void ur_update_time(undo_redo_t ur, loff_t offset, unsigned long long id) {
  int ret;
  assert(ur_is_block_present(ur, offset));
  loff_t timeOffset;

  timeOffset = COW_OFFSET_TO_IDX(offset);
  timeOffset *= sizeof(id);

  ret = util_write(ur->timeFile, &id, sizeof(id), timeOffset);
  assert(ret == sizeof(id));
}

unsigned long long ur_get_time(undo_redo_t ur, loff_t offset) {
  unsigned long long id;
  int ret;

  loff_t timeOffset = COW_OFFSET_TO_IDX(offset);
  timeOffset *= sizeof(id);

  assert(ur_is_block_present(ur,offset));

  ret = util_read(ur->timeFile, &id, sizeof(id), timeOffset);

  return id;
}

void ur_invalidate(undo_redo_t ur, unsigned long idx) {
  struct ur_meta_data_struct meta;

  df_get_rec_by_idx(ur->metaData, idx, &meta, NULL);
  
  df_invalidate(ur->metaData, COW_OFFSET_TO_IDX(meta.offset));
  df_invalidate(ur->blockData, COW_OFFSET_TO_IDX(meta.offset));
}
void ur_remove_invalid(undo_redo_t ur) {
  df_remove_invalid(ur->metaData);
  df_remove_invalid(ur->blockData);
}



// now this is for our sparse file.  There is only one of these per cow disk
#define SPECDATA_FILE_EXT ".spd"
#define SPECBIT_FILE_EXT  ".spb"
#define SPEC_NUM_BYTES_EXT 4
#define NUM_SPEC_BLOCKS 4000

struct sparse_file_struct {
  int fd;
  struct sort_element offsetList[NUM_SPEC_BLOCKS];
  unsigned long numBlocks;
  bitmap_t bitmap;
  unsigned long long numBlockFlush; // for debugging
};

sparse_file_t sf_new(char *baseName) {
  sparse_file_t sf;
  char *fileName;
  unsigned long len;
  mm_segment_t old_fs;

  sf = util_alloc_large(sizeof(struct sparse_file_struct));

  sf->numBlocks = 0;
  sf->numBlockFlush = 0;
  
  // open up the sparse file
  len = strnlen(baseName,UTIL_MAX_PATH);
  len += SPEC_NUM_BYTES_EXT + 1;
  fileName = (char *) util_alloc_small(len);

  sprintf(fileName,"%s%s",baseName, SPECDATA_FILE_EXT);
  
  old_fs = get_fs();
  set_fs(get_ds());
  sf->fd = sys_open(fileName, O_LARGEFILE | O_RDWR | O_CREAT |O_TRUNC, 0600);
  set_fs(old_fs);
  assert(sf->fd >= 0);
  
  // now open up the bitmap we are going to use
  sprintf(fileName, "%s%s", baseName, SPECBIT_FILE_EXT);
  sf->bitmap = util_new_bitmap(fileName);
  
  util_free_small(fileName);

  return sf;
}

void sf_delete(sparse_file_t sf, struct file *diskFile) {
  if(diskFile != NULL) {
    sf_flush_all_blocks(sf, diskFile);
  } else {
    assert(sf->numBlocks == 0);
  }

  sys_close(sf->fd);
  util_delete_bitmap(sf->bitmap);
  printk("sparse file had to flush %llu blocks\n",sf->numBlockFlush);
  util_free_large(sf);
}

int sf_get_fd(sparse_file_t sf) {
  return sf->fd;
}

int sf_is_block_present(sparse_file_t sf, loff_t offset) {
  return util_is_bit_set(sf->bitmap, COW_OFFSET_TO_IDX(offset));
}

static void sf_set_block_present(sparse_file_t sf, loff_t offset) {
  util_set_bit(sf->bitmap, COW_OFFSET_TO_IDX(offset));
}
static void sf_set_block_not_present(sparse_file_t sf, loff_t offset) {
  util_clear_bit(sf->bitmap, COW_OFFSET_TO_IDX(offset));
}

void sf_add_blocks(sparse_file_t sf, loff_t offset, unsigned long numBytes,
		   struct file *diskFile) {
  unsigned long numNewBlocks;
  unsigned long bytes;

  numNewBlocks = 0;
  for(bytes = 0; bytes < numBytes; bytes += COW_BLOCK_SIZE) {
    if(!sf_is_block_present(sf, offset + bytes)) {
      numNewBlocks++;
    }
  }
  
  if(numNewBlocks == 0) {
    // no new blocks to add
    return;
  }

  // check if our buffer is full
  if((sf->numBlocks + numNewBlocks) > NUM_SPEC_BLOCKS) {
    sf_flush_all_blocks(sf, diskFile);
    assert(sf->numBlocks == 0);
  }

  // we dont have to do any actual work, the guest is about to
  // write to our file
  for(bytes = 0; bytes < numBytes; bytes += COW_BLOCK_SIZE) {
    if(!sf_is_block_present(sf, offset + bytes)) {
      sf_set_block_present(sf, offset + bytes);
      sf->offsetList[sf->numBlocks].key = offset + bytes;
      sf->offsetList[sf->numBlocks].idx = 0xdeadbeef;
      sf->numBlocks++;
    }
  }
}

static void sf_copy_one_block(sparse_file_t sf, loff_t offset, 
			      struct file *diskFile) {
  int ret;
  void *data;
  struct file *sparseFile;

  assert(sf_is_block_present(sf, offset));

  sparseFile = fget(sf->fd); assert(sparseFile != NULL);
  data = util_alloc_small(COW_BLOCK_SIZE);

  ret = util_kernel_read(sparseFile, data, COW_BLOCK_SIZE, offset);
  assert(ret == COW_BLOCK_SIZE);

  writeDiskFile(diskFile, offset, data);

  fput(sparseFile);
  util_free_small(data);     
}

// this will simply flush the block to the backing disk before it is read
// the rotation time on the disk will make the subsequent read slow
// so this should not be the common case.  however, we will not 
// remove this block from the sparse file  
void sf_flush_block(sparse_file_t sf, loff_t offset, int diskFd) {
  struct file *diskFile;

  assert(sf_is_block_present(sf, offset));
  sf->numBlockFlush++;

  diskFile = fget(diskFd); assert(diskFile != NULL);

  sf_copy_one_block(sf, offset, diskFile);	

  fput(diskFile);
}
void sf_flush_all_blocks(sparse_file_t sf, struct file *diskFile) {
  unsigned long idx;
  struct file *sparseFile;

  util_heap_sort(sf->offsetList, (int) sf->numBlocks);

  for(idx = 0; idx < sf->numBlocks; idx++) {
    sf_copy_one_block(sf, sf->offsetList[idx].key, diskFile);
    sf_set_block_not_present(sf, sf->offsetList[idx].key);
  }
  
  sparseFile = fget(sf->fd); assert(sparseFile != NULL);
  do_truncate(sparseFile->f_dentry, 0);
  fput(sparseFile);

  sf->numBlocks = 0;
}
void sf_set_file_pos(sparse_file_t sf, loff_t offset) {
  struct file *sparseFile;

  sparseFile = fget(sf->fd); assert(sparseFile != NULL);
  kernel_llseek(sparseFile, offset, UTIL_SEEK_SET);
  fput(sparseFile);
}

struct data_file_struct {
  unsigned long recordSize;
  fileUtil_t dataFile;
  bitmap_t bitmap;
  fileUtil_t keyHash;
  fileUtil_t listFile;
  char *dataFileName;
  char *bitmapFileName;
  char *keyHashFileName;
  char *listFileName;
  unsigned long objectRefCount;
  unsigned long unlink;
  unsigned long numRec;
};

#define DF_DATA_EXT ".dfd"
#define DF_BIT_EXT  ".dfb"
#define DF_HASH_EXT ".dfh"
#define DF_LIST_EXT ".dfl"
#define DF_MAX_EXT_LEN   4

#define DF_REC_META_SIZE (sizeof(loff_t) + sizeof(unsigned long))

static unsigned long df_calc_num_rec(data_file_t df) {
  loff_t size;
  loff_t bytes;
  unsigned long idx;

  size = util_seek(df->dataFile, 0, UTIL_SEEK_END);
  assert(size >= 0);
  
  // sour long long / and % ...
  bytes = 0;
  idx = 0;
  while(size > bytes) {
    bytes += (df->recordSize + DF_REC_META_SIZE);
    assert(idx < (idx+1));
    idx++;
  }

  assert(bytes == size);

  return idx;
}

static data_file_t df_new_open(char *fileName, unsigned long recordSize, int new) {
  unsigned long len;
  data_file_t df;

  df = util_alloc_small(sizeof(*df));
  memset(df, 0, sizeof(*df));
  df->objectRefCount = 1;
  df->recordSize = recordSize;
  df->unlink = 0;
  df->numRec = 0;

  len = strlen(fileName);
  len += DF_MAX_EXT_LEN + 1;
  df->dataFileName = util_alloc_small(len);
  sprintf(df->dataFileName,"%s%s",fileName, DF_DATA_EXT);
  df->bitmapFileName = util_alloc_small(len);
  sprintf(df->bitmapFileName,"%s%s", fileName, DF_BIT_EXT);
  df->keyHashFileName = util_alloc_small(len);
  sprintf(df->keyHashFileName,"%s%s", fileName, DF_HASH_EXT);
  df->listFileName = util_alloc_small(len);
  sprintf(df->listFileName,"%s%s", fileName, DF_LIST_EXT);

  if(new) {
    df->dataFile = util_open(df->dataFileName, O_LARGEFILE | O_RDWR | 
			     O_CREAT | O_TRUNC, 0600);
    assert(df->dataFile != NULL);
    df->keyHash = util_open(df->keyHashFileName, O_LARGEFILE | O_RDWR | 
			    O_CREAT | O_TRUNC, 0600);
    assert(df->keyHash != NULL);
    df->bitmap = util_new_bitmap(df->bitmapFileName);
    assert(df->bitmap != NULL);

    df->listFile = util_open(df->listFileName,  O_LARGEFILE | O_RDWR | 
			     O_CREAT | O_TRUNC, 0600);
    assert(df->listFile != NULL);
  } else {
    df->dataFile = util_open(df->dataFileName, O_LARGEFILE | O_RDWR, 0600);
    if(df->dataFile == NULL) {
      df_close_unlink(df);
      df = NULL;
    } else {
      // if the data file is there, the rest should be also
      df->keyHash = util_open(df->keyHashFileName, O_LARGEFILE | O_RDWR, 0600);
      assert(df->keyHash != NULL);
      df->bitmap = util_open_bitmap(df->bitmapFileName);
      assert(df->bitmap != NULL);
      df->listFile = util_open(df->listFileName, O_LARGEFILE | O_RDWR, 0600);
      assert(df->listFile != NULL);
    }
    df->numRec = df_calc_num_rec(df);
  }
  
  return df;
}
data_file_t df_new(char *fileName, unsigned long recordSize) {
  return df_new_open(fileName, recordSize, 1);
}


data_file_t df_open(char *fileName, unsigned long recordSize) {
  data_file_t df = df_new_open(fileName, recordSize, 0);
  return df;
}
void df_inc_object_ref_count(data_file_t df) {
  df->objectRefCount++;
}

void df_close(data_file_t df) {
  assert(df->objectRefCount > 0);

  df->objectRefCount--;
  if(df->objectRefCount == 0) {
    // free up resources;
    util_delete_bitmap(df->bitmap);
    util_close(df->dataFile);
    util_close(df->keyHash);
    util_close(df->listFile);

    if(df->unlink) {
      util_unlink(df->dataFileName);
      util_unlink(df->bitmapFileName);
      util_unlink(df->keyHashFileName);
      util_unlink(df->listFileName);
    }

    util_free_small(df->dataFileName);
    util_free_small(df->bitmapFileName);
    util_free_small(df->keyHashFileName);
    util_free_small(df->listFileName);

    util_free_small(df);
  }
}

void df_close_unlink(data_file_t df) {
  df->unlink = 1;
  df_close(df);
}

int df_is_rec_present(data_file_t df, loff_t key) {
  return util_is_bit_set(df->bitmap, key);
}

static loff_t df_lookup_pos(data_file_t df, loff_t key) {
  int ret;
  loff_t pos;
  
  assert(util_is_bit_set(df->bitmap, key));
  // lookup the position of the data block
  ret = util_read(df->keyHash, &pos, sizeof(pos), key*sizeof(pos));
  assert(ret == sizeof(pos));
  return pos;
}

static void df_set_pos(data_file_t df, loff_t key, loff_t pos) {
  int ret;
  
  ret = util_write(df->keyHash, &pos, sizeof(pos), key*sizeof(pos));
  assert(ret == sizeof(pos));
}

void df_add_rec(data_file_t df, loff_t key, void *data) {
  int ret;
  unsigned long refCount = 1;
  loff_t pos;
  
  // update the bitmap
  assert(!util_is_bit_set(df->bitmap, key));
  util_set_bit(df->bitmap, key);
  
  // add the data
  pos = util_seek(df->dataFile, 0, UTIL_SEEK_END);
  ret = util_append(df->dataFile, &refCount, sizeof(refCount));
  assert(ret == sizeof(refCount));
  ret = util_append(df->dataFile, &key, sizeof(key));
  assert(ret == sizeof(key));
  ret = util_append(df->dataFile, data, df->recordSize);
  assert(ret == df->recordSize);

  // add the key to the list
  ret = util_append(df->listFile, &key, sizeof(key));
  assert(ret == sizeof(key));

  // now the key hash
  df_set_pos(df, key, pos);

  df->numRec++;
}

unsigned long df_get_num_rec(data_file_t df) {
  return df->numRec;
}

void df_get_rec(data_file_t df, loff_t key, void *data) {
  int ret;
  loff_t pos;

  pos = df_lookup_pos(df, key);

  // read the data
  pos += DF_REC_META_SIZE;
  ret = util_read(df->dataFile, data, df->recordSize, pos);
  assert(ret == df->recordSize);
}

static unsigned long df_get_ref_count(data_file_t df, loff_t key) {
  int ret;
  loff_t pos;
  unsigned long refCount;

  pos = df_lookup_pos(df, key);
  ret = util_read(df->dataFile, &refCount, sizeof(refCount), pos);
  assert(ret == sizeof(refCount));

  return refCount;
}

static void df_set_ref_count(data_file_t df, loff_t key, unsigned long refCount) {
  int ret;
  loff_t pos;

  pos = df_lookup_pos(df, key);
  ret = util_write(df->dataFile, &refCount, sizeof(refCount), pos);
  assert(ret == sizeof(refCount));
}

void df_invalidate(data_file_t df, loff_t key) {
  unsigned long refCount;

  refCount = df_get_ref_count(df, key);
  assert(refCount > 0);
  df_set_ref_count(df, key, refCount-1);
}

void df_inc_rec_ref_count(data_file_t df, loff_t key) {
  unsigned long refCount;
  
  refCount = df_get_ref_count(df, key);
  assert(refCount > 0);
  df_set_ref_count(df, key, refCount+1);
}

void df_remove_invalid(data_file_t df) {
  loff_t pos;
  loff_t fileSize;
  loff_t writePos;
  void *data;
  loff_t key;
  int ret;
  unsigned long recLen;
  unsigned long refCount;

  recLen = df->recordSize + DF_REC_META_SIZE;
  data = util_alloc_small(recLen);

  fileSize = util_seek(df->dataFile, 0, UTIL_SEEK_END);
  writePos = 0;

  // clear the key hash, we will build this back up
  util_truncate(df->keyHash, 0);
  util_truncate(df->listFile, 0);

  df->numRec = 0;

  for(pos = 0; pos < fileSize; pos += recLen) {
    ret = util_read(df->dataFile, &refCount, sizeof(refCount), pos);
    assert(ret == sizeof(refCount));
    ret = util_read(df->dataFile, &key, sizeof(key), pos+sizeof(unsigned long));
    assert(ret == sizeof(key));
    assert(util_is_bit_set(df->bitmap, key));
    
    if(refCount) {
      if(writePos != pos) {
	ret = util_read(df->dataFile, data, recLen, pos);
	assert(ret == recLen);
	ret = util_write(df->dataFile, data, recLen, writePos);
	assert(ret == recLen);
      }
      
      // setup key hash
      df_set_pos(df, key, writePos);
      writePos += recLen;      
      
      // setup key list
      ret = util_append(df->listFile, &key, sizeof(key));
      assert(ret == sizeof(key));

      df->numRec++;
    } else {
      util_clear_bit(df->bitmap,key);
    }
  }
  
  assert(writePos <= fileSize);
  assert(pos == fileSize);
  if(writePos < fileSize) {
    util_truncate(df->dataFile, writePos);
  }
  
  util_free_small(data);
}

void df_get_key_by_idx(data_file_t df, unsigned long idx, loff_t *keyOut) {
  loff_t pos = idx;
  int ret;
  loff_t key;

  pos *= sizeof(key);
  
  ret = util_read(df->listFile, &key, sizeof(key), pos);
  assert(ret == sizeof(key));

  *keyOut = key;
}

void df_get_rec_by_idx(data_file_t df, unsigned long idx, void *data, loff_t *keyOut){
  loff_t pos = idx;
  int ret;
  loff_t key;
  loff_t tmpKey;
  unsigned long refCount;

  pos *= (df->recordSize + DF_REC_META_SIZE);

  // get some debugging info
  ret = util_read(df->dataFile, &refCount, sizeof(refCount),pos);
  assert(ret == sizeof(refCount));
  ret = util_read(df->dataFile, &key, sizeof(key), pos + sizeof(unsigned long));
  assert(ret == sizeof(key));

  assert(pos == df_lookup_pos(df, key));
  assert(refCount > 0);
  df_get_key_by_idx(df, idx, &tmpKey);
  assert(tmpKey == key);

  if(data != NULL) {
    pos += DF_REC_META_SIZE;
    ret = util_read(df->dataFile, data, df->recordSize, pos);
    assert(ret == df->recordSize);
  }

  if(keyOut != NULL) {
    *keyOut = key;
  }
}
unsigned long df_find_rec(data_file_t df, loff_t key) {
  loff_t pos;
  unsigned long idx;

  pos = df_lookup_pos(df, key);
  assert(pos < (loff_t) 0x7fffffff);
  assert(pos >= 0);
  idx = (unsigned long) pos;
  return idx / (df->recordSize + DF_REC_META_SIZE);
}


struct list_file_struct {
  unsigned long recordSize;
  fileUtil_t dataFile;
  fileUtil_t listFile;
  bitmap_t bitmap;
  char *dataFileName;
  char *bitmapFileName;
  char *listFileName;
  unsigned long unlink;
  unsigned long numRec;
};

#define LST_DATA_EXT ".lstd"
#define LST_BIT_EXT  ".lstb"
#define LST_LIST_EXT  ".lstl"
#define LST_MAX_EXT_LEN   5

list_file_t lst_new(char *fileName, unsigned long recordSize, int useMemBitmap) {
  unsigned long len;
  list_file_t lst;

  lst = util_alloc_small(sizeof(*lst));
  memset(lst, 0, sizeof(*lst));

  lst->recordSize = recordSize;

  len = strlen(fileName) + LST_MAX_EXT_LEN + 1;
  lst->dataFileName = util_alloc_small(len);
  sprintf(lst->dataFileName, "%s%s", fileName, LST_DATA_EXT);
  lst->bitmapFileName = util_alloc_small(len);
  sprintf(lst->bitmapFileName, "%s%s", fileName, LST_BIT_EXT);
  lst->listFileName = util_alloc_small(len);
  sprintf(lst->listFileName, "%s%s", fileName, LST_LIST_EXT);

  lst->dataFile = util_open(lst->dataFileName, O_LARGEFILE | O_RDWR | 
			    O_CREAT | O_TRUNC, 0600);
  
  lst->listFile = util_open(lst->listFileName, O_LARGEFILE | O_RDWR | 
			    O_CREAT | O_TRUNC, 0600);
  
  assert((lst->dataFile != NULL) && (lst->listFile != NULL));
  
  if(useMemBitmap) {
    lst->bitmap = util_new_mem_bitmap();
  } else {
    lst->bitmap = util_new_bitmap(lst->bitmapFileName);
  }
  assert(lst->bitmap != NULL);

  return lst;
}
void lst_close(list_file_t lst) {
  util_delete_bitmap(lst->bitmap);
  util_close(lst->dataFile);
  util_close(lst->listFile);

  if(lst->unlink) {
    util_unlink(lst->dataFileName);
    util_unlink(lst->bitmapFileName);
    util_unlink(lst->listFileName);
  }

  util_free_small(lst->listFileName);
  util_free_small(lst->dataFileName);
  util_free_small(lst->bitmapFileName);
  
  util_free_small(lst);
}
void lst_close_unlink(list_file_t lst) {
  lst->unlink = 1;
  lst_close(lst);
}

int lst_is_rec_present(list_file_t lst, loff_t key) {
  assert(key >= 0);

  return util_is_bit_set(lst->bitmap, key);
}

void lst_add_rec(list_file_t lst, loff_t key, void *data) {
  int ret;

  assert(key >= 0);

  assert(!util_is_bit_set(lst->bitmap, key));
  util_set_bit(lst->bitmap, key);

  ret = util_append(lst->listFile, &key, sizeof(key));
  assert(ret == sizeof(key));

  ret = util_append(lst->dataFile, data, lst->recordSize);
  assert(ret == lst->recordSize);

  lst->numRec++;
}

void lst_get_rec_by_idx(list_file_t lst, unsigned long idx, 
			void *data, loff_t *keyOut) {
  loff_t pos;
  int ret;
 
  if(keyOut != NULL) {
    pos = idx * sizeof(*keyOut);
    ret = util_read(lst->listFile, keyOut, sizeof(*keyOut), pos);
    assert(ret == sizeof(*keyOut));
  }

  if(data != NULL) {
    pos = idx * lst->recordSize;
    ret = util_read(lst->dataFile, data, lst->recordSize, pos);
    assert(ret == lst->recordSize);
  }
}

unsigned long lst_get_num_rec(list_file_t lst) {
  return lst->numRec;
}
