#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/unistd.h>
#include <linux/string.h>
#include <linux/types.h>
#include <linux/param.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/file.h>

#include <asm/uaccess.h>

#include <asm/signal.h>
#include <linux/mman.h>

#include <asm/system.h>
#include <asm/segment.h>

#include <asm/semaphore.h>

#include "cow_disk.h"
#include "myutil.h"
#include "checkpoint.h"

#include <revirt/rlog.h>

#define CURR_FILE_EXT ".cur"

extern int util_kernel_read(struct file *file, void *buf, 
			    unsigned long numBytes, loff_t offset);
extern int util_kernel_write(struct file *file, void *buf, 
			     unsigned long numBytes, loff_t offset);

extern asmlinkage long sys_mmap2(unsigned long addr, unsigned long len,
                                 unsigned long prot, unsigned long flags,
                                 unsigned long fd, unsigned long pgoff);
extern asmlinkage long sys_munmap(unsigned long addr, size_t len);


//static void flushSpeculativeData(cow_disk_t cow, struct file *diskFile);

typedef enum {
  COW_MODE_LOG,
  COW_MODE_REPLAY
} cow_mode_t;

#define UML_COW_NEGATIVE_INFINITY ((unsigned long long)(-1LL))
#define UML_COW_POSITIVE_INFINITY ((unsigned long long)(-2LL))

// before the first write on a new cow disk, head, tail , curr, and currUndo
// should all be null.  After the first write or the first checkpont has been
// added, none of them should be null
struct cow_disk_struct {
  struct checkpoint_struct *head;
  struct checkpoint_struct *tail;
  struct checkpoint_struct *curr;
  unsigned long long currId;
  char *baseName;
  undo_redo_t currUndo;
  char *currUndoFileName;
  char *currUndoDataName;
  cow_mode_t mode;
  sparse_file_t sf;
};

cow_disk_t newCowDisk(char *baseName) {
  unsigned long len;

  cow_disk_t cow = (cow_disk_t) util_alloc_small(sizeof(*cow));
  memset(cow, 0, sizeof(*cow));
  
  cow->mode = COW_MODE_LOG;

  // now set up the filename data
  len = strnlen(baseName, UTIL_MAX_PATH);
  assert(len < UTIL_MAX_PATH);
  len += 1;
  cow->baseName = (char *) util_alloc_small(len);
  sprintf(cow->baseName, "%s", baseName);

  len = strnlen(baseName, UTIL_MAX_PATH);
  assert(len < UTIL_MAX_PATH);
  len += strnlen(CURR_FILE_EXT, UTIL_MAX_PATH);
  len += 1;
  cow->currUndoFileName = (char *) util_alloc_small(len);
  sprintf(cow->currUndoFileName, "%s%s", baseName, CURR_FILE_EXT);

  cow->currUndoDataName = (char *) util_alloc_small(len);
  sprintf(cow->currUndoDataName, "%s.dat", baseName);

  cow->sf = sf_new(baseName);

  return cow;
}

void closeCowDisk(cow_disk_t cow, struct file *file) {
  checkpoint_t cp, tmp;

  if(cow->tail != NULL) {
    assert(cow->head != NULL);
    assert(cow->curr != NULL);
    // this means that there is live data after the last checkpoint, we should
    // save it
    if(cow->currId > cp_get_id(cow->tail)) {
      setNewCheckpoint(cow, file);
    }

    // now close up everything
    cp = cow->head;
    while(cp != NULL) {
      tmp = cp;
      cp = cp_get_next(cp);

      while(cp_delete(tmp,0) > 0)
	;
    }
    
    if(cow->currUndo != NULL) ur_close(cow->currUndo);
  }

  util_free_small(cow->baseName);
  util_free_small(cow->currUndoFileName);
  util_free_small(cow->currUndoDataName);
  sf_delete(cow->sf, file);
  util_free_small(cow);
}

sparse_file_t getCowSf(cow_disk_t cow) {
  return cow->sf;
}


static void readDiskFile(struct file *file, loff_t offset, void *data) {
  unsigned char *userMem;
  int ret;

  // to deal with raw disks
  userMem = (unsigned char *) sys_mmap2(0, PAGE_SIZE, PROT_READ | PROT_WRITE,
					MAP_PRIVATE | MAP_ANONYMOUS, 0, 0);
  assert(userMem != NULL);
  assert(PAGE_SIZE >= COW_BLOCK_SIZE);
  assert(((unsigned long) userMem & (COW_BLOCK_SIZE-1)) == 0);
  
  ret = util_kernel_read(file, userMem, COW_BLOCK_SIZE, offset);
  assert(ret == COW_BLOCK_SIZE);
  copy_from_user(data, userMem, COW_BLOCK_SIZE);
  
  sys_munmap((unsigned long) userMem, PAGE_SIZE);
}
static void readDiskFd(int diskFd, loff_t offset, void *data) {
  struct file *file = fget(diskFd);
  assert(file != NULL);
  readDiskFile(file, offset, data);
  fput(file);
}

void writeDiskFile(struct file *file, loff_t offset, void *data) {
  unsigned char *userMem;
  int ret;
  
  // to deal with raw disks
  userMem = (unsigned char *) sys_mmap2(0, PAGE_SIZE, PROT_READ | PROT_WRITE,
					MAP_PRIVATE | MAP_ANONYMOUS, 0, 0);
  assert(userMem != NULL);
  assert(PAGE_SIZE >= COW_BLOCK_SIZE);
  assert(((unsigned long) userMem & (COW_BLOCK_SIZE-1)) == 0);
  
  copy_to_user(userMem, data, COW_BLOCK_SIZE);
  ret = util_kernel_write(file, userMem, COW_BLOCK_SIZE, offset);
  assert(ret == COW_BLOCK_SIZE);
  
  sys_munmap((unsigned long) userMem, PAGE_SIZE);
}

// called only once we are sure that we are not going to open existing
// checkpoints
static void initNewCowDisk(cow_disk_t cow) {
  // we need to initialize the cow data
  assert(cow->head == NULL); assert(cow->tail == NULL); 
  assert(cow->curr == NULL); assert(cow->currUndo == NULL);
  cow->curr = cp_new(cow->baseName, 0);
  cow->head = cow->curr;
  cow->tail = cow->curr;
  
  // open up a new undo for the current data
  ur_new(cow->currUndoFileName, NULL, cow->currUndoDataName, &cow->currUndo, NULL);
}

void writeCowDisk(cow_disk_t cow, int diskFd, loff_t offset) {
  void *tmpData;

  assert(offset >= 0);
  assert((((unsigned long) offset) % COW_BLOCK_SIZE) == 0);

  if(revirt_options.replayflags & REPLAY_OPT_STK_NOWRITE) {
    // this should never get called if the nowrite flags is set
    assert(0);
  }

  if(cow->curr == NULL) {
    initNewCowDisk(cow);
  } else {
    // see if we are logging or replaying.  If we are replaying
    // we need to reset the current undo every time we pass a checkpoint
    if(cow->mode == COW_MODE_REPLAY) {
      assert(cow->curr != cow->tail);
      assert(cp_get_next(cow->curr) != NULL);
      assert(cp_get_id(cp_get_next(cow->curr)) > cow->currId);
    } else if(cow->mode == COW_MODE_LOG) {
      assert(cow->currId >= cp_get_id(cow->tail));
    } else {
      assert(0);
    }
  }

  // now process the new data
  if(!ur_is_block_present(cow->currUndo, offset)) {
    // XXX FIXME we do not need this read when we are replaying
    // optimize this later....
    tmpData = util_alloc_small(COW_BLOCK_SIZE);

    readDiskFd(diskFd, offset, tmpData);
    ur_add_block(cow->currUndo, offset, cow->currId, tmpData);

    util_free_small(tmpData);
  }

  ur_update_time(cow->currUndo, offset, cow->currId);

  cow->currId++;

  if(cow->mode == COW_MODE_REPLAY) {
    // the head and tail do not change when replaying and there
    // is always a next checkpoint when we get a new write
    if(cp_get_id(cp_get_next(cow->curr)) == cow->currId) {
      ur_reset(&cow->currUndo, cow->currUndoDataName);
      cow->curr = cp_get_next(cow->curr);
      if(cow->curr == cow->tail) {
	printk("WARNING: you have gone beyond your disk cow log while\n"
	       "in replay mode.  Make sure you know what you are doing\n"
	       "You probably do not want to use these logs any more!!!!!!!!\n"
	       "cow disk mode now set to LOG\n");
	cow->mode = COW_MODE_LOG;
      }
    }    
  }
}

static void copy_ur(undo_redo_t to, undo_redo_t blocklist_ur,
		    struct file *file, undo_redo_t timelist_ur) {
  unsigned long idx, numRec;
  loff_t offset;
  unsigned long long id;
  void *data;
  
  data = util_alloc_small(COW_BLOCK_SIZE);

  numRec = ur_get_num_rec(blocklist_ur);
  for(idx = 0; idx < numRec; idx++) {
    // get the offset from the blocklist.  If there is a file, get the
    // data from there.  otherwise just use the data from the blocklist
    if(file == NULL) {
      ur_get_rec(blocklist_ur, idx, &offset, &id, data);
    } else {
      ur_get_rec_meta(blocklist_ur, idx, &offset, &id);
      readDiskFile(file, offset, data);
    }

    // get the time from the timelist if there is one, otherwise just 
    // use the id from the record
    if(timelist_ur != NULL) {
      id = ur_get_time(timelist_ur, offset);
    }
    // insert the new block
    ur_add_block(to, offset, id, data);
  }
  util_free_small(data);
}

static unsigned long long setLogCheckpoint(cow_disk_t cow, struct file *file) {
  checkpoint_t prevCp;

  assert(cow->curr != NULL);
  assert(cow->curr == cow->tail);
  assert(cow->currId == cp_get_id(cow->curr));
  assert(cp_get_next(cow->curr) == NULL);
  assert(cp_get_prev(cow->curr) != NULL);

  prevCp = cp_get_prev(cow->curr);

  // copy the current undo information to the previous checkpoint undo log
  assert(ur_get_num_rec(cow->currUndo) > 0);
  assert(ur_get_num_rec(cp_get_undo(prevCp)) == 0);
  copy_ur(cp_get_undo(prevCp), cow->currUndo, NULL, NULL);

  // set up the redo log for the current checkpoint.  It is based on
  // the current state of the file, the block list from the prev check
  // undo list, and the times from the current undo.  Also, try first 
  // to get the data from the sparse file, if it is not there then 
  // use the disk file
  copy_ur(cp_get_redo(cow->curr), cp_get_undo(prevCp), file, cow->currUndo);

  assert(cp_get_id(cow->curr) == cow->currId);
  return cow->currId;
}

// this function is specifically for adding a new checkpoint when
// replaying.  This is how the undo(N-1) gets updated to undo(N-1)'
static void remove_undo_blocks(undo_redo_t undo, undo_redo_t blocklist_ur) {
  unsigned long idx, numRec;
  loff_t offset;
  unsigned long long id;
  void *data;

  data = util_alloc_small(COW_BLOCK_SIZE);
  
  numRec = ur_get_num_rec(undo);
  for(idx = 0; idx < numRec; idx++) {
    ur_get_rec_meta(undo, idx, &offset, &id);
    // remove any blocks from the undo record that are not in our blocklist
    if(!ur_is_block_present(blocklist_ur,offset)) {
      assert(ur_is_block_present(undo, offset));
      ur_invalidate(undo, idx);
    }
  }

  ur_remove_invalid(undo);
  util_free_small(data);
}

// this function is specifically for adding a new checkpoint when
// replaying.  This is how the redo(N+1) gets updated to redo(N+1)'
static void remove_redo_blocks(undo_redo_t redo, undo_redo_t currRedo) {
  unsigned long idx, numRec, currBlockIdx;
  loff_t offset, currOffset;
  unsigned long long id, currId;
  void *data;

  data = util_alloc_small(COW_BLOCK_SIZE);
  
  numRec = ur_get_num_rec(redo);
  for(idx = 0; idx < numRec; idx++) {
    ur_get_rec_meta(redo, idx, &offset, &id);
    if(ur_is_block_present(currRedo,offset)) {
      currBlockIdx = ur_find_block(currRedo, offset);
      ur_get_rec_meta(currRedo, currBlockIdx, &currOffset, &currId);
      assert(currOffset == offset);
      // remove it if the times are equal for this block
      if(currId == id) {	
	ur_invalidate(redo, idx);
      }
    }
  }
  
  ur_free_find_data(currRedo);
  ur_remove_invalid(redo);
  util_free_small(data);
}

static unsigned long long setReplayCheckpoint(cow_disk_t cow, struct file *file) {
  checkpoint_t prevCp, nextCp;
  assert(cow->curr != NULL);
  assert(cp_get_next(cow->curr) != NULL);
  assert(cp_get_prev(cow->curr) != NULL);
  assert(cp_get_id(cow->curr) == cow->currId);

  prevCp = cp_get_prev(cow->curr);
  nextCp = cp_get_next(cow->curr);

  assert(cow->currId > cp_get_id(prevCp));
  //assert(cow->currId < cp_get_id(nextCp));
  if(cow->currId >= cp_get_id(nextCp)) {
    printk("currId = %llu, next id = %llu\n",cow->currId, cp_get_id(nextCp));
    assert(0);
  }


  // the redo for the new checkpoint and the undo for the previous
  copy_ur(cp_get_redo(cow->curr), cow->currUndo, file, cow->currUndo);
  remove_undo_blocks(cp_get_undo(prevCp), cow->currUndo);

  // now the redo for the next checkpoint
  remove_redo_blocks(cp_get_redo(nextCp), cp_get_redo(cow->curr));
  copy_ur(cp_get_undo(cow->curr), cp_get_redo(nextCp), file, NULL);

  return cow->currId;
}

unsigned long long setNewCheckpoint(cow_disk_t cow, struct file *file) {
  unsigned long long id = (unsigned long long) -1;
  checkpoint_t cp;

  static int warn=1;
  
  if(revirt_options.replayflags & REPLAY_OPT_STK_NOWRITE) {
    if(warn) {
      printk("setNewCheckpoint: STK_NOWRITE set, not writing file.\n");
      warn=0;
    }
    return 0;
  }
  warn=1;

  if(cow->curr == NULL) {
    initNewCowDisk(cow);
  }

  //printk("taking checkpoint at currId %llu\n",cow->currId);
  if(cp_get_next(cow->curr) != NULL) {
    //printk("next id %llu\n",cp_get_id(cp_get_next(cow->curr)));
  }

  sf_flush_all_blocks(cow->sf, file);

  // if we have not had any disk writes since the last checkpoint, simply
  // increment the reference count on the current checkpoint
  if(cow->currId == cp_get_id(cow->curr)) {
    cp_increment_ref_count(cow->curr);
    return cow->currId;
  }

  // we need to add a new checkpoint
  cp = cp_new(cow->baseName, cow->currId);

  // insert it into the list and make it the current checkpoint
  cp_insert(cow->curr, cp);
  if(cow->curr == cow->tail) {    
    cow->tail = cp;
  }
  cow->curr = cp;

  if(cow->mode == COW_MODE_LOG) {
    id = setLogCheckpoint(cow, file);
  } else if(cow->mode == COW_MODE_REPLAY) {
    id = setReplayCheckpoint(cow, file);
  } else {
    assert(0);
  }

  ur_reset(&cow->currUndo, cow->currUndoDataName);

  return id;
}

void appendCowCheckpoint(cow_disk_t cow, unsigned long long id) {
  checkpoint_t cp;

  // we are now replaying if this is happening
  cow->mode = COW_MODE_REPLAY;

  cp = cp_open(cow->baseName, id);

  if(cow->tail == NULL) {
    assert((cow->head == NULL) && (cow->curr == NULL));
    cow->tail = cp;
    cow->head = cp;
    cow->curr = cp;
    cow->currId = id;
  } else {
    assert((cow->head != NULL) && (cow->curr != NULL));
    assert(cp_get_id(cow->tail) < id);
    cp_insert(cow->tail, cp);
    cow->tail = cp;
  }
  
}

static void restore_ur(struct file *toFile, undo_redo_t ur, bitmap_t visitedBitmap) {
  unsigned long idx, numRec;
  void *data;
  loff_t offset;
  unsigned long long tmpId;
  unsigned long numBlocksRest = 0;

  data = util_alloc_small(COW_BLOCK_SIZE);
  assert(ur != NULL);
  assert(toFile != NULL);
  numRec = ur_get_num_rec(ur);
  printk("about to restore %lu blocks\n",numRec);
  for(idx = 0; idx < numRec; idx++) {
    // try two things here
    // 1) defer the data read
    // 2) warm up the file cache for the undo / redo record
    //    since we will be reading it out of order
    ur_get_rec_meta_sorted(ur, idx, &offset, &tmpId);
    
    if(visitedBitmap == NULL) {
      ur_get_rec_sorted(ur, idx, &offset, &tmpId, data);
      writeDiskFile(toFile, offset, data);
      numBlocksRest++;
    } else if(!util_is_bit_set(visitedBitmap,offset >> COW_BLOCK_SHIFT)) {
      util_set_bit(visitedBitmap,offset >> COW_BLOCK_SHIFT);
      ur_get_rec_sorted(ur, idx, &offset, &tmpId, data);
      writeDiskFile(toFile, offset, data);
      numBlocksRest++;
    }
  }
  printk("done restored %lu blocks\n", numBlocksRest);

  ur_free_find_data(ur);
  util_free_small(data);
}

checkpoint_t cow_find_checkpoint(cow_disk_t cow, unsigned long long id) {
  checkpoint_t p, q, ret=NULL, origCp;

  p=origCp=cow->curr;
  if(id < cp_get_id(p)) {
    /* If the target id is below us, walk backwards... */
    do {
	q=cp_get_prev(p);
	if(!q) {
	    printk("cow_find_checkpoint: started at id %lld, wanted id %lld, last id %lld, cp_get_prev() returned null!\n",
	    	cp_get_id(origCp), id, cp_get_id(p));
	    goto out;
	}
	if(id > cp_get_id(q)) {
	    printk("cow_find_checkpoint: target id %lld between checkpoints %lld and %lld!\n",
	    	id, cp_get_id(p), cp_get_id(q));
	    goto out;
	}
	p=q;
    } while(cp_get_id(p) != id);
  } else if (id > cp_get_id(p)) {
    /* Otherwise, walk forwards... */
    do {
	q=cp_get_next(p);
	if(!q) {
	    printk("cow_find_checkpoint: started at id %lld, wanted id %lld, last id %lld, cp_get_prev() returned null!\n",
	    	cp_get_id(origCp), id, cp_get_id(p));
	    goto out;
	}
	if(id < cp_get_id(q)) {
	    printk("cow_find_checkpoint: target id %lld between checkpoints %lld and %lld!\n",
	    	id, cp_get_id(p), cp_get_id(q));
	    goto out;
	}
	p=q;
    } while(cp_get_id(p) != id);
  }
  ret=p;

  out:
   return ret;
}


static int moveToCheckpoint(cow_disk_t cow, unsigned long long id, 
			     struct file *file) {
  checkpoint_t origCp, targetCp;
  checkpoint_t p;
  bitmap_t visitedBlocks;
  char *bitmapName;
  int adjustOrigCp = 1;
  int err=0;

  assert((cow->tail != NULL) && (cow->head != NULL) && (cow->curr != NULL));

  origCp = cow->curr;

  if(cp_get_id(cow->curr) == cow->currId) {
    // we are at a checkpoint already, no adjustment needed
    adjustOrigCp = 0;
  }

  /*
   * Move "cow->curr" to the target id...
   */
  if(!(p=cow_find_checkpoint(cow, id))) {
    err=-EINVAL; 
    goto out;
  }

  cow->curr = p;
  cow->currId = cp_get_id(cow->curr);
  assert(cow->currId == id);
  targetCp = cow->curr;

  /* FIXME -- encapsulate this in a more versatile function */

    bitmapName = util_alloc_small(strlen(cow->baseName)+8);
    sprintf(bitmapName,"%s.tmpbit",cow->baseName);
    visitedBlocks = util_new_bitmap(bitmapName);
    util_unlink(bitmapName);
    util_free_small(bitmapName);

    // this adjustment takes care of the restore listed above.  No more go 
    // back to the current checkpoint and go from there.
    if(adjustOrigCp) {
      assert((cp_get_next(origCp) != NULL) && (cp_get_prev(origCp) != NULL));
      origCp = (cp_get_id(origCp) < cp_get_id(targetCp)) ? cp_get_prev(origCp) : 
	cp_get_next(origCp);
    }

    while(targetCp != origCp) {
      if(cp_get_id(targetCp) < cp_get_id(origCp)) {
	restore_ur(file,cp_get_undo(targetCp),visitedBlocks);
	targetCp = cp_get_next(targetCp);
	assert(cp_get_id(targetCp) <= cp_get_id(origCp));
      } else if(cp_get_id(targetCp) > cp_get_id(origCp)) {
	restore_ur(file,cp_get_redo(targetCp),visitedBlocks);
	targetCp = cp_get_prev(targetCp);
	assert(cp_get_id(targetCp) >= cp_get_id(origCp));
      } else {
	assert(0);
      }
    }

    util_delete_bitmap(visitedBlocks);


  assert(cow->currId == id);
  return 0;

 out:
  return err;
}

int setCowCurrId(cow_disk_t cow, unsigned long long id) {
  checkpoint_t p;
  int err=0;
  if(!(p=cow_find_checkpoint(cow, id))) {
  	err=-EINVAL;
	goto out;
  }
  if(cp_get_id(p) != id) {
    err=-EINVAL;
    goto out;
  }
  cow->curr = p;
  cow->currId = id;
  out:
   return err;
}

int goToCheckpoint(cow_disk_t cow, unsigned long long id, struct file *file) {
  static int warn=1;
  int err;

  if(revirt_options.replayflags & REPLAY_OPT_STK_NOWRITE) {
    if(warn) {
    	printk("goToCheckpoint: STK_NOWRITE set, not writing file.\n");
	warn=0;
    }
    err=0;
    goto out;
  }
  warn=1;

  sf_flush_all_blocks(cow->sf, file);

  if(cow->currId == id) {
    // there is nothing to do, just return
  }

  // as soon as we go to a checkpoint, we are in replay mode
  if(cow->mode == COW_MODE_LOG) {
    // set a checkpoint at the current position, this will be the last one
    if(cp_get_id(cow->tail) < cow->currId) {
      printk("setting a checkpoint at the end\n");
      setNewCheckpoint(cow, file);
    }
  }

  cow->mode = COW_MODE_REPLAY;
  printk("moving from %llu to the checkpoint at %llu\n",cow->currId, id);
  if((err=moveToCheckpoint(cow, id, file))<0) {
    printk("goToCheckpoint: moveToCheckpoint returned %d!\n", err);
    goto out;
  }

  assert(cow->currId == id);
  if(cow->curr == cow->tail) {
    // we just switched back to log mode
    printk("switching back to log mode\n");
    cow->mode = COW_MODE_LOG;
  }
 out:
  return err;
}

// this function is used to merge two undo redo logs.  This merge is based
// soley on block number.  If there is a block in extra, that is not
// in base, then it is added to base.
static void merge_ur(undo_redo_t base, undo_redo_t extra) {
  unsigned long idx, numRec;
  void *data;
  loff_t offset;
  unsigned long long id;
  
  data = util_alloc_small(COW_BLOCK_SIZE);
  numRec = ur_get_num_rec(extra);
  for(idx = 0; idx < numRec; idx++) {
    ur_get_rec_meta(extra, idx, &offset, &id);
    if(!ur_is_block_present(base, offset)) {
      ur_get_rec(extra, idx, &offset, &id, data);
      ur_add_block(base, offset, id, data);
    }
  }

  util_free_small(data);
}

// this is used to update the currUndo when we are deleting checkpoints
static void update_currUndo(cow_disk_t cow) {
  undo_redo_t cRedo, pUndo;
  unsigned long idx, numRec;
  void *data;
  loff_t offset;
  unsigned long long id;

  cRedo = cp_get_redo(cow->curr);
  pUndo = cp_get_undo(cp_get_prev(cow->curr));

  data = util_alloc_small(COW_BLOCK_SIZE);
 
  // first store the correct times from the current redo into the
  // previous undo.  Undo logs do not use them, so it will just
  // store them for us
  numRec = ur_get_num_rec(cRedo);
  for(idx = 0; idx < numRec; idx++) {
    ur_get_rec_meta(cRedo, idx, &offset, &id);
    assert(ur_is_block_present(pUndo,offset));
    ur_update_time(pUndo, offset, id);
  }

  // now copy the data from the currUndo into pUndo
  numRec = ur_get_num_rec(cow->currUndo);
  for(idx = 0; idx < numRec; idx++) {
    ur_get_rec_meta(cow->currUndo, idx, &offset, &id);
    if(!ur_is_block_present(pUndo, offset)) {
      ur_get_rec(cow->currUndo, idx, &offset, &id, data);
      ur_add_block(pUndo, offset, id, data);
      id = ur_get_time(cow->currUndo, offset);
      ur_update_time(pUndo, offset, id);
    }
  }

  // and copy the data back from pUndo to currUndo
  ur_reset(&cow->currUndo, cow->currUndoDataName);
  numRec = ur_get_num_rec(pUndo);
  for(idx = 0; idx < numRec; idx++) {
    ur_get_rec(pUndo, idx, &offset, &id, data);
    id = ur_get_time(pUndo, offset);
    ur_add_block(cow->currUndo, offset, id, data);
    ur_update_time(cow->currUndo, offset, id);
  }


  ur_reset_time(pUndo);
  
  util_free_small(data);
}
void deleteCheckpoint(cow_disk_t cow, unsigned long long id) {
  checkpoint_t cp;
  unsigned long idx,numRec;

  static int warn=1;

  if(revirt_options.replayflags & REPLAY_OPT_STK_NOWRITE) {
    if(warn) {
      printk("deleteCheckpoint: STK_NOWRITE set, not writing file.\n");
      warn=0;
    }
    return;
  }
  warn=1;
  
  assert(cow->curr != NULL);

  // find the checkpoint in question
  cp = cow->curr;
  while(id != cp_get_id(cp)) {
    if(id > cp_get_id(cp)) {
      assert(cp_get_next(cp) != NULL);
      assert(cp_get_id(cp_get_next(cp)) > cp_get_id(cp));
      cp = cp_get_next(cp);
    } else {
      assert(cp_get_prev(cp) != NULL);
      assert(cp_get_id(cp_get_prev(cp)) < cp_get_id(cp));
      cp = cp_get_prev(cp);
    }
  }

  if(cp_get_ref_count(cp) > 1) {
    cp_delete(cp,0);
    return;
  } 

  // we know that the checkpoint will be deleted

  // we cannot delete the first checkpoint.  We will need it to get back
  // to the beginning, plus it would invalidata SEVERAL assumptions made
  // throught the cow_disk code
  if(cp == cow->head) return;

  // check if we need to update cow->currUndo
  if(cp == cow->curr) {
    update_currUndo(cow);
    cow->curr = cp_get_prev(cow->curr);
  }
  
  assert(cp_get_prev(cp) != NULL);
  if(cp_get_next(cp) != NULL) {
    merge_ur(cp_get_redo(cp_get_next(cp)), cp_get_redo(cp));
    merge_ur(cp_get_undo(cp_get_prev(cp)), cp_get_undo(cp));
  } else {
    // we are deleting the last checkpoint, make sure the tail is still up
    // to date
    assert(cp == cow->tail);
    cow->tail = cp_get_prev(cp);
    // clean up the undo record for the new tail
    numRec = ur_get_num_rec(cp_get_undo(cow->tail));
    for(idx = 0; idx < numRec; idx++) {
      ur_invalidate(cp_get_undo(cow->tail),idx);
    }
    ur_remove_invalid(cp_get_undo(cow->tail));
  }



  // now remove it from the list
  cp_remove(cp);
  cp_delete(cp,1);

  // check if we just got switched back into log mode
  if(cow->curr == cow->tail) {
    if(cow->mode == COW_MODE_REPLAY) {
      printk("switching cow disk back to log mode\n");
    }
    cow->mode = COW_MODE_LOG;
  }

  assert(cp_get_next(cow->tail) == NULL);
}

/*
static void installCurrUndo(cow_disk_t cow, struct file *file) {
  unsigned long idx, numRec;
  void *data;
  loff_t offset;
  unsigned long long id;
  char *undoName;
  int ret;
  
  assert(cow->curr->id == cow->currId);

  // set up the undo record for the current checkpoint
  numRec = ur_get_num_rec(cow->currUndo);
  assert(numRec > 0);
  data = util_alloc_small(COW_BLOCK_SIZE);

  undoName = createBasename(cow->baseName, cow->currId,NAME_UNDO);
  if(cow->curr->undo != NULL) close_undo_redo_file(cow->curr->undo);
  cow->curr->undo = new_undo_redo_file(undoName);
  util_free_small(undoName);

  for(idx = 0; idx < numRec; idx++) {
    ur_get_rec(cow->currUndo, idx, &offset, &id, data);
    ur_add_block(cow->curr->undo, offset, data, id);
  }

  // set up redo record for current checkpoint
  assert(cow->curr->prev != NULL);
  if(cow->curr->prev->redo != NULL) close_undo_redo_file(cow->curr->prev->redo);  
  undoName = createBasename(cow->baseName, cow->curr->prev->id,NAME_REDO);
  cow->curr->prev->redo = new_undo_redo_file(undoName);
  util_free_small(undoName);

  for(idx = 0; idx < numRec; idx++) {
    ur_get_rec_meta(cow->currUndo, idx, &offset, &id);
    // read the current disk
    ret = util_kernel_read(file, data, COW_BLOCK_SIZE, offset);
    assert(ret == COW_BLOCK_SIZE);
    ur_add_block(cow->curr->prev->redo, offset, data, id);
  }

  // reset the current undo log and update the curr and tail pointers
  close_undo_redo_file(cow->currUndo);
  cow->currUndo = new_undo_redo_file(cow->currUndoFileName);

  util_free_small(data);
}

static void installCheckpointAtCurrId(cow_disk_t cow) {
  struct checkpoint_struct *cp;

  // set up the undo record and the new checkpoint
  cp = newCheckpointStruct();
  cp->id = cow->currId;
  
  // insert into list
  assert(cow->curr != NULL);
  assert(cow->curr->id < cp->id);
  cp->prev = cow->curr;
  cp->next = cow->curr->next;
  if(cow->curr->next == NULL) {
    // end of the list
    cow->tail = cp;
  } else {
    assert(cow->curr->next->id > cp->id);
    cow->curr->next->prev = cp;
  }

  cow->curr->next = cp;
  cow->curr = cp;
}

static void addNewCheckpointToEnd(cow_disk_t cow, struct file *file) {
  installCheckpointAtCurrId(cow);

  // install the actuall data
  installCurrUndo(cow, file);
}

static int areBlocksEqual(unsigned long *block0, unsigned long *block1) {
  unsigned long idx;
  for(idx = 0; idx < (COW_BLOCK_SIZE / sizeof(unsigned long)); idx++) {
    if(block0[idx] != block1[idx]) return 0;
  }
  
  return 1;
}

static undo_redo_t calcInsertRedo(cow_disk_t cow, struct file *file) {
  unsigned long idx, numRec, foundIdx;
  void *diskData;
  void *redoData;
  loff_t offset;
  unsigned long long id;
  char *redoName;
  int ret;
  undo_redo_t redo;

  // this is based on the data of the current hard disk and the 
  // data in the current redo log for the previous checkpoint
  assert(cow->curr->prev != NULL); 
  assert(cow->currId == cow->curr->id);
  assert(cow->curr->prev->redo != NULL);
  redoName = createBasename(cow->baseName, cow->currId,NAME_REDO);
  redo = new_undo_redo_file(redoName);
  util_free_small(redoName);

  diskData = util_alloc_small(COW_BLOCK_SIZE);
  redoData = util_alloc_small(COW_BLOCK_SIZE);

  numRec = ur_get_num_rec(cow->curr->prev->redo);
  for(idx = 0; idx < numRec; idx++) {
    ur_get_rec(cow->curr->prev->redo, idx, &offset,&id,redoData);
    if(!ur_is_block_present(cow->currUndo, offset)) {
      // the block is not present, add it in
      ur_add_block(redo, offset, redoData, id);
    } else {
      // the block is in the current undo log, check the data 
      // to see if they are different
      foundIdx = ur_find_block(cow->currUndo, offset);
      // get the redo value from the current disk
      ret = util_kernel_read(file, diskData, COW_BLOCK_SIZE, offset);
      assert(ret == COW_BLOCK_SIZE);
      if(!areBlocksEqual(diskData, redoData)) {
	ur_add_block(redo, offset, redoData, id);
      }
    }
  }

  util_free_small(diskData);
  util_free_small(redoData);

  return redo;
}

static undo_redo_t calcInsertUndo(cow_disk_t cow, struct file *file) {
  unsigned long idx, numRec;
  void *data;
  loff_t offset;
  unsigned long long id;
  char *undoName;
  int ret;
  undo_redo_t undo;

  assert(cow->curr->next != NULL);
  assert(cow->curr->next->id > cow->curr->id);
  assert(cow->curr->redo != NULL);

  data = util_alloc_small(COW_BLOCK_SIZE);

  undoName = createBasename(cow->baseName, cow->curr->next->id,NAME_UNDO);
  undo = new_undo_redo_file(undoName);
  util_free_small(undoName);

  // for each of the blocks that were written to , as denoted
  // by curr->redo, copy the current state of the hard disk and
  // store it in the undo log of the next checkpoint
  numRec = ur_get_num_rec(cow->curr->redo);
  for(idx = 0; idx < numRec; idx++) {
    ur_get_rec_meta(cow->curr->redo, idx, &offset, &id);
    ret = util_kernel_read(file, data, COW_BLOCK_SIZE, offset);
    ur_add_block(undo, offset, data, id);
  }

  util_free_small(data);

  return undo;
}

static void insertNewCheckpoint(cow_disk_t cow, struct file *file) {
  assert((cow->curr != NULL) && (cow->curr->next != NULL));
  installCheckpointAtCurrId(cow);

  // first, we need to figure out what the redo log is going to 
  // be for the current checkpoint since we have not actually executed
  // it yet
  cow->curr->redo = calcInsertRedo(cow, file);

  // now that we have the redo for the current checkpoint, we can
  // figure out the new undo for the next checkpoint;
  if(cow->curr->next->undo != NULL) close_undo_redo_file(cow->curr->next->undo);
  cow->curr->next->undo = calcInsertUndo(cow, file);


  installCurrUndo(cow, file);
}

unsigned long long setNewCheckpoint(cow_disk_t cow, struct file *file) {
  assert(file != NULL);
  // check if we are setting a new checkpoint beyond the end of the curr
  // checkpoint, or if we are setting a new one between two existing
  // checkpoints.  Also, if there were not any writes during the current
  // checkpoint interval, we simply increment the reference count for
  // the checkpoint
  if(cow->currId == cow->curr->id) {
    assert(ur_get_num_rec(cow->currUndo) == 0);
    incrementRefCount(cow->curr);
  } else if(cow->curr->next == NULL) {
    // new checkpoint at end
    assert(cow->curr->id < cow->currId);
    addNewCheckpointToEnd(cow, file);
  } else {
    // new checkpoint between two existing checkpoints
    assert((cow->currId > cow->curr->id) && (cow->currId <= cow->curr->next->id));
    if(cow->currId == cow->curr->next->id) {
      cow->curr = cow->curr->next;
      incrementRefCount(cow->curr);
    } else {
      insertNewCheckpoint(cow, file);
    }
  }

  assert(cow->curr->id == cow->currId);

  return cow->curr->id;
}

static void restoreURData(undo_redo_t ur, struct file *file) {
  unsigned long idx, numRec;
  void *data;
  loff_t offset;
  unsigned long long tmpId;
  
  data = util_alloc_small(COW_BLOCK_SIZE);
  assert(ur != NULL);
  numRec = ur_get_num_rec(ur);
  for(idx = 0; idx < numRec; idx++) {
    ur_get_rec(ur, idx, &offset, &tmpId, data);
    util_kernel_write(file, data, COW_BLOCK_SIZE, offset);
  }

  util_free_small(data);
}


void mergeURRec(undo_redo_t base, undo_redo_t extra) {
  // any blocks that are in extra, but not in base, will be added to base
  unsigned long idx, numRec;
  loff_t offset;
  unsigned long long id;
  void *data;

  data = util_alloc_small(COW_BLOCK_SIZE);

  numRec = ur_get_num_rec(extra);
  for(idx = 0; idx < numRec; idx++) {
    ur_get_rec_meta(extra, idx, &offset, &id);
    if(!ur_is_block_present(base, offset)) {
      ur_get_rec(extra, idx, &offset, &id, data);
      ur_add_block(base, offset, data, id);
    }
  }

  util_free_small(data);
}

void copyURRec(undo_redo_t to, undo_redo_t from) {
  unsigned long idx, numRec;
  loff_t offset;
  unsigned long long id;
  void *data;
  
  data = util_alloc_small(COW_BLOCK_SIZE);

  numRec = ur_get_num_rec(from);
  for(idx = 0; idx < numRec; idx++) {
    ur_get_rec(from, idx, &offset, &id, data);
    ur_add_block(to, offset, data, id);
  }

  util_free_small(data);
}


void resetAndCopyURRec(cow_disk_t cow, undo_redo_t *to, undo_redo_t from, 
		       unsigned long long toId, undo_redo_name_t urn) {
  
  char *tmpName = createBasename(cow->baseName, toId, urn);
  if(*to != NULL) close_undo_redo_file(*to);
  *to = new_undo_redo_file(tmpName);
  util_free_small(tmpName);

  copyURRec(*to, from);
}

void deleteCheckpoint(cow_disk_t cow, unsigned long long id) {
  struct checkpoint_struct *cpDel;

  assert(cow->curr != NULL);

  // find the checkpoint in question
  cpDel = cow->curr;
  while(id != cpDel->id) {
    if(id > cpDel->id) {
      assert(cpDel->next != NULL);
      assert(cpDel->next->id > cpDel->id);
      cpDel = cpDel->next;
    } else {
      assert(cpDel->prev != NULL);
      assert(cpDel->prev->id < cpDel->id);
      cpDel = cpDel->prev;
    }
  }

  if(cpRefCount(cpDel) > 1) {
    deleteCheckpointStruct(cpDel);
    return;
  } 

  // we know that the checkpoint will be deleted

  // we cannot delete the first checkpoint.  We will need it to get back
  // to the beginning, plus it would invalidata SEVERAL assumptions made
  // throught the cow_disk code
  if(cpDel == cow->head) return;

  // see if we have to add anything to the current undo log
  if( (cow->currId >= cpDel->id) && 
      ( (cpDel->next == NULL) || (cpDel->next->id > cow->currId) ) ) {
    // we have to update the currUndo
    assert(cow->curr == cpDel);
    // check if we are the head (which will not have an undo so no mergin is needed)
    if(cpDel->undo != NULL) {
      mergeURRec(cpDel->undo, cow->currUndo);
      close_undo_redo_file(cow->currUndo);
      cow->currUndo = new_undo_redo_file(cow->currUndoFileName);
      copyURRec(cow->currUndo, cpDel->undo);
    }
  }

  // now set up the adjacent checkpoints
  assert(cpDel->prev != NULL);
  assert(cpDel->prev->redo != NULL);
  if(cpDel->redo != NULL) {
    mergeURRec(cpDel->redo, cpDel->prev->redo);
    resetAndCopyURRec(cow, &cpDel->prev->redo, cpDel->redo, cpDel->prev->id,
		      NAME_REDO);
  } else {
    assert(cpDel->next == NULL);
  }

  if(cpDel->next != NULL) {
    mergeURRec(cpDel->undo, cpDel->next->undo);
    resetAndCopyURRec(cow, &cpDel->next->undo, cpDel->undo, cpDel->next->id,
		      NAME_UNDO);
  }

  // adjust the curr and tail pointers if needed
  if(cpDel->next == NULL) {
    // we are the tail
    assert(cpDel == cow->tail);
    cow->tail = cpDel;
  }

  if(cpDel == cow->curr) {
    cow->curr = cpDel->prev;
  }

  // now remove the checkpoint from the list
  cpDel->prev->next = cpDel->next;
  if(cpDel->next != NULL) cpDel->next->prev = cpDel->prev;

  // we want to free up the file data also
  if(cpDel->undo != NULL) close_unlink_undo_redo_file(cpDel->undo);
  if(cpDel->redo != NULL) close_unlink_undo_redo_file(cpDel->redo);
  cpDel->undo = NULL;
  cpDel->redo = NULL;
  deleteCheckpointStruct(cpDel);

}
*/
