/*
 * linux/drivers/char/dbraw.c
 *
 * Front-end dbraw character devices.  These can be bound to any block
 * devices to provide genuine Unix dbraw character device semantics.
 *
 * We reserve minor number 0 for a control interface.  ioctl()s on this
 * device are used to bind the other minor numbers to block devices.
 */

#include <linux/fs.h>
#include <linux/iobuf.h>
#include <linux/major.h>
#include <linux/blkdev.h>
#include <linux/raw.h>
#include <linux/capability.h>
#include <linux/smp_lock.h>
#include <asm/uaccess.h>

#include <linux/module.h>
#include <asm/mman.h>

#include <revirt/rlog.h> /* For revirt_suppress_sched() */

#include "dbraw.h"
#include "myutil.h"
#include "checkpoint.h"

#define DBRAW_DEFAULT_BLOCK_SIZE 512

#define DBRAW_PRENAME "/checkpoint/cow-log"
#define BITS_PER_BYTE 8

#define L1_SIZE   (65536*2)
#define L1_NUM_ENTRIES (L1_SIZE / sizeof(unsigned long *))
#define L2_SIZE   (PAGE_SIZE) 
#define L2_NUM_ENTRIES (L2_SIZE / sizeof(unsigned long))
#define DBRAW_MAP_SIZE (L1_NUM_ENTRIES * L2_SIZE)
#define DBRAW_NUM_MAP_ENTRIES (DBRAW_MAP_SIZE / sizeof(unsigned long))
#define MAP_ACC_SIZE (DBRAW_NUM_MAP_ENTRIES / BITS_PER_BYTE)

//#define dprintk(x...) 
#define dprintk printk

//#define SNAPSHOT_MAP
//#define DEBUG_COW_DISK

/*
struct block_map {
 	unsigned long *map;
};
*/

struct block_map {
	unsigned long **map;
};

typedef struct dbraw_device_data_s {
	struct block_device *binding;
	int inuse, sector_size, sector_bits;
	unsigned long limit;
	unsigned long guestNumBlocks; // the size of the guest disk
	struct semaphore mutex;
	long long rw_count;
	struct block_map *bMap;
	unsigned char *freeBlockBitmap;
	unsigned char *mapAcc;
	unsigned long currFreeBlock;
	int mapInit;
	struct list_head cp_list;
	struct semaphore cleaning_mutex;
} dbraw_device_data_t;

static dbraw_device_data_t dbraw_devices[256];

struct cow_disk_cp {
	struct list_head link;
	bitmap_t undo_bm;
	bitmap_t redo_bm;
	list_file_t dataFile;
	long long rw_count;
	unsigned long refCount;
#ifdef SNAPSHOT_MAP
	fileUtil_t tmpMap;
#endif
};

extern asmlinkage long sys_mmap2(unsigned long addr, unsigned long len,
                                 unsigned long prot, unsigned long flags,
                                 unsigned long fd, unsigned long pgoff);
extern asmlinkage long sys_munmap(unsigned long addr, size_t len);

static ssize_t rw_dbraw_dev(int rw, struct file *, char *, size_t, loff_t *,int);

ssize_t	dbraw_read(struct file *, char *, size_t, loff_t *);
ssize_t	dbraw_write(struct file *, const char *, size_t, loff_t *);
int	dbraw_open(struct inode *, struct file *);
int	dbraw_release(struct inode *, struct file *);
int	dbraw_ctl_ioctl(struct inode *, struct file *, unsigned int, unsigned long);
int	dbraw_ioctl(struct inode *, struct file *, unsigned int, unsigned long);

static int freeBlockIsSet(int minor, unsigned long block);
static void reclaim_free_blocks(int minor, struct file *filp);
static void reset_free_blocks(int minor);
//static void freeVisitedBlocks(struct cow_disk_cp *cp, int minor);
static unsigned long checkpointCb(int action, unsigned long data, 
				  int minor);
static void getCurrNextCp(struct cow_disk_cp **curr,
			  struct cow_disk_cp **next,
			  long long rw_count,
			  int minor);

static struct file_operations dbraw_fops = {
	read:		dbraw_read,
	write:		dbraw_write,
	open:		dbraw_open,
	release:	dbraw_release,
	ioctl:          dbraw_ioctl,
};

static struct file_operations dbraw_ctl_fops = {
	ioctl:		dbraw_ctl_ioctl,
	open:		dbraw_open,
};


static unsigned long getGuestNumBlocks(int minor) {
	return dbraw_devices[minor].guestNumBlocks;
}

static unsigned long getTotalNumBlock(int minor) {
	return dbraw_devices[minor].limit;
}

static unsigned long getFreeBlockSize(int minor) {
	return getTotalNumBlock(minor) / BITS_PER_BYTE;
}

static struct block_map *alloc_block_map(void) {
	struct block_map *bMap;
	
	bMap = util_alloc_small(sizeof(*bMap));
	bMap->map = NULL;

	return bMap;
}

static void free_block_map(struct block_map *bMap) {
	unsigned long idx;

	if(bMap->map) {
		for(idx = 0; idx < L1_NUM_ENTRIES; idx++) {
			if(bMap->map[idx]) {
				free_page((unsigned long) bMap->map[idx]);
			}
		}
		util_free_large(bMap->map);
		bMap->map = NULL;
	}

	util_free_small(bMap);
}

static int freeBlockListSet = 0;

static unsigned long lookup_block_map(int minor, 
				      unsigned long guestAddr) {
	struct block_map *bMap = dbraw_devices[minor].bMap;
	unsigned long l1_offset = (guestAddr / L2_NUM_ENTRIES);
	unsigned long l2_offset = (guestAddr % L2_NUM_ENTRIES);
	assert(guestAddr < DBRAW_NUM_MAP_ENTRIES);
	if(bMap->map == NULL) {
		bMap->map = util_alloc_large(L1_SIZE);
		memset(bMap->map, 0, L1_SIZE);
	}
	if(bMap->map[l1_offset] == NULL) {
		bMap->map[l1_offset] = (unsigned long *) __get_free_page(GFP_KERNEL);
		assert(bMap->map[l1_offset] != NULL);
		memset(bMap->map[l1_offset], 0, PAGE_SIZE);
	}

	assert(bMap->map[l1_offset][l2_offset] < getTotalNumBlock(minor));
	if(bMap->map[l1_offset][l2_offset] && freeBlockListSet) {
		assert(!freeBlockIsSet(minor, bMap->map[l1_offset][l2_offset]));
	}
	return bMap->map[l1_offset][l2_offset];
}

static void set_block_map(int minor, unsigned long guestAddr,
			  unsigned long hostAddr) {
	unsigned long l1_offset = (guestAddr / L2_NUM_ENTRIES);
	unsigned long l2_offset = (guestAddr % L2_NUM_ENTRIES);
	struct block_map *bMap = dbraw_devices[minor].bMap;
	assert(guestAddr < DBRAW_NUM_MAP_ENTRIES);
	assert(hostAddr < getTotalNumBlock(minor));
	if(bMap->map == NULL) {
		bMap->map = util_alloc_large(L1_SIZE);
		memset(bMap->map, 0, L1_SIZE);
	}
	if(bMap->map[l1_offset] == NULL) {
		bMap->map[l1_offset] = (unsigned long *) __get_free_page(GFP_KERNEL);
		assert(bMap->map[l1_offset] != NULL);
		memset(bMap->map[l1_offset], 0, PAGE_SIZE);
	}
	
	bMap->map[l1_offset][l2_offset] = hostAddr;
}

/*
static struct block_map *alloc_block_map(void) {
	struct block_map *bMap;

	printk("dbraw: map size = %lu\n",DBRAW_MAP_SIZE);

	bMap = util_alloc_small(sizeof(*bMap));
	bMap->map = util_alloc_large(DBRAW_MAP_SIZE);

	return bMap;
}

static void free_block_map(struct block_map *bMap) {
	if(bMap->map) {
		util_free_large(bMap->map);
		bMap->map = NULL;
	}

	util_free_small(bMap);
}

static unsigned long lookup_block_map(struct block_map *bMap, 
				      unsigned long guestAddr) {
	return bMap->map[guestAddr];
}

static void set_block_map(struct block_map *bMap, unsigned long guestAddr,
			  unsigned long hostAddr) {
	
	bMap->map[guestAddr] = hostAddr;
}
*/

static void bitmapClear(unsigned char *buf, unsigned long block) {
	unsigned long byte = block / BITS_PER_BYTE;
	unsigned long bit = block % BITS_PER_BYTE;

	buf[byte] &= ~(0x1 << bit);
}

static void bitmapSet(unsigned char *buf, unsigned long block) {
	unsigned long byte = block / BITS_PER_BYTE;
	unsigned long bit = block % BITS_PER_BYTE;

	buf[byte] |= (0x1 << bit);
}

static int isBitmapSet(unsigned char *buf, unsigned long block) {
	unsigned long byte = block / BITS_PER_BYTE;
	unsigned long bit = block % BITS_PER_BYTE;

	return buf[byte] & (0x1 << bit);
}

static void freeBlockClear(int minor, unsigned long block) {
	bitmapClear(dbraw_devices[minor].freeBlockBitmap, block);
}

static void freeBlockSet(int minor, unsigned long block) {
	bitmapSet(dbraw_devices[minor].freeBlockBitmap, block);
}

static int freeBlockIsSet(int minor, unsigned long block) {	
	return isBitmapSet(dbraw_devices[minor].freeBlockBitmap, block);
}

/*
static int __init dbraw_init(void)
{
	int i;
	register_chrdev(DBRAW_MAJOR, "dbraw", &dbraw_fops);

	for (i = 0; i < 256; i++)
		init_MUTEX(&dbraw_devices[i].mutex);

	return 0;
}

__initcall(dbraw_init);
*/


int init_module(void) {
	int i;
	register_chrdev(DBRAW_MAJOR, "dbraw", &dbraw_fops);
	
	for (i = 0; i < 256; i++) {
		init_MUTEX(&dbraw_devices[i].mutex);
		init_MUTEX(&dbraw_devices[i].cleaning_mutex);
		
		dbraw_devices[i].limit = 0;
		dbraw_devices[i].guestNumBlocks = 0;
		dbraw_devices[i].rw_count = 0;
		dbraw_devices[i].bMap = NULL;
		dbraw_devices[i].freeBlockBitmap = NULL;
		dbraw_devices[i].mapAcc = NULL;
		dbraw_devices[i].currFreeBlock = 0;
		dbraw_devices[i].mapInit = 0;
		
		INIT_LIST_HEAD(&dbraw_devices[i].cp_list);
	}
	
	return 0;
}

void cleanup_module(void) {
	int i;

	for (i = 0; i < 256; i++) {
		if(dbraw_devices[i].bMap) {
			free_block_map(dbraw_devices[i].bMap);
			dbraw_devices[i].bMap = NULL;
		}

		if(dbraw_devices[i].mapAcc) {
			util_free_large(dbraw_devices[i].mapAcc);
			dbraw_devices[i].mapAcc = NULL;
		}

		if(dbraw_devices[i].freeBlockBitmap) {
			util_free_large(dbraw_devices[i].freeBlockBitmap);
			dbraw_devices[i].freeBlockBitmap = NULL;
		}
	}
	unregister_chrdev(DBRAW_MAJOR, "dbraw");

	registerCowDiskCpCb(NULL, -1);
}

/* 
 * Open/close code for dbraw IO.
 */

int dbraw_open(struct inode *inode, struct file *filp)
{
	int minor;
	struct block_device * bdev;
	kdev_t rdev;	/* it should eventually go away */
	int err;
	int sector_size;
	int sector_bits;
	unsigned long limit;

	minor = MINOR(inode->i_rdev);

	/* 
	 * Is it the control device? 
	 */
	
	if (minor == 0) {
		filp->f_op = &dbraw_ctl_fops;
		return 0;
	}
	
	if (!filp->f_iobuf) {
		err = alloc_kiovec(1, &filp->f_iobuf);
		if (err)
			return err;
	}

	down(&dbraw_devices[minor].mutex);
	/*
	 * No, it is a normal dbraw device.  All we need to do on open is
	 * to check that the device is bound, and force the underlying
	 * block device to a sector-size blocksize. 
	 */
	
	bdev = dbraw_devices[minor].binding;
	err = -ENODEV;
	if (!bdev)
		goto out;

	atomic_inc(&bdev->bd_count);
	rdev = to_kdev_t(bdev->bd_dev);
	err = blkdev_get(bdev, filp->f_mode, 0, BDEV_RAW);
	if (err)
		goto out;
	
	/*
	 * Don't change the blocksize if we already have users using
	 * this device 
	 */

	if (dbraw_devices[minor].inuse++)
		goto out;

	/* 
	 * Don't interfere with mounted devices: we cannot safely set
	 * the blocksize on a device which is already mounted.  
	 */
	
	sector_size = 512;
	if (is_mounted(rdev)) {
		if (blksize_size[MAJOR(rdev)])
			sector_size = blksize_size[MAJOR(rdev)][MINOR(rdev)];
	} else {
		if (hardsect_size[MAJOR(rdev)])
			sector_size = hardsect_size[MAJOR(rdev)][MINOR(rdev)];
	}

	if(sector_size != DBRAW_DEFAULT_BLOCK_SIZE) {
	  dprintk("block device block size != %d (it is %d)\n",
		  DBRAW_DEFAULT_BLOCK_SIZE,sector_size);
	}
	
	if(sector_size > PAGE_SIZE) {
	  dprintk("Warning, block size is larger than a page, you will not\n");
	  dprintk("be able to reclaim free blocks on the dbraw device\n");
	}

	set_blocksize(rdev, sector_size);
	dbraw_devices[minor].sector_size = sector_size;

	for (sector_bits = 0; !(sector_size & 1); )
		sector_size>>=1, sector_bits++;
	dbraw_devices[minor].sector_bits = sector_bits;

	if (blk_size[MAJOR(rdev)])
		limit = (((loff_t) blk_size[MAJOR(rdev)][MINOR(rdev)]) << BLOCK_SIZE_BITS) >> sector_bits;
	else
		limit = 0;

	dbraw_devices[minor].limit = limit;
	dprintk("raw device %d, limit = %lu, sector size = %d\n",
		minor,limit,dbraw_devices[minor].sector_size);

 out:
	up(&dbraw_devices[minor].mutex);

	if(!err)
		MOD_INC_USE_COUNT;

	return err;
}

int dbraw_release(struct inode *inode, struct file *filp)
{
	int minor;
	struct block_device *bdev;

	MOD_DEC_USE_COUNT;

	minor = MINOR(inode->i_rdev);

	down(&dbraw_devices[minor].mutex);
	bdev = dbraw_devices[minor].binding;
	dbraw_devices[minor].inuse--;
	up(&dbraw_devices[minor].mutex);
	blkdev_put(bdev, BDEV_RAW);
	return 0;
}

static int flushMap(int minor, struct dbraw_cmd_struct *cmd,  struct file *filp) {
	unsigned long idx;
	loff_t offset;
	int ret;
	unsigned long guestAddr, hostAddr,i;
	unsigned long size;
	unsigned long blockSize;

	blockSize = dbraw_devices[minor].sector_size;

	assert(blockSize > 0);

	if(cmd->bufSize < blockSize) {
		dprintk("block size < DBRAW_BLOCK_SIZE\n");
		return -EINVAL;
	}
	if(cmd->bufSize % blockSize) {
		dprintk("buffer size must be multiple of block size\n");
		return -EINVAL;
	}
	if(((unsigned long) cmd->buf) & (blockSize-1)) {
		dprintk("buffer not block aligned\n");
		return -EINVAL;
	}
	if(!dbraw_devices[minor].bMap) {
		dprintk("no block map loaded in memory\n");
		return -EINVAL;
	}

	dprintk("flushing map\n");
	guestAddr = 0;
	for(idx = 0; idx < DBRAW_MAP_SIZE; idx += cmd->bufSize) {

		if((DBRAW_MAP_SIZE - idx) < cmd->bufSize) {
			size = (DBRAW_MAP_SIZE - idx);
		} else {
			size = cmd->bufSize;
		}

		for(i = 0; i < size; i += sizeof(unsigned long)) {
			hostAddr = lookup_block_map(minor,guestAddr);
						    
			ret = copy_to_user(cmd->buf + i,
					   &hostAddr,
					   sizeof(unsigned long));
			if(ret != 0) {
				dprintk("could not copy buffer data\n");
				return ret;
			}
			guestAddr++;
		}

		offset = idx;		
		ret = rw_dbraw_dev(WRITE,filp, cmd->buf, size, &offset,1);
		if(ret != size) {
			dprintk("error writing out map\n");
			return ret;
		}
	}

	return 0;
}

static int initMap(int minor, struct dbraw_cmd_struct *cmd,  struct file *filp) {
	unsigned long idx;
	loff_t offset;
	int ret;
	unsigned long guestAddr, hostAddr, i;
	unsigned long size;
	unsigned long blockSize;

	blockSize = dbraw_devices[minor].sector_size;
	assert(blockSize > 0);

	if(cmd->bufSize < blockSize) {
		dprintk("block size < DBRAW_BLOCK_SIZE\n");
		return -EINVAL;
	}
	if(cmd->bufSize % blockSize) {
		dprintk("buffer size must be multiple of block size\n");
		return -EINVAL;
	}
	if(((unsigned long) cmd->buf) & (blockSize-1)) {
		dprintk("buffer not block aligned\n");
		return -EINVAL;
	}

	// setup the guest blocks
	if(cmd->guestNumBlocks >= getTotalNumBlock(minor)) {
		dprintk("too many guest blocks, there are only %lu pysical blocks\n",
			getTotalNumBlock(minor));
		return -EINVAL;
	}


	// initialize the map
	dprintk("init map\n");	
	if(!dbraw_devices[minor].bMap) {
		dbraw_devices[minor].bMap = alloc_block_map();
		if(dbraw_devices[minor].bMap == NULL) {
			panic("could not alloc buffer\n");
		}
	}
	// zero out the initial map
	for(idx = 0; idx < DBRAW_NUM_MAP_ENTRIES; idx++) {
		set_block_map(minor,idx, 0);
	}

	dprintk("read in init map\n");
	guestAddr = 0;
	for(idx = 0; idx < DBRAW_MAP_SIZE; idx += cmd->bufSize) {
		offset = idx;
		
		if((DBRAW_MAP_SIZE - idx) < cmd->bufSize) {
			size = (DBRAW_MAP_SIZE - idx);
		} else {
			size = cmd->bufSize;
		}

		ret = rw_dbraw_dev(READ,filp, cmd->buf, size, &offset,1);
		if(ret != size) {
			dprintk("error reading in map\n");
			return ret;
		}

		for(i = 0; i < size; i += sizeof(unsigned long)) {
			ret = copy_from_user(&hostAddr, cmd->buf + i, 
					     sizeof(unsigned long));
			if(ret != 0) {
				dprintk("could not copy buffer data\n");
				return ret;
			}
			set_block_map(minor,guestAddr, hostAddr);
			guestAddr++;
		}
	}
	

	// setup the bitmap for the map of accessed blocks
	if(dbraw_devices[minor].mapAcc) 
		util_free_large(dbraw_devices[minor].mapAcc);
	dbraw_devices[minor].mapAcc = util_alloc_large(MAP_ACC_SIZE);
	memset(dbraw_devices[minor].mapAcc,0,MAP_ACC_SIZE);

	// setup the freeblock bitmap
	reset_free_blocks(minor);

	dbraw_devices[minor].mapInit = 1;
	dbraw_devices[minor].currFreeBlock = 0;

	dbraw_devices[minor].guestNumBlocks = cmd->guestNumBlocks;

	return 0;
}
static int checksumHostBlock(int minor, struct dbraw_cmd_struct *cmd,  
			     struct file *filp) {
	unsigned long checksum, idx;
	int ret;
	loff_t offset;
	unsigned long *data;
	unsigned long blockSize;

	blockSize = dbraw_devices[minor].sector_size;
	assert(blockSize > 0);

	if(cmd->bufSize != blockSize) {
		dprintk("block size != DBRAW_BLOCK_SIZE\n");
		return -EINVAL;
	}
	if(((unsigned long) cmd->buf) & (blockSize-1)) {
		dprintk("buffer not block aligned\n");
		return -EINVAL;
	}

	offset = cmd->hostAddr * blockSize;
	ret = rw_dbraw_dev(READ,filp, cmd->buf, blockSize, &offset,1);
	if(ret != blockSize) {
		dprintk("error reading in map\n");
		return ret;
	}

	data = util_alloc_small(blockSize);
	
	ret = copy_from_user(data, cmd->buf, blockSize);
	if(ret != 0) {
		dprintk("could not copy from user\n");
		util_free_small(data);
		return ret;
	}
	
	checksum = 0;
	for(idx = 0; idx < (blockSize / sizeof(unsigned long)); idx++) {
		checksum += data[idx];
	}
	util_free_small(data);

	dprintk("checksum for %lu = %lu\n",cmd->hostAddr, checksum);

	return 0;
}

static int getSize(int minor, struct dbraw_size_struct *size) {
	unsigned long idx;

	printk("getting the size\n");
	size->blockSize = dbraw_devices[minor].sector_size;
	size->numVirt = getGuestNumBlocks(minor);
	size->numPhys = getTotalNumBlock(minor);

	printk("calc the number of free blocks, numPys = %lu, blockSize = %lu\n",
	       size->numPhys, size->blockSize);
	size->numFree = 0;
	for(idx = 0; idx < getTotalNumBlock(minor); idx++) {
		if(freeBlockIsSet(minor, idx)) {
			size->numFree++;
		}
	}
	printk("done, found %lu free blocks\n",size->numFree);

	return 0;
}

int dbraw_ioctl(struct inode *inode,
		struct file *filp,
		unsigned int command,
		unsigned long arg)
{
	int err = 0;
	int minor;
	struct dbraw_cmd_struct cmd;
	struct dbraw_size_struct size;
	int ret;

	minor = MINOR(filp->f_dentry->d_inode->i_rdev);
	
	if((command == DBRAW_INIT_MAP) ||
	   (command == DBRAW_FLUSH_MAP) ||
	   (command == DBRAW_CHECKSUM_HOST_BLOCK)) {
		ret = copy_from_user(&cmd, (void *) arg, sizeof(cmd));
		if(ret != 0)
			return ret;
	} 
	
	if((command != DBRAW_INIT_MAP) && 
	   !dbraw_devices[minor].mapInit &&
	   (command != DBRAW_GET_BLOCK_SIZE) ) {
	  dprintk("map not init\n");
	  return -EINVAL;
	}

	switch (command) {
	case DBRAW_GET_BLOCK_SIZE:
		ret = copy_to_user((void *) arg, &dbraw_devices[minor].sector_size,
				   sizeof(dbraw_devices[minor].sector_size));
		if(ret != 0)
			return ret;
		break;
	case DBRAW_INIT_MAP:
		err = initMap(minor, &cmd, filp);
		break;
	case DBRAW_FLUSH_MAP:
		err = flushMap(minor, &cmd, filp);
		break;
	case DBRAW_CHECKSUM_HOST_BLOCK:
		err = checksumHostBlock(minor, &cmd, filp);
		break;
	case DBRAW_GET_SIZES:
		err = getSize(minor, &size);
		ret = copy_to_user((void *) arg, &size, sizeof(size));
		if(ret != 0) 
			return ret;
		break;
	case DBRAW_RECLAIM_FREE_BLOCKS:
		err = 0;
		reclaim_free_blocks(minor, filp);
		break;
	default:
		err = -EINVAL;
	}
	
	return err;
}

/*
 * Deal with ioctls against the dbraw-device control interface, to bind
 * and unbind other dbraw devices.  
 */

int dbraw_ctl_ioctl(struct inode *inode, 
		  struct file *flip,
		  unsigned int command, 
		  unsigned long arg)
{
	struct raw_config_request rq;
	int err = 0;
	int minor;

	switch (command) {
	case RAW_SETBIND:
	case RAW_GETBIND:

		/* First, find out which dbraw minor we want */

		err = copy_from_user(&rq, (void *) arg, sizeof(rq));
		if (err)
			break;
		
		minor = rq.raw_minor;
		if (minor <= 0 || minor > MINORMASK) {
			err = -EINVAL;
			break;
		}

		if (command == RAW_SETBIND) {
			/*
			 * This is like making block devices, so demand the
			 * same capability
			 */
			if (!capable(CAP_SYS_ADMIN)) {
				err = -EPERM;
				break;
			}

			/* 
			 * For now, we don't need to check that the underlying
			 * block device is present or not: we can do that when
			 * the dbraw device is opened.  Just check that the
			 * major/minor numbers make sense. 
			 */

			if ((rq.block_major == NODEV && 
			     rq.block_minor != NODEV) ||
			    rq.block_major > MAX_BLKDEV ||
			    rq.block_minor > MINORMASK) {
				err = -EINVAL;
				break;
			}
			
			down(&dbraw_devices[minor].mutex);
			if (dbraw_devices[minor].inuse) {
				up(&dbraw_devices[minor].mutex);
				err = -EBUSY;
				break;
			}
			if (dbraw_devices[minor].binding)
				bdput(dbraw_devices[minor].binding);
			dbraw_devices[minor].binding = 
				bdget(kdev_t_to_nr(MKDEV(rq.block_major, rq.block_minor)));
			up(&dbraw_devices[minor].mutex);
			
			registerCowDiskCpCb(checkpointCb, minor);
		} else {
			struct block_device *bdev;
			kdev_t dev;

			bdev = dbraw_devices[minor].binding;
			if (bdev) {
				dev = to_kdev_t(bdev->bd_dev);
				rq.block_major = MAJOR(dev);
				rq.block_minor = MINOR(dev);
			} else {
				rq.block_major = rq.block_minor = 0;
			}
			err = copy_to_user((void *) arg, &rq, sizeof(rq));
		}
		break;
		
	default:
		err = -EINVAL;
	}
	
	return err;
}



ssize_t	dbraw_read(struct file *filp, char * buf, 
		 size_t size, loff_t *offp)
{
	return rw_dbraw_dev(READ, filp, buf, size, offp,0);
}

ssize_t	dbraw_write(struct file *filp, const char *buf, 
		  size_t size, loff_t *offp)
{
	return rw_dbraw_dev(WRITE, filp, (char *) buf, size, offp,0);
}

#define SECTOR_BITS 9
#define SECTOR_SIZE (1U << SECTOR_BITS)
#define SECTOR_MASK (SECTOR_SIZE - 1)

static void reset_free_blocks(int minor) {
	unsigned long idx, hostAddr;
	unsigned long blockSize;

	blockSize = dbraw_devices[minor].sector_size;
	assert(blockSize > 0);

	freeBlockListSet = 0;

	dprintk("setting up freeblock bitmap\n");
	if(dbraw_devices[minor].freeBlockBitmap == NULL) {
		dbraw_devices[minor].freeBlockBitmap = 
			util_alloc_large(getFreeBlockSize(minor));
	}

	memset(dbraw_devices[minor].freeBlockBitmap, 0, getFreeBlockSize(minor));

	dprintk("adding all blocks to free list\n");
	for(idx = 0; idx < getTotalNumBlock(minor); idx++) {
		freeBlockSet(minor, idx);
	}
	
	dprintk("removing mapped blocks from free list\n");
	for(idx = 0; idx < DBRAW_NUM_MAP_ENTRIES; idx++) {
		hostAddr = lookup_block_map(minor, idx);
		if(hostAddr) {
			freeBlockClear(minor, hostAddr);
		}
	}

	dprintk("removing map blocks from free list\n");
	for(idx = 0; idx < (DBRAW_MAP_SIZE / blockSize); idx++) {
		freeBlockClear(minor, idx);
	}

	freeBlockListSet = 1;
}

static void reclaim_free_blocks(int minor, struct file *filp) {
	struct cow_disk_cp *cp;
	unsigned long idx, numRec, hostAddr, guestAddr, i;
	unsigned char *userMem;
	loff_t offset;
	int ret;
	unsigned long blockSize;

	down(&dbraw_devices[minor].cleaning_mutex);

	blockSize = dbraw_devices[minor].sector_size;
	assert(blockSize > 0);

	reset_free_blocks(minor);

	// now make sure any block being held by checkpoints are
	// taken off the free list
	dprintk("removing checkpoint blocks from free list\n");
	list_for_each_entry(cp, &dbraw_devices[minor].cp_list, link) {
		numRec = lst_get_num_rec(cp->dataFile);
		for(idx = 0; idx < numRec; idx++) {
			lst_get_rec_by_idx(cp->dataFile, idx, &hostAddr, NULL);
			freeBlockClear(minor, hostAddr);
		}
	}

	// read the inital map off of the disk and make sure we do
	// not allow the blocks to be re-allocated
	dprintk("removing initial map from free list\n");
	userMem = (unsigned char *) sys_mmap2(0, PAGE_SIZE, PROT_READ | PROT_WRITE,
					      MAP_PRIVATE | MAP_ANONYMOUS, 0, 0);
	assert(userMem != NULL);
	assert(PAGE_SIZE >= blockSize);
	assert(((unsigned long) userMem & (blockSize-1)) == 0);
	guestAddr = 0;
	for(idx = 0; idx < DBRAW_MAP_SIZE; idx += blockSize) {
		offset = idx;
		ret = rw_dbraw_dev(READ,filp, userMem, blockSize, &offset,1);
		if(ret != blockSize) {
			dprintk("error reading in map\n");
			return;
		}

		for(i = 0; i < blockSize; i += sizeof(unsigned long)) {
			ret = copy_from_user(&hostAddr, userMem + i, 
					     sizeof(unsigned long));
			if(ret != 0) {
				dprintk("could not copy buffer data\n");
				return;
			}
			if(hostAddr) {
				freeBlockClear(minor, hostAddr);
			}
			guestAddr++;
		}
	}

	sys_munmap((unsigned long) userMem, PAGE_SIZE);

	up(&dbraw_devices[minor].cleaning_mutex);
}

static unsigned long alloc_new_block(int minor, struct file *filp) {
	unsigned long count = 0;
	unsigned long ret;

	down(&dbraw_devices[minor].cleaning_mutex);
	while(!freeBlockIsSet(minor, dbraw_devices[minor].currFreeBlock)) {
		count++;
		dbraw_devices[minor].currFreeBlock++;
		if(dbraw_devices[minor].currFreeBlock >= getTotalNumBlock(minor)) {
			printk("rolling over free block list\n");
			dbraw_devices[minor].currFreeBlock = 0;
		}

		if(count == getTotalNumBlock(minor)) {
			dprintk("no more free blocks, running reclamation\n");
			up(&dbraw_devices[minor].cleaning_mutex);
			reclaim_free_blocks(minor, filp);
			down(&dbraw_devices[minor].cleaning_mutex);
		}

		if(count > (2*getTotalNumBlock(minor))) {
			dprintk("OOM, no more free blocks\n");
			up(&dbraw_devices[minor].cleaning_mutex);
			assert(0);
			return (unsigned long) -1;
		}
	}

	freeBlockClear(minor, dbraw_devices[minor].currFreeBlock);
	ret = dbraw_devices[minor].currFreeBlock++;
	if(dbraw_devices[minor].currFreeBlock >= getTotalNumBlock(minor)) {
		printk("rolling over free block list\n");
		dbraw_devices[minor].currFreeBlock = 0;
	}

	up(&dbraw_devices[minor].cleaning_mutex);

	return ret;
}

static void update_undo(struct cow_disk_cp *cp, unsigned long guest_block,
			unsigned long host_block) {
	if(!util_is_bit_set(cp->undo_bm, guest_block)) {
		util_set_bit(cp->undo_bm, guest_block);

		if(!lst_is_rec_present(cp->dataFile, guest_block)) {
			lst_add_rec(cp->dataFile, guest_block, &host_block);
		}
	}
}

static unsigned long get_block(unsigned long guest_block, int rw, int minor,
			       int ignoreMap, struct file *filp) {
	unsigned long ret = (unsigned long) -1;
	struct cow_disk_cp *curr, *next;
	unsigned long hostAddr;

	if((guest_block >= getGuestNumBlocks(minor)) && !ignoreMap) {
		dprintk("deny requested guest block %lu\n",guest_block);
		return (unsigned long) -1;
	}


	// before the map is initialized, just a 1 to 1 mapping between
	// guest_blocks and host_blocks
	if(!dbraw_devices[minor].mapInit || ignoreMap) {
		ret = guest_block;
	} else {
		if(rw == READ) {
			hostAddr = lookup_block_map(minor,guest_block);
			if(hostAddr == 0) {
				ret = alloc_new_block(minor, filp);
				set_block_map(minor,guest_block, ret);
			} else {
				ret = hostAddr;
			}
		} else {
			hostAddr = lookup_block_map(minor,guest_block);
			if(isBitmapSet(dbraw_devices[minor].mapAcc, guest_block)) {
				// resuse old block, already new one alloc
				ret = hostAddr;
				if(ret == 0) {
					dprintk("inconsistent mapAcc state");
					assert(0);
				}
			} else {
				ret = alloc_new_block(minor, filp);

				getCurrNextCp(&curr, &next,
					      dbraw_devices[minor].rw_count,
					      minor);

				// only update the undo of the cp
				// when we are at the tail.  otherwise
				// it is already set when it was taken
				if((curr != NULL) && (next == NULL)) {
					update_undo(curr, guest_block,hostAddr);
				}
						
				set_block_map(minor,guest_block, ret);
				bitmapSet(dbraw_devices[minor].mapAcc, guest_block);
			}
		}
	}

	return ret;
}

static void resetVisitedBlocks(int minor) {
	memset(dbraw_devices[minor].mapAcc, 0, MAP_ACC_SIZE);
}

ssize_t	rw_dbraw_dev(int rw, struct file *filp, char *buf, 
		   size_t size, loff_t *offp, int ignoreMap)
{
	struct kiobuf * iobuf;
	int		new_iobuf;
	int		err = 0;
	unsigned long	blocknr, blocks;
	size_t		transferred;
	int		iosize;
	int		i;
	int		minor;
	kdev_t		dev;
	unsigned long	limit;

	int		sector_size, sector_bits, sector_mask;
	int		max_sectors;
	
	if(rw == READ) modifiedUserAddr(current, (unsigned long) buf, size);

	/*
	 * First, a few checks on device size limits 
	 */

	minor = MINOR(filp->f_dentry->d_inode->i_rdev);

	new_iobuf = 0;
	iobuf = filp->f_iobuf;
	if (test_and_set_bit(0, &filp->f_iobuf_lock)) {
		/*
		 * A parallel read/write is using the preallocated iobuf
		 * so just run slow and allocate a new one.
		 */
		err = alloc_kiovec(1, &iobuf);
		if (err)
			goto out;
		new_iobuf = 1;
	}

	dev = to_kdev_t(dbraw_devices[minor].binding->bd_dev);
	sector_size = dbraw_devices[minor].sector_size;
	sector_bits = dbraw_devices[minor].sector_bits;
	sector_mask = sector_size- 1;
	max_sectors = KIO_MAX_SECTORS >> (sector_bits - 9);

	/*
	if (blk_size[MAJOR(dev)])
		limit = (((loff_t) blk_size[MAJOR(dev)][MINOR(dev)]) << BLOCK_SIZE_BITS) >> sector_bits;
	else
		limit = INT_MAX;
	*/
	if(ignoreMap) {
		limit = getTotalNumBlock(minor);
	} else {
		limit = getGuestNumBlocks(minor);
	}

	if(limit == 0) {
		dprintk("invalid: guest size not set\n");
		err = -EINVAL;
		goto out_free;
	}

	/*
	dprintk ("rw_dbraw_dev: dev %d:%d (+%d)\n",
		 MAJOR(dev), MINOR(dev), limit);
	*/

	err = -EINVAL;
	if ((*offp & sector_mask) || (size & sector_mask))
		goto out_free;
	err = 0;
	if (size)
		err = -ENXIO;
	if ((*offp >> sector_bits) >= limit)
		goto out_free;

	/*
	 * Split the IO into KIO_MAX_SECTORS chunks, mapping and
	 * unmapping the single kiobuf as we go to perform each chunk of
	 * IO.  
	 */

	transferred = 0;
	blocknr = *offp >> sector_bits;
	while (size > 0) {
		blocks = size >> sector_bits;
		if (blocks > max_sectors)
			blocks = max_sectors;
		if (blocks > limit - blocknr)
			blocks = limit - blocknr;
		if (!blocks)
			break;

		iosize = blocks << sector_bits;

		err = map_user_kiobuf(rw, iobuf, (unsigned long) buf, iosize);
		if (err)
			break;

		for (i=0; i < blocks; i++) {
			struct cow_disk_cp *curr, *next;
			dbraw_devices[minor].rw_count++;
			
			iobuf->blocks[i] = get_block(blocknr++,rw,minor,
						     ignoreMap, filp);

			getCurrNextCp(&curr, &next,
				      dbraw_devices[minor].rw_count,
				      minor);

			// check if we need to reset our visited block list
			if(curr && 
			   (curr->rw_count == dbraw_devices[minor].rw_count)) {
				//freeVisitedBlocks(curr, minor);
				resetVisitedBlocks(minor);
			}

			if(iobuf->blocks[i] == ((unsigned long) -1)) {
				err = 0;
				goto out_free;
			}
		}
		if(err)
			break;

		err = brw_kiovec(rw, 1, &iobuf, dev, iobuf->blocks, sector_size);

		if (rw == READ && err > 0)
			mark_dirty_kiobuf(iobuf, err);
		
		if (err >= 0) {
			transferred += err;
			size -= err;
			buf += err;
		}

		unmap_kiobuf(iobuf);

		if (err != iosize)
			break;
	}
	
	if (transferred) {
		*offp += transferred;
		err = transferred;
	}

 out_free:
	if (!new_iobuf)
		clear_bit(0, &filp->f_iobuf_lock);
	else
		free_kiovec(1, &iobuf);
 out:	

	return err;
}

static struct cow_disk_cp *alloc_cow_disk_cp(long long rw_count) {
	char *tmpName;
	struct cow_disk_cp *cp = util_alloc_small(sizeof(*cp));
	tmpName = util_alloc_small(strlen(DBRAW_PRENAME)+50);

	sprintf(tmpName,"%s.%lld.undo",DBRAW_PRENAME, rw_count);
	cp->undo_bm = util_new_bitmap(tmpName);
	util_unlink(tmpName);

	sprintf(tmpName,"%s.%lld.redo",DBRAW_PRENAME, rw_count);
	cp->redo_bm = util_new_bitmap(tmpName);
	util_unlink(tmpName);

	sprintf(tmpName,"%s.%lld.data",DBRAW_PRENAME, rw_count);
	cp->dataFile = lst_new(tmpName, sizeof(unsigned long), 0);

	cp->rw_count = rw_count;
	cp->refCount = 1;

#ifdef SNAPSHOT_MAP
	sprintf(tmpName,"%s.%lld.tmp",DBRAW_PRENAME, rw_count);
	cp->tmpMap = util_open(tmpName, 
			       O_RDWR | O_LARGEFILE | O_EXCL | O_CREAT | O_TRUNC,
			       0600);
	util_unlink(tmpName);
#endif

	util_free_small(tmpName);
	return cp;
}

static void free_cow_disk_cp(struct cow_disk_cp *cp) {
#ifdef SNAPSHOT_MAP
	util_close(cp->tmpMap);
#endif
	util_delete_bitmap(cp->undo_bm);
	util_delete_bitmap(cp->redo_bm);
	
	lst_close_unlink(cp->dataFile);
	util_free_small(cp);
}

static void getCurrNextCp(struct cow_disk_cp **curr,
			  struct cow_disk_cp **next,
			  long long rw_count,
			  int minor) {
	struct cow_disk_cp *cp;

#ifdef DEBUG_COW_DISK
	struct cow_disk_cp *prev = NULL;
	list_for_each_entry(cp, &dbraw_devices[minor].cp_list, link) {
		if(prev != NULL) {
			assert(prev->rw_count < cp->rw_count);
		}
		prev = cp;
	}
#endif
	
	*curr = NULL;
	*next = NULL;

	list_for_each_entry(cp, &dbraw_devices[minor].cp_list, link) {
		if(cp->rw_count > rw_count) {
			*next = cp;
			if(cp->link.prev != &dbraw_devices[minor].cp_list) {
				*curr = list_entry(cp->link.prev,
						   struct cow_disk_cp,
						   link);
				return;
			}
		}
	}

	// if there are any entries, we are beyond the tail of the list
	if(dbraw_devices[minor].cp_list.prev != 
	   &dbraw_devices[minor].cp_list) {
		*curr = list_entry(dbraw_devices[minor].cp_list.prev,
				   struct cow_disk_cp,
				   link);
	}
}

static void merge_undo_redo(bitmap_t target_ur,
			    list_file_t target_data,
			    bitmap_t source_ur,
			    list_file_t source_data) {
	unsigned long idx, numRec;
	loff_t guest_addr;
	unsigned long data;
	unsigned long numCopied = 0;

	numRec = lst_get_num_rec(source_data);
	for(idx = 0; idx < numRec; idx++) {
		lst_get_rec_by_idx(source_data, idx, &data, &guest_addr);
		if(util_is_bit_set(source_ur, guest_addr)) {
			if(!util_is_bit_set(target_ur, guest_addr)) {
				util_set_bit(target_ur, guest_addr);

				if(!lst_is_rec_present(target_data, guest_addr)) {
					lst_add_rec(target_data, guest_addr, &data);
					numCopied++;
				} 
			}
		}
	}
	printk("merge_undo_redo, examined %lu records, copied %lu\n",
	       numRec, numCopied);
}

static void copy_undo_redo(bitmap_t target_ur,
			   list_file_t target_data,
			   bitmap_t source_ur,
			   list_file_t source_data,
			   int minor) {
	unsigned long idx, numRec;
	loff_t guest_addr;
	unsigned long hostAddr;
	
	numRec = lst_get_num_rec(source_data);
	for(idx = 0; idx < numRec; idx++) {
		lst_get_rec_by_idx(source_data, idx, NULL, &guest_addr);
		if(util_is_bit_set(source_ur, guest_addr)) {
			util_set_bit(target_ur, guest_addr);

			if(!lst_is_rec_present(target_data, guest_addr)) {
				hostAddr = lookup_block_map(minor, guest_addr);
				lst_add_rec(target_data, guest_addr, &hostAddr);
			}
		}
	}
}

static unsigned long takeCheckpoint(int minor) {
	struct cow_disk_cp *cp;
	struct cow_disk_cp *curr, *next;
	long long rw_count;

	rw_count = dbraw_devices[minor].rw_count;

	getCurrNextCp(&curr, &next, rw_count, minor);

	if(curr && (curr->rw_count > rw_count))
		assert(0);

	if(next && (next->rw_count <= rw_count))
		assert(0);

	if(curr && (curr->rw_count == rw_count)) {
		dprintk("incrementing refCount\n");
		curr->refCount++;
		return (unsigned long) curr;
	}

	cp = alloc_cow_disk_cp(rw_count);

	if(curr != NULL) {
		list_add(&cp->link,&curr->link);
	} else {
		list_add_tail(&cp->link, &dbraw_devices[minor].cp_list);
	}

	if(curr) {
		copy_undo_redo(cp->redo_bm,
			       cp->dataFile,
			       curr->undo_bm,
			       curr->dataFile,
			       minor);
	}

	if(next) {
		copy_undo_redo(cp->undo_bm,
			       cp->dataFile,
			       next->redo_bm,
			       next->dataFile,
			       minor);
	}
#ifdef SNAPSHOT_MAP
	dprintk("copying entire map\n");
	{
#define ONE_M (1024 * 1024)
		unsigned long guestAddr, hostAddr, idx,i;
		int ret;
		unsigned long *buf;
		buf = util_alloc_large(ONE_M);
		guestAddr = 0;
		for(idx = 0; idx < DBRAW_MAP_SIZE; idx += (ONE_M)) {
			for(i = 0; i < ((ONE_M) / sizeof(unsigned long)); i++) {
				hostAddr = lookup_block_map(minor,
							    guestAddr);
				buf[i] = hostAddr;
				guestAddr++;
			}
			ret = util_append(cp->tmpMap,
					  buf,
					  ONE_M);
			if(ret != ONE_M) {
				dprintk("did not copy map\n");
			}
		}
		assert(guestAddr == DBRAW_NUM_MAP_ENTRIES);
		util_free_large(buf);
	}
#endif
	
	// reset our record of blocks written to in the current
	// checkpoint interval
	resetVisitedBlocks(minor);

	return (unsigned long) cp;
}

static void restoreCowDiskMap(list_file_t dataFile, bitmap_t undoRedo, 
			      unsigned char *prevRest, int minor) {
	unsigned long numRec, idx;
	loff_t guest_addr;
	unsigned long host_addr;

	numRec = lst_get_num_rec(dataFile);

	for(idx =0; idx < numRec; idx++) {
		lst_get_rec_by_idx(dataFile, idx, &host_addr, &guest_addr);

		if(util_is_bit_set(undoRedo, guest_addr) &&
		   !isBitmapSet(prevRest, (unsigned long) guest_addr)) {

			bitmapSet(prevRest, (unsigned long) guest_addr);
			set_block_map(minor, guest_addr, host_addr);
		}
	}
}

static void restoreCheckpoint(struct cow_disk_cp *cp, int minor) {
	long long new_rw_count;
	int goBack;
	list_file_t dataFile;
	bitmap_t undoRedo;
	struct list_head *list;
#ifdef SNAPSHOT_MAP
	//fileUtil_t tmpMap = cp->tmpMap;
#endif
	//freeVisitedBlocks(NULL, minor);
	resetVisitedBlocks(minor);	       
	new_rw_count = cp->rw_count;

	if(dbraw_devices[minor].rw_count == cp->rw_count)
		return;

	goBack = dbraw_devices[minor].rw_count > cp->rw_count;
	list = &cp->link;
	while(list != &dbraw_devices[minor].cp_list) {
		cp = list_entry(list, struct cow_disk_cp, link);	       
		
		if(goBack && (cp->rw_count >= dbraw_devices[minor].rw_count)) {
			break;
		}
		if(!goBack && (cp->rw_count <= dbraw_devices[minor].rw_count)) {
			break;
		}

		assert(cp->refCount > 0);
		dataFile = cp->dataFile;
		undoRedo = (goBack) ? cp->undo_bm : cp->redo_bm;
		restoreCowDiskMap(dataFile, undoRedo,
				  dbraw_devices[minor].mapAcc,
				  minor);

		list = (goBack) ? list->next : list->prev;
	}
	

#ifdef SNAPSHOT_MAP
#warning "SNAPSHOT WILL NOT BE RESTORED (but cow map will)"
	/*
	dprintk("restoring entire map\n");
	{
		loff_t offset = 0;
		int ret = util_read(tmpMap, 
				    dbraw_devices[minor].blockMap,
				    DBRAW_MAP_SIZE,
				    offset);
		if(ret != DBRAW_MAP_SIZE) {
			dprintk("could not restore map\n");
		}
	}
	*/
#endif

	resetVisitedBlocks(minor);	       
	dbraw_devices[minor].rw_count = new_rw_count;

}

/*
static void freeBlocksAfterDel(struct cow_disk_cp *cp,
			       struct cow_disk_cp *prev,
			       struct cow_disk_cp *next,
			       unsigned char *freeBlockBitmap) {
	unsigned long hostAddr, otherHostAddr;
	loff_t guestAddr;
	unsigned long idx, numRec;
	int match;

	return;

	dprintk("starting freeBlocksAfterDel\n");

	numRec = lst_get_num_rec(cp->dataFile);
	for(idx = 0; idx < numRec; idx++) {
		lst_get_rec_by_idx(cp->dataFile, idx, &hostAddr, &guestAddr);
		
		match = 0;
		if((prev != NULL) && util_is_bit_set(cp->redo_bm, guestAddr)) {
			assert(util_is_bit_set(prev->undo_bm, guestAddr));
			assert(lst_is_rec_present(prev->dataFile, guestAddr));
			lst_get_rec(prev->dataFile, guestAddr, &otherHostAddr);
			if(otherHostAddr == hostAddr) {
				match = 1;
			}
		}

		if((next != NULL) && util_is_bit_set(cp->undo_bm, guestAddr)) {
			assert(util_is_bit_set(next->redo_bm, guestAddr));
			assert(lst_is_rec_present(next->dataFile, guestAddr));
			lst_get_rec(next->dataFile, guestAddr, &otherHostAddr);
			if(otherHostAddr == hostAddr) {
				match = 1;
			}
		}
		
		if(!match) {
			bitmapSet(freeBlockBitmap, hostAddr);
		}
	}

	dprintk("done freeBlocksAfterDel\n");
}
 
static void freeVisitedBlocks(struct cow_disk_cp *cp, int minor) {
	unsigned char *freeBlockList;
	unsigned char *mapAcc;
	struct block_map *bMap;
	unsigned long hostAddr, guestAddr;

	return;

	dprintk("starting freeVisitedBlocks\n");

	freeBlockList = dbraw_devices[minor].freeBlockBitmap;
	mapAcc = dbraw_devices[minor].mapAcc;
	bMap = dbraw_devices[minor].bMap;

	// if we are at a new checkpoint, use the values from that map
	// to replace the values of the freed blocks.  Otherwise, just clear
	// and they will be replaced when the undo or redo log is applied

	for(guestAddr = 0; guestAddr < DBRAW_NUM_MAP_ENTRIES; guestAddr++) {
		if(isBitmapSet(mapAcc, guestAddr)) {
			hostAddr = lookup_block_map(minor, guestAddr);
			bitmapSet(freeBlockList, hostAddr);
			if(cp != NULL) {
				lst_get_rec(cp->dataFile, guestAddr, &hostAddr);
			} else {
				hostAddr = 0;
			}
			set_block_map(bMap, guestAddr, hostAddr);
		}
	}

	dprintk("done freeVisitedBlocks\n");
}
*/

static void deleteCheckpoint(struct cow_disk_cp *cp, int minor) {
	struct cow_disk_cp *prev, *next;

	prev = NULL;
	next = NULL;
	
	assert(cp->refCount > 0);
	cp->refCount--;
	if(cp->refCount > 0) 
		return;

	if(cp->link.next != &dbraw_devices[minor].cp_list) {
		next = list_entry(cp->link.next,
				  struct cow_disk_cp,
				  link);
	}

	if(cp->link.prev != &dbraw_devices[minor].cp_list) {
		prev = list_entry(cp->link.prev,
				  struct cow_disk_cp,
				  link);
	}

	if(next) {
		merge_undo_redo(next->redo_bm,
				next->dataFile,
				cp->redo_bm,
				cp->dataFile);
	}

	if(prev) {
		merge_undo_redo(prev->undo_bm,
				prev->dataFile,
				cp->undo_bm,
				cp->dataFile);
	}
	
	//freeBlocksAfterDel(cp, prev, next, 
	//	   dbraw_devices[minor].freeBlockBitmap);

	// remove from list and free mem
	list_del(&cp->link);
	free_cow_disk_cp(cp);


}

static unsigned long checkpointCb(int action, unsigned long data, 
				  int minor) {

	unsigned long ret = 0;

	down(&dbraw_devices[minor].cleaning_mutex);	

	if(action == COW_DISK_TAKE_CP) {
		ret = takeCheckpoint(minor);
	} else if(action == COW_DISK_RESTORE_CP) {
		restoreCheckpoint((struct cow_disk_cp *) data, minor);
	} else if(action == COW_DISK_DELETE_CP) {
		deleteCheckpoint((struct cow_disk_cp *) data, minor);
	} else {
		up(&dbraw_devices[minor].cleaning_mutex);	
		assert(0);
		return (unsigned long) -1;
	}

	up(&dbraw_devices[minor].cleaning_mutex);	

	return ret;
}
