#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <assert.h>
#include <fcntl.h>
#include <errno.h>

#include "../include/linux/raw.h"
#include "dbraw.h"

#define BUF_SIZE (1024 * 1024)

static void usage(void) {
  printf("usage: [-g] [-s] [-i num_guest_blocks] [-f] [-b] [-r] [-j major] [-n minor]\n");
  printf("\t-h     print help\n");
  printf("\t-g     get raw mapping\n");
  printf("\t-s     set raw mapping\n");
  printf("\t-i     initialize block map\n");
  printf("\t-f     flush block map to disk\n");
  printf("\t-b     get the disk size information\n");
  printf("\t-r     reclaim free blocks\n");
}

static int handleGetSet(int get, int set, int major, int minor) {
  int fd;
  struct raw_config_request rq;
  int ret;

  fd = open("/dev/raw/dbraw0",O_RDONLY);
  if(fd < 0) {
    perror("could not open dbraw0");
    return 1;
  }

  // assume the same major minor numbers
  rq.raw_minor = minor;

  if(set) {
    rq.block_major = major;
    rq.block_minor = minor;

    ret = ioctl(fd, RAW_SETBIND, &rq);
    assert(ret == 0);

    printf("binding dbraw device %d to major %d minor %d\n",
    		minor,major, minor);
  } 

  if(get) {
    ret = ioctl(fd, RAW_GETBIND, &rq);
    assert(ret == 0);

    printf("dbraw device %d block major %llu block minor %llu\n",
	   rq.raw_minor, rq.block_major, rq.block_minor);
  }

  close(fd);

  return 0;
}

static unsigned long handleSize(int getSize, int minor, int print) {
  struct dbraw_size_struct size;
  int fd;
  int ret;
  char fileName[128];
  unsigned long long virt, phys, free;

  sprintf(fileName,"/dev/raw/dbraw%d",minor);
  fd = open(fileName,O_RDONLY);
  if(fd < 0) {
    perror(fileName);
    return 0;
  }

  if(getSize) {
    ret = ioctl(fd, DBRAW_GET_SIZES, &size);
    assert(ret == 0);

    if(print) {
      virt = size.numVirt;
      virt *= size.blockSize;
      phys = size.numPhys;
      phys *= size.blockSize;
      free = size.numFree;
      free *= size.blockSize;
      
      printf("all sizes are in bytes\n");
      printf("blockSize = %lu\nvirtual = %llu\nphysical = %llu\nfree = %llu\n",
	     size.blockSize, virt, phys, free);
    }
  } else {
    assert(0);
  }

  close(fd);

  return size.blockSize;
}
static void handleInitFlushChecksum(int init, int flush, int checksum, int minor,
				    unsigned long numGuestBlocks) {
  int fd;
  struct dbraw_cmd_struct cmd;
  unsigned long buf;
  int ret;
  char fileName[128];
  unsigned long blockSize;

  sprintf(fileName,"/dev/raw/dbraw%d",minor);
  fd = open(fileName,O_RDONLY);
  if(fd < 0) {
    perror(fileName);
    return;
  }

  // read in the block size first
  blockSize = 0;
  ret = ioctl(fd, DBRAW_GET_BLOCK_SIZE, &blockSize);
  assert(ret == 0);
  assert(blockSize > 0);

  cmd.bufSize = BUF_SIZE;
  buf = (unsigned long) malloc(BUF_SIZE + blockSize);
  assert(buf != 0);
  buf += blockSize;
  buf &= ~(blockSize-1);
  cmd.buf = (char *) buf;
  cmd.hostAddr = checksum; //this is the host addr

  cmd.guestNumBlocks = numGuestBlocks;

  if(init) {
    assert(!flush);
    ret = ioctl(fd, DBRAW_INIT_MAP, &cmd);
  } else if(flush) {
    assert(!init);
    ret = ioctl(fd, DBRAW_FLUSH_MAP, &cmd);
  } else if(checksum) {
    ret = ioctl(fd, DBRAW_CHECKSUM_HOST_BLOCK, &cmd);
  }
  assert(ret == 0);

  // dont worry about memory leak

  close(fd);
}

static void handleReclaim(int minor) {
  int fd;
  int ret;
  char fileName[128];

  sprintf(fileName,"/dev/raw/dbraw%d",minor);
  fd = open(fileName,O_RDONLY);
  if(fd < 0) {
    perror(fileName);
    return;
  }

  ret = ioctl(fd, DBRAW_RECLAIM_FREE_BLOCKS, NULL);
  assert(ret == 0);


  close(fd);
}

int main(int argc, char *argv[]) {
  int opt;
  int get = 0;
  int set = 0;
  int init = 0;
  int flush = 0;
  int getSize = 0;
  unsigned long checksum = 0;
  int reclaim = 0;
  int major = -1;
  int minor = -1;
  unsigned long numGuestBlocks = 0;

  if(argc == 1) {
    usage();
    return 1;
  }

  while((opt = getopt(argc,argv,"ghsi:fc:brj:n:")) != -1) {
    if(opt == 'g') {
      get = 1;
    } else if(opt == 'h') {
      usage();
    } else if(opt == 's') {
      set = 1;
    } else if(opt == 'i') {
      if(sscanf(optarg, "%lu", &numGuestBlocks) != 1) {
	printf("could not parse the numGuestBlocks: %s\n",optarg);
	return 1;
      }
      init = 1;
    } else if(opt == 'f') {
      flush = 1;
    } else if(opt == 'c') {
      if(sscanf(optarg, "%lu", &checksum) != 1) {
	printf("could not parse checksum host address\n");
      }
    } else if(opt == 'b') {
      getSize = 1;
    } else if(opt == 'r') {
      reclaim = 1;
    } else if(opt == 'j') {
      if(sscanf(optarg, "%d", &major) != 1) {
	printf("coult not parse major: %s\n",optarg);
	return 1;
      }
    } else if(opt == 'n') {
      if(sscanf(optarg, "%d", &minor) != 1) {
	printf("coult not parse minor: %s\n",optarg);
	return 1;
      }
    }

  }

  if(get || set) {
    if((major == -1) || (minor == -1)) {
      printf("must specifiy both major and minor numbers for get and set cmds\n");
      return 1;
    }

    if(handleGetSet(get, set, major, minor) != 0) {
      return 1;
    }
  }

  if(init || flush || checksum) {
    if(minor == -1) {
      printf("must specify minor device number\n");
      return 1;
    }
    if(init && (numGuestBlocks == 0)) {
      printf("must set num_guest_blocks");
      return 1;
    }
    handleInitFlushChecksum(init, flush, checksum, minor, numGuestBlocks);
  }

  if(getSize) {
    if(minor == -1) {
      printf("must specify minor device number\n");
      return 1;
    }
    handleSize(getSize, minor, 1);
  }

  if(reclaim) {
    if(minor == -1) {
      printf("must specify minor device number\n");
      return 1;
    }
    handleReclaim(minor);
  }

  return 0;
}
