/****************************************************************************

		THIS SOFTWARE IS NOT COPYRIGHTED

   HP offers the following for use in the public domain.  HP makes no
   warranty with regard to the software or it's performance and the
   user accepts the software "AS IS" with all faults.

   HP DISCLAIMS ANY WARRANTIES, EXPRESS OR IMPLIED, WITH REGARD
   TO THIS SOFTWARE INCLUDING BUT NOT LIMITED TO THE WARRANTIES
   OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

****************************************************************************/

/****************************************************************************
 *  Header: remcom.c,v 1.34 91/03/09 12:29:49 glenne Exp $
 *
 *  Module name: remcom.c $
 *  Revision: 1.34 $
 *  Date: 91/03/09 12:29:49 $
 *  Contributor:     Lake Stevens Instrument Division$
 *
 *  Description:     low level support for gdb debugger. $
 *
 *  Considerations:  only works on target hardware $
 *
 *  Written by:      Glenn Engel $
 *  ModuleState:     Experimental $
 *
 *  NOTES:           See Below $
 *
 *  Modified for 386 by Jim Kingdon, Cygnus Support.
 *
 *  To enable debugger support, two things need to happen.  One, a
 *  call to set_debug_traps() is necessary in order to allow any breakpoints
 *  or error conditions to be properly intercepted and reported to gdb.
 *  Two, a breakpoint needs to be generated to begin communication.  This
 *  is most easily accomplished by a call to breakpoint().  Breakpoint()
 *  simulates a breakpoint by executing a trap #1.
 *
 *  The external function exceptionHandler() is
 *  used to attach a specific handler to a specific 386 vector number.
 *  It should use the same privilege level it runs at.  It should
 *  install it as an interrupt gate so that interrupts are masked
 *  while the handler runs.
 *
 *  Because gdb will sometimes write to the stack area to execute function
 *  calls, this program cannot rely on using the supervisor stack so it
 *  uses it's own stack area reserved in the int array remcomStack.
 *
 *************
 *
 *    The following gdb commands are supported:
 *
 * command          function                               Return value
 *
 *    g             return the value of the CPU registers  hex data or ENN
 *    G             set the value of the CPU registers     OK or ENN
 *
 *    mAA..AA,LLLL  Read LLLL bytes at address AA..AA      hex data or ENN
 *    MAA..AA,LLLL: Write LLLL bytes at address AA.AA      OK or ENN
 *
 *    c             Resume at current address              SNN   ( signal NN)
 *    cAA..AA       Continue at address AA..AA             SNN
 *
 *    s             Step one instruction                   SNN
 *    sAA..AA       Step one instruction from AA..AA       SNN
 *
 *    k             kill
 *
 *    ?             What was the last sigval ?             SNN   (signal NN)
 *
 * All commands and responses are sent with a packet which includes a
 * checksum.  A packet consists of
 *
 * $<packet info>#<checksum>.
 *
 * where
 * <packet info> :: <characters representing the command or response>
 * <checksum>    :: < two hex digits computed as modulo 256 sum of <packetinfo>>
 *
 * When a packet is received, it is first acknowledged with either '+' or '-'.
 * '+' indicates a successful transfer.  '-' indicates a failed transfer.
 *
 * Example:
 *
 * Host:                  Reply:
 * $m0,10#2a               +$00010203040506070809101112131415#42
 *
 ****************************************************************************/


/************************************************************************
 *
 * external low-level support routines
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/unistd.h>
#include <linux/string.h>
#include <linux/types.h>
#include <linux/param.h>
#include <linux/fs.h>

#include <asm/uaccess.h>

#include <asm/signal.h>
#include <linux/mman.h>

#include <asm/system.h>
#include <asm/segment.h>

#include <asm/semaphore.h>

#include <linux/mm.h>

#include <revirt/checkpoint.h>

#include "myutil.h"
#include "uml.h"

extern void putDebugChar();	/* write a single character      */
extern int getDebugChar();	/* read and return a single char */
extern void exceptionHandler();	/* assign an exception handler   */


/************************************************************************/
/* BUFMAX defines the maximum number of characters in inbound/outbound buffers*/
/* at least NUMREGBYTES*2 are needed for register packets */
#define BUFMAX 400


#define CHECKPOINT_CMD   1
#define ROLLBACK_CMD     2

//static char initialized;  /* boolean flag. != 0 means we've been initialized */

int     remote_debug;
/*  debug >  0 prints ill-formed commands in valid packets & checksum errors */

static const char hexchars[]="0123456789abcdef";

/* Number of registers.  */
#define NUMREGS	16

/* Number of bytes of registers.  */
#define NUMREGBYTES (NUMREGS * 4)

enum regnames {GDB_EAX, GDB_ECX, GDB_EDX, GDB_EBX, 
	       GDB_ESP, GDB_EBP, GDB_ESI, GDB_EDI,
	       GDB_PC /* also known as eip */,
	       GDB_PS /* also known as eflags */,
	       GDB_CS, GDB_SS, GDB_DS, GDB_ES, GDB_CFS, GDB_GS};


/*
 * these should not be static cuz they can be used outside this module
 */
int registers[NUMREGS];


/* Put the error code here just in case the user cares.  */
//int gdb_i386errcode;
/* Likewise, the vector number here (since GDB only gets the signal
   number through the usual means, and that's not very specific).  */
//int gdb_i386vector = -1;

/* GDB stores segment registers in 32-bit words (that's just the way
   m-i386v.h is written).  So zero the appropriate areas in registers.  */

#if 0
#define SAVE_REGISTERS1() \
  asm ("movl %eax, _registers");                                   	  \
  asm ("movl %ecx, _registers+4");			  		     \
  asm ("movl %edx, _registers+8");			  		     \
  asm ("movl %ebx, _registers+12");			  		     \
  asm ("movl %ebp, _registers+20");			  		     \
  asm ("movl %esi, _registers+24");			  		     \
  asm ("movl %edi, _registers+28");			  		     \
  asm ("movw $0, %ax");							     \
  asm ("movw %ds, _registers+48");			  		     \
  asm ("movw %ax, _registers+50");					     \
  asm ("movw %es, _registers+52");			  		     \
  asm ("movw %ax, _registers+54");					     \
  asm ("movw %fs, _registers+56");			  		     \
  asm ("movw %ax, _registers+58");					     \
  asm ("movw %gs, _registers+60");			  		     \
  asm ("movw %ax, _registers+62");
#define SAVE_ERRCODE() \
  asm ("popl %ebx");                                  \
  asm ("movl %ebx, _gdb_i386errcode");
#define SAVE_REGISTERS2() \
  asm ("popl %ebx"); /* old eip */			  		     \
  asm ("movl %ebx, _registers+32");			  		     \
  asm ("popl %ebx");	 /* old cs */			  		     \
  asm ("movl %ebx, _registers+40");			  		     \
  asm ("movw %ax, _registers+42");                                           \
  asm ("popl %ebx");	 /* old eflags */		  		     \
  asm ("movl %ebx, _registers+36");			 		     \
  /* Now that we've done the pops, we can save the stack pointer.");  */   \
  asm ("movw %ss, _registers+44");					     \
  asm ("movw %ax, _registers+46");     	       	       	       	       	     \
  asm ("movl %esp, _registers+16");
#endif

static int hex (char ch)
{
  if ((ch >= 'a') && (ch <= 'f'))
    return (ch - 'a' + 10);
  if ((ch >= '0') && (ch <= '9'))
    return (ch - '0');
  if ((ch >= 'A') && (ch <= 'F'))
    return (ch - 'A' + 10);
  return (-1);
}

static char remcomInBuffer[BUFMAX];
static char remcomOutBuffer[BUFMAX];

/* scan for the sequence $<data>#<checksum>     */

// for hardware watchpoints, maybe revisit this later
#define loaddebug(thread,register) \
		__asm__("movl %0,%%db" #register  \
			: /* no output */ \
			:"r" ((thread)->debugreg[register]))

#define DBG_L(reg) (0x1 << ((reg) * 2))
//loaddebug(&current->thread, 7);

static void setBreakAddr(unsigned long dbgNo, unsigned long addr) {
  assert((dbgNo >= 1) && (dbgNo <= 3));
  current->thread.debugreg[dbgNo] = addr;  
  current->thread.debugreg[7] |= DBG_L(dbgNo);
  loaddebug(&current->thread, 7);
  loaddebug(&current->thread, 1);
  loaddebug(&current->thread, 2);
  loaddebug(&current->thread, 3);

  printk("setting breakpoint for dbreg %lu at 0x%08lx\n",dbgNo, addr);
}

static void clearBreakAddr(unsigned long dbgNo) {
  assert((dbgNo >= 1) && (dbgNo <= 3));
  current->thread.debugreg[dbgNo] = 0;
  current->thread.debugreg[7] &= ~DBG_L(dbgNo);
  loaddebug(&current->thread, 7);
  loaddebug(&current->thread, 1);
  loaddebug(&current->thread, 2);
  loaddebug(&current->thread, 3);

  printk("clearing breakpoint for dbreg %lu\n",dbgNo);
}




static unsigned char * getpacket (void)
{
  unsigned char *buffer = &remcomInBuffer[0];
  unsigned char checksum;
  unsigned char xmitcsum;
  int count;
  int ret;
  char ch;

  while (1)
    {
      /* wait around for the start character, ignore all other characters */
      while ((ret = getDebugChar ()) != '$') {
	ch = ret;
	if(ret == -1) {
	  return buffer;
	}
      }


    retry:
      checksum = 0;
      xmitcsum = -1;
      count = 0;

      /* now, read until a # or end of buffer is found */
      while (count < BUFMAX) {
	ret = getDebugChar ();
	ch = ret;
	if(ret == -1) {
	  return NULL;
	}

	if (ch == '$')
	  goto retry;
	if (ch == '#')
	  break;
	checksum = checksum + ch;
	buffer[count] = ch;
	count = count + 1;
      }

      buffer[count] = 0;

      if (ch == '#')
	{
	  ch = getDebugChar (); if(ch == -1) return NULL;
	  xmitcsum = hex (ch) << 4;
	  ch = getDebugChar (); if(ch == -1) return NULL;
	  xmitcsum += hex (ch);

	  if (checksum != xmitcsum)
	    {
	      if (remote_debug)
		{
		  printk("bad checksum.  My count = 0x%x, sent=0x%x. buf=%s\n",
			 checksum, xmitcsum, buffer);
		}
	      putDebugChar ('-');	/* failed checksum */
	    }
	  else
	    {
	      putDebugChar ('+');	/* successful transfer */

	      /* if a sequence char is present, reply the sequence ID */
	      if (buffer[2] == ':')
		{
		  putDebugChar (buffer[0]);
		  putDebugChar (buffer[1]);

		  return &buffer[3];
		}

	      return &buffer[0];
	    }
	}
    }
}

/* send the packet in buffer.  */

static void putpacket (unsigned char *buffer)
{
  unsigned char checksum;
  int count;
  char ch;

  /*  $<packet info>#<checksum>. */
  do {
    putDebugChar ('$');
    checksum = 0;
    count = 0;
    
    while ((ch = buffer[count]) != 0) {
      putDebugChar (ch);
      checksum += ch;
      count += 1;
    }
    
    putDebugChar ('#');
    putDebugChar (hexchars[checksum >> 4]);
    putDebugChar (hexchars[checksum % 16]);
   
    ch = getDebugChar(); if(ch == -1) return;
  } while(ch != '+');
}


/* convert the memory pointed to by mem into hex, placing result in buf */
/* return a pointer to the last char put in buf (null) */
/* If MAY_FAULT is non-zero, then we should set mem_err in response to
   a fault; if zero treat a fault like any other fault in the stub.  */
static char *mem2hex (char *mem, char *buf, int count, int isUserMode)
{
  int i;
  unsigned char ch;

  for (i = 0; i < count; i++) {
    if(isUserMode) {
      ch = 0;
      copy_from_user(&ch, mem+i, sizeof(char));
    } else {
      ch = mem[i];
    }
    *buf++ = hexchars[ch >> 4];
    *buf++ = hexchars[ch % 16];
  }
  *buf = 0;
  
  return (buf);
}

static char *guestKernelHex2Mem(char *buf, char *mem, int count, 
				int isUserMode, int *err) {
  
  int i;
  unsigned char ch;
  struct gdb_mod_mem *node, *tmp;
  int ret;

  for (i = 0; i < count; i++) {
    ch = hex (*buf++) << 4;
    ch = ch + hex (*buf++);
    if(isUserMode) {
      //printk("writing 0x%02x to 0x%08lx\n",ch, (unsigned long) mem);

      // check if this memory has already been written by gdb
      node = current->uml->gdb_mem_list;
      while(node != NULL) {
	if(node->addr == mem) break;
	node = node->next;
      }

      if(node != NULL) {
	// memory already written make sure gdb is just restoring the
	// original value
	assert(node->origVal == ch);
	// restoring the original value
	//assert(ch == 0xcc);
	ch = node->origVal;
	tmp = current->uml->gdb_mem_list;
	if(tmp == node) {
	  // at the head of the list
	  current->uml->gdb_mem_list = node->next;
	} else {
	  while(tmp->next != NULL) {
	    if(tmp->next == node) {
	      tmp->next = node->next;
	      break;
	    }
	    tmp = tmp->next;
	  }
	}
	util_free_small(node);
      } else {
	// not found must be writing in a break point	
	if(ch == 0xcc) { // break point instruction
	  node = util_alloc_small(sizeof(*node));
	  ret = access_process_vm(current, (unsigned long) mem, &node->origVal,
				  sizeof(node->origVal), 0);
	  if(ret != sizeof(node->origVal)) {
	    util_free_small(node);
	    if(err != NULL) *err = 1;
	  } else {
	    assert(ret == sizeof(node->origVal));
	    node->newVal = ch;
	    node->next = current->uml->gdb_mem_list;
	    node->addr = mem;
	    current->uml->gdb_mem_list = node;
	  }
	}
      }

      //ret = copy_to_user(mem++, &ch, sizeof(char));
      //assert(ret == 0);
      ret = access_process_vm(current, (unsigned long) (mem++),
			      &ch, sizeof(ch), 1);
      if(ret != sizeof(ch)) {
	if(err != NULL) *err = 1;
      }
    } else {
      mem[i] = ch;
    }
  }

  return (mem);
}

#define FIRST_UBP 1
#define LAST_UBP  3
static char *guestUserHex2Mem(char *buf, char *mem, int count, 
				int isUserMode, int *err) {
  int i,idx;
  unsigned char ch;

  for (i = 0; i < count; i++) {
    ch = hex (*buf++) << 4;
    ch = ch + hex (*buf++);
    if(isUserMode) {
      if(ch == 0xcc) {
	// break point

	for(idx = FIRST_UBP; idx <= LAST_UBP; idx++) {
	  if(current->thread.debugreg[idx] == 0) {
	    break;
	  }
	}
	if((idx >= FIRST_UBP) && (idx <= LAST_UBP)) {
	  setBreakAddr(idx, (unsigned long) mem);
	} else {
	  printk("!!!!!!!!!!!!!!!ran out of breakpoints!!!!!!!!!!!!\n");
	  printk("not setting 0x%08lx\n",(unsigned long) mem);
	  for(idx=FIRST_UBP; idx <= LAST_UBP; idx++) {
	    printk("dbgreg[%d] = 0x%08lx\n",idx, current->thread.debugreg[idx]);
	  }
	  printk("\n\n");
	}
      } else {
	// clearing a breakpoint
	for(idx = FIRST_UBP; idx <= LAST_UBP; idx++) {
	  if(current->thread.debugreg[idx] == (unsigned long) mem) {
	    break;
	  }
	}
	if((idx >= FIRST_UBP) && (idx <= LAST_UBP)) {
	  clearBreakAddr(idx);
	} else {	  
	  printk("ignoring write at 0x%08lx\n",(unsigned long) mem);
	  for(idx = FIRST_UBP; idx <= LAST_UBP; idx++) {
	    printk("debugreg[%d] = 0x%08lx\n",idx,current->thread.debugreg[idx]);
	  }
	  printk("debugreg[%d] = 0x%08lx\n",7,current->thread.debugreg[7]);
	}
      }
      mem++;
    }
  }

  return (mem);
}

extern int isGuestKernelDebug(void);

/* convert the hex array pointed to by buf into binary to be placed in mem */
/* return a pointer to the character AFTER the last byte written */
static char *hex2mem (char *buf, char *mem, int count, int isUserMode, int *err) {

  if(isGuestKernelDebug()) {
    return guestKernelHex2Mem(buf, mem, count, isUserMode, err);
  } else {
    return guestUserHex2Mem(buf, mem, count, isUserMode, err);
  }

  return NULL;
}

/**********************************************/
/* WHILE WE FIND NICE HEX CHARS, BUILD AN INT */
/* RETURN NUMBER OF CHARS PROCESSED           */
/**********************************************/
static int hexToInt (char **ptr, int *intValue)
{
  int numChars = 0;
  int hexValue;

  *intValue = 0;

  while (**ptr)
    {
      hexValue = hex (**ptr);
      if (hexValue >= 0)
	{
	  *intValue = (*intValue << 4) | hexValue;
	  numChars++;
	}
      else
	break;

      (*ptr)++;
    }

  return (numChars);
}

/*
static int hexToUnsigned (char *ptr, unsigned long *val)
{
  int numChars = 0;
  int hexValue;
  
  *val = 0;
  
  while (*ptr) {
    hexValue = hex (*ptr);
    if (hexValue >= 0) {
      *val = (*val << 4) | hexValue;
      numChars++;
    }
    else {
      break;
    }
    
    ptr++;
  }
  
  return (numChars);
}
*/

void copyRegsToBuf(struct pt_regs *regs) {
  char *cregs = (char *) registers;
  memset(cregs, 0, sizeof(registers));

  memcpy(cregs, &regs->eax, 4);
  memcpy(cregs + 4, &regs->ecx, 4);
  memcpy(cregs + 8, &regs->edx, 4);
  memcpy(cregs + 12, &regs->ebx, 4);
  memcpy(cregs + 20, &regs->ebp, 4);
  memcpy(cregs + 24, &regs->esi, 4);
  memcpy(cregs + 28, &regs->edi, 4);
  memcpy(cregs + 48, &regs->xds, 2);
  memcpy(cregs + 52, &regs->xes, 2);
  memcpy(cregs + 32, &regs->eip, 4);
  memcpy(cregs + 40, &regs->xcs, 2);
  memcpy(cregs + 36, &regs->eflags, 4);
  memcpy(cregs + 44, &regs->xss, 2);
  memcpy(cregs + 16, &regs->esp, 4);
}

#define revmemcpy(a,b,c) memcpy((b),(a),(c))
void copyBufToRegs(struct pt_regs *regs) {
  char *cregs = (char *) registers;

  revmemcpy(cregs, &regs->eax, 4);
  revmemcpy(cregs + 4, &regs->ecx, 4);
  revmemcpy(cregs + 8, &regs->edx, 4);
  revmemcpy(cregs + 12, &regs->ebx, 4);
  revmemcpy(cregs + 20, &regs->ebp, 4);
  revmemcpy(cregs + 24, &regs->esi, 4);
  revmemcpy(cregs + 28, &regs->edi, 4);
  revmemcpy(cregs + 48, &regs->xds, 2);
  revmemcpy(cregs + 52, &regs->xes, 2);
  revmemcpy(cregs + 32, &regs->eip, 4);
  revmemcpy(cregs + 40, &regs->xcs, 2);
  revmemcpy(cregs + 36, &regs->eflags, 4);
  revmemcpy(cregs + 44, &regs->xss, 2);
  revmemcpy(cregs + 16, &regs->esp, 4);
}


/*
static void parseCustomCommand(struct pt_regs *regs, char *ptr) {
  unsigned long cmd=0;

  hexToUnsigned(ptr, &cmd);

  printk("writing custom command 0x%08lx\n",cmd);

  assert(0);
}
*/

/*
static int clearHwBreak(char *buf) {
  unsigned long hwCmd, hwAddr, hwLen;
  int err = 0;

  hexToInt(&buf, (int *) &hwCmd);
  buf++;
  hexToInt(&buf, (int *) &hwAddr);
  buf++;
  hexToInt(&buf, (int *) &hwLen);

  printk("hardware break cmd = %lu, addr = 0x%08lx, len = %lu\n",
	 hwCmd, hwAddr, hwLen);

  //if(numFreeDbg == 0) return -1;

  switch(hwCmd) {
  case 0:
    // this is simply a software breakpoint
    assert(hwLen == 1);
    hex2mem("cc",(char *) hwAddr, 1, 1, &err);
    if(err) {
      printk("got a memory error from hex2mem\n");
    }
    return (err) ? -1 : 0;
  }

  return -1;
}

static int setHwBreak(char *buf) {
  unsigned long hwCmd, hwAddr, hwLen;
  int err = 0;

  hexToInt(&buf, (int *) &hwCmd);
  buf++;
  hexToInt(&buf, (int *) &hwAddr);
  buf++;
  hexToInt(&buf, (int *) &hwLen);

  printk("hardware break cmd = %lu, addr = 0x%08lx, len = %lu\n",
	 hwCmd, hwAddr, hwLen);

  //if(numFreeDbg == 0) return -1;

  switch(hwCmd) {
  case 0:
    // this is simply a software breakpoint
    assert(hwLen == 1);
    hex2mem("cc",(char *) hwAddr, 1, 1, &err);
    if(err) {
      printk("got a memory error from hex2mem\n");
    }
    return (err) ? -1 : 0;
  }

  return -1;
}
*/

static int gdbStepping = 0;

/*
 * This function does all command procesing for interfacing to gdb.
 */
void gdb_handle_exception(struct pt_regs *regs, int sigval, pid_t threadPid,
			  int gotSIGINT)
{
  int stepping;
  int addr, length;
  char *ptr;
  int newPC;
  extern int debugOn;
  int mem_err;

  copyRegsToBuf(regs);

  ptr = remcomOutBuffer;

  if(gotSIGINT) {
    *ptr++ = 'S';
    *ptr++ = hexchars[SIGINT >> 4];
    *ptr++ = hexchars[SIGINT & 0xf];
  } else {
    *ptr++ = 'T';			/* notify gdb with signo, PC, FP and SP */
    *ptr++ = hexchars[sigval >> 4];
    *ptr++ = hexchars[sigval & 0xf];
    
    *ptr++ = hexchars[GDB_ESP]; 
    *ptr++ = ':';
    ptr = mem2hex((char *)&registers[GDB_ESP], ptr, 4, 0);	/* SP */
    *ptr++ = ';';
    
    *ptr++ = hexchars[GDB_EBP]; 
    *ptr++ = ':';
    ptr = mem2hex((char *)&registers[GDB_EBP], ptr, 4, 0); 	/* FP */
    *ptr++ = ';';
    
    *ptr++ = hexchars[GDB_PC]; 
    *ptr++ = ':';
    ptr = mem2hex((char *)&registers[GDB_PC], ptr, 4, 0); 	/* PC */
    *ptr++ = ';';
    
    if(threadPid) {
      printk("threadPid = %d\n",threadPid);
      *ptr++ = 't';
      *ptr++ = 'h';
      *ptr++ = 'r';
      *ptr++ = 'e';
      *ptr++ = 'a';
      *ptr++ = 'd';
      *ptr++ = ':';
      *ptr++ = hexchars[((threadPid & 0xff000000) >> 24) >> 4];
      *ptr++ = hexchars[((threadPid & 0xff000000) >> 24) & 0xf];
      *ptr++ = hexchars[((threadPid & 0x00ff0000) >> 16) >> 4];
      *ptr++ = hexchars[((threadPid & 0x00ff0000) >> 16) & 0xf];
      *ptr++ = hexchars[((threadPid & 0x0000ff00) >> 8) >> 4];
      *ptr++ = hexchars[((threadPid & 0x0000ff00) >> 8) & 0xf];
      *ptr++ = hexchars[((threadPid & 0x000000ff)) >> 4];
      *ptr++ = hexchars[((threadPid & 0x000000ff)) & 0xf];
      //ptr = mem2hex((char *) &threadPid, ptr, 4, 0);
      *ptr++ = ';';
    }
  }

  *ptr = '\0';

  putpacket (remcomOutBuffer);

  stepping = 0;

  while (debugOn)
    {
      remcomOutBuffer[0] = 0;
      ptr = getpacket ();
      
      if(ptr == NULL) return;

      //printk("cmdstr: %s\n",ptr);

      switch (*ptr++)
	{
	case '?':
	  remcomOutBuffer[0] = 'S';
	  remcomOutBuffer[1] = hexchars[sigval >> 4];
	  remcomOutBuffer[2] = hexchars[sigval % 16];
	  remcomOutBuffer[3] = 0;
	  break;
	case 'd':
	  remote_debug = !(remote_debug);	/* toggle debug flag */
	  break;
	case 'g':		/* return the value of the CPU registers */
	  mem2hex ((char *) registers, remcomOutBuffer, NUMREGBYTES, 0);
	  break;
	case 'G':		/* set the value of the CPU registers - return OK */
	  hex2mem (ptr, (char *) registers, NUMREGBYTES, 0, NULL);
	  strcpy (remcomOutBuffer, "OK");
	  copyBufToRegs(regs);
	  break;
	case 'P':		/* set the value of a single CPU register - return OK */
	  {
	    int regno;

	    if (hexToInt (&ptr, &regno) && *ptr++ == '=') {
	      if (regno >= 0 && regno < NUMREGS) {
		//printk("setting reg %d from 0x%08lx to %s\n",
		//     regno, (unsigned long) registers[regno],ptr);
		hex2mem (ptr, (char *) &registers[regno], 4, 0, NULL);
		strcpy (remcomOutBuffer, "OK");
		copyBufToRegs(regs);
		break;
	      } else if(regno == 0x29) {
		// ignore this for now
		strcpy(remcomOutBuffer, "OK");
		break;
	      }
	    }

	    strcpy (remcomOutBuffer, "E01");
	    break;
	  }

	  /* mAA..AA,LLLL  Read LLLL bytes at address AA..AA */
	case 'm':
	  /* TRY TO READ %x,%x.  IF SUCCEED, SET PTR = 0 */
	  if (hexToInt (&ptr, &addr))
	    if (*(ptr++) == ',')
	      if (hexToInt (&ptr, &length))
		{
		  ptr = 0;
		  //mem_err = 0;
		  mem2hex ((char *) addr, remcomOutBuffer, length, 1);
		  //XXX FIXME!
		  //if (mem_err)
		  //{
		  //  strcpy (remcomOutBuffer, "E03");
		  //  debug_error ("memory fault");
		  //}
		}

	  if (ptr)
	    {
	      printk("E01:2\n");
	      strcpy (remcomOutBuffer, "E01");
	    }
	  break;

	  /* MAA..AA,LLLL: Write LLLL bytes at address AA.AA return OK */
	case 'M':
	  /* TRY TO READ '%x,%x:'.  IF SUCCEED, SET PTR = 0 */
	  if (hexToInt (&ptr, &addr))
	    if (*(ptr++) == ',')
	      if (hexToInt (&ptr, &length))
		if (*(ptr++) == ':')
		  {

		    //printk("setting 0x%08lx to %s\n",(unsigned long) addr, ptr);
		    //if(((unsigned long) addr) == 0) {
		    //parseCustomCommand(regs, ptr);
		    //} else {
		    mem_err = 0;
		    hex2mem (ptr, (char *) addr, length, 1, &mem_err);
		      //}
		    if (mem_err) {
		      // XXX FIXME just let gdb think it is set
		      //strcpy (remcomOutBuffer, "E03");
		      //printk("umlgdb: user mode memory fault\n");
		      strcpy (remcomOutBuffer, "OK");
		    } else {
		      strcpy (remcomOutBuffer, "OK");
		    }

		    ptr = 0;
		  }
	  if (ptr)
	    {
	      strcpy (remcomOutBuffer, "E02");
	    }
	  break;

	  /* cAA..AA    Continue at address AA..AA(optional) */
	  /* sAA..AA   Step one instruction from AA..AA(optional) */
	case 's':
	  stepping = 1;
	case 'c':
	  /* try to read optional parameter, pc unchanged if no parm */
	  if (hexToInt (&ptr, &addr))
	    registers[GDB_PC] = addr;

	  newPC = registers[GDB_PC];

	  /* set the trace bit if we're stepping */
	  if (stepping) {
	    if(!gdbStepping) {
	      setTrapFlag(current);
	    }
	    gdbStepping = 1;
	  } else {
	    if(gdbStepping) {
	      clearTrapFlag(current);
	    }
	    gdbStepping = 0;
	  }

	  copyBufToRegs(regs);
	  return;
	  break;

	  /* kill the program */
	case 'k':		/* do nothing */
	  debugOn = 0;
	  break;
	  /*
	case 'Z':
	  printk("set hardware break: %s\n",ptr);
	  if(setHwBreak(ptr) == 0) {
	    strcpy(remcomOutBuffer,"OK");
	  } else {
	    printk("could not set hw breakpoint\n");
	  }
	  break;
	case 'z':
	  printk("clear hardware break: %s\n",ptr);
	  if(clearHwBreak(ptr) == 0) {
	    strcpy(remcomOutBuffer,"OK");
	  } else {
	    printk("could not clear hw breakpoint\n");
	  }
	  break;
	case 'H':	  
	  printk("setting thread with %s cmd\n",ptr);
	  if((*ptr == 'c') || (*ptr == 'g')) {
	    ptr++;
	    hexToInt(&ptr,&threadNo);
	    printk("threadNo = %d\n",threadNo);
	    strcpy (remcomOutBuffer, "OK");
	  }
	  break;
	  */
	}			/* switch */

      /* reply to the request */
      if(debugOn) {
	putpacket (remcomOutBuffer);
      }
    }
}


