// this is for the kernel implementation of the file io utilities
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/unistd.h>
#include <linux/string.h>
#include <linux/types.h>
#include <linux/param.h>
#include <linux/fs.h>
#include <linux/slab.h>

#include <asm/uaccess.h>

#include <asm/signal.h>
#include <linux/mman.h>

#include <asm/system.h>
#include <asm/segment.h>

#include <asm/semaphore.h>

#include "myutil.h"

#define BITS_PER_BYTE 8

#define INIT_MEM_SIZE (4096 * 2)

struct bitmapStruct {
  fileUtil_t file;
  unsigned char *memData;
  loff_t fileSize;
  loff_t cacheIdx;
  unsigned char cacheData;
  int cacheDirty;
};

bitmap_t util_new_mem_bitmap(void) {
  bitmap_t bmap = (bitmap_t) util_alloc_small(sizeof(struct bitmapStruct));

  bmap->memData = util_alloc_large(INIT_MEM_SIZE);
  memset(bmap->memData, 0, INIT_MEM_SIZE);
  bmap->fileSize = INIT_MEM_SIZE;
  
  bmap->file = NULL;
  bmap->cacheIdx = -1;
  bmap->cacheData = 0;
  bmap->cacheDirty = 0;

  return bmap;
}

bitmap_t util_new_bitmap(char *fileName) {
  bitmap_t bmap = (bitmap_t) util_alloc_small(sizeof(struct bitmapStruct));
  bmap->file = util_open(fileName, O_LARGEFILE | O_RDWR | O_CREAT | O_TRUNC,
			 0600);
  bmap->fileSize = 0;
  bmap->cacheIdx = -1;
  bmap->cacheData = 0;
  bmap->cacheDirty = 0;
  bmap->memData = NULL;

  return bmap;
}

bitmap_t util_open_bitmap(char *fileName) {
  bitmap_t bmap = (bitmap_t) util_alloc_small(sizeof(struct bitmapStruct));
  bmap->file = util_open(fileName, O_LARGEFILE | O_RDWR, 0600);

  if(bmap->file == NULL) {
    util_free_small(bmap);
    bmap = NULL;
  } else {
    bmap->fileSize = util_seek(bmap->file, 0, UTIL_SEEK_END);
    bmap->cacheIdx = -1;
    bmap->cacheData = 0;
    bmap->cacheDirty = 0;
    bmap->memData = NULL;
  }

  return bmap;
}
  
void util_delete_bitmap(bitmap_t bitmap) {
  if(bitmap->file) {
    util_close(bitmap->file);
  }
  if(bitmap->memData) {
    util_free_large(bitmap->memData);
  }
  
  util_free_small(bitmap);
}

static void util_get_idx_and_bit(loff_t offset, loff_t *idx, unsigned long *bit) {
  if(offset < 0) {
    printk("invalid offset %lld, 0x%016llx\n",offset, offset);
    assert(0);
  }
  assert(offset >= 0);
  *idx = offset / BITS_PER_BYTE;
  *bit = ((unsigned long) offset) % BITS_PER_BYTE;
}

int util_is_bit_set(bitmap_t bitmap, loff_t offset) {
  loff_t idx;
  unsigned long bit;
  unsigned char data;
  int ret;

  util_get_idx_and_bit(offset, &idx, &bit);

  // check if it goes beyond the size of the file
  if(idx >= bitmap->fileSize) return 0;

  if(bitmap->memData != NULL) {
    // using kernel memory
    data = bitmap->memData[idx];
  } else {
    // using host file backup
    if(bitmap->cacheIdx == idx) {
      data = bitmap->cacheData;
    } else {
      // read in the byte that the bit it in
      ret = util_read(bitmap->file, &data, sizeof(data), idx);
      assert(ret == sizeof(data));
    
      if(bitmap->cacheDirty) {
	ret = util_write(bitmap->file, &bitmap->cacheData, 
			 sizeof(bitmap->cacheData), bitmap->cacheIdx);
	assert(ret == sizeof(bitmap->cacheData));
	bitmap->cacheDirty = 0;
      }
      bitmap->cacheData = data;
      bitmap->cacheIdx = idx;
    }
  }


  // check the bit
  return ((0x1 << bit) & data);
}

#define TMP_DATA_SIZE 4096

static void util_set_clear_bit(bitmap_t bitmap, loff_t offset, int set) {
  loff_t idx;
  unsigned long bit;
  unsigned char data;
  int ret;
  loff_t i;
  unsigned char *tmpData;

  util_get_idx_and_bit(offset, &idx, &bit);

  // check if we need to expand our file
  if(idx >= bitmap->fileSize) {
    if(bitmap->memData != NULL) {
      tmpData = util_alloc_large(bitmap->fileSize * 2);
      memset(tmpData, 0, bitmap->fileSize * 2);
      memcpy(tmpData, bitmap->memData, bitmap->fileSize);
      util_free_large(bitmap->memData);
      bitmap->memData = tmpData;
      bitmap->fileSize *= 2;
    } else {
      tmpData = util_alloc_small(TMP_DATA_SIZE);
      memset(tmpData,0,TMP_DATA_SIZE);
      for(i = bitmap->fileSize; i <= idx; i+=TMP_DATA_SIZE) {
	ret = util_write(bitmap->file, tmpData, TMP_DATA_SIZE, i);
	assert(ret == TMP_DATA_SIZE);
	bitmap->fileSize += TMP_DATA_SIZE;
      }
      util_free_small(tmpData);
    }
  }

  if(bitmap->memData != NULL) {
    data = bitmap->memData[idx];
  } else {
    if(bitmap->cacheIdx == idx) {
      data = bitmap->cacheData;
    } else {
      // now read the byte that our bit is in
      ret = util_read(bitmap->file,&data, sizeof(data), idx);
      assert(ret == sizeof(data));
    }
  }

  // adjust the bit and write it back
  if(set) {
    data |= (0x1 << bit);
  } else {
    data &= ~(0x1 << bit);
  }

  if(bitmap->memData != NULL) {
    bitmap->memData[idx] = data;
  } else {
    if(bitmap->cacheIdx == idx) {
      bitmap->cacheData = data;
      bitmap->cacheDirty = 1;
    } else {
      // write through
      ret = util_write(bitmap->file, &data, sizeof(data), idx);
      assert(ret == sizeof(data));
    }
  }
}

void util_set_bit(bitmap_t bitmap, loff_t offset) {
  util_set_clear_bit(bitmap, offset, 1);
}

void util_clear_bit(bitmap_t bitmap, loff_t offset) {
  util_set_clear_bit(bitmap, offset, 0);
}
