#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <assert.h>
#include <sys/ioctl.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <pthread.h>

#include <sys/types.h>
#include <regex.h>

#include <string>
#include <vector>
#include <iostream>

using namespace std;

#include "introspect_gdb.h"
#include "uml.h"

#define CONT_STR "c\n"

#define REVBREAK_STR  "revbreak"
#define REVSTEP_STR   "revstep"
#define LISTCHECK_STR "lscheck"
#define HDWATCH_STR   "hdwatch"
#define GOTO_STR      "goto"
#define CHECKPER_STR  "chperiod"
#define RUNTEST_STR   "runtest"
#define DBGUSR_STR    "dbgusr"

#define UML_DEVICE    "/dev/uml"
#define UMLGDB_DEV    "/dev/umlgdb"

#define DEFAULT_CHECK_PERIOD_SEC 10
#define MS_PER_NS 1000000
#define MS_SLEEP 100
#define SLEEP_PERIOD_PER_SEC 10
#define USEC_PER_SEC 1000000

#define CHECKPOINT_CMD_NONE 0
#define CHECKPOINT_CMD_STOP 1
#define CHECKPOINT_CMD_RESTART 2

#define CHECKPOINT_STATE_RUNNING 0
#define CHECKPOINT_STATE_STOPPED 1

static bool debugPrintOn = false;

static void sigusr1(int signo) {
  cerr << "setting debugPrint to " << debugPrintOn << endl;
  debugPrintOn = !debugPrintOn;
}

static void sigalrm(int signo) {
  // dont do anything this is just to interrupts a semaphore in the
  // kernel uml_read function that may be taking too long
}

introspect_gdb::introspect_gdb(int gdbStdin, int gdbStdout, int gdbStderr) {
  this->gdbStdin = gdbStdin;
  this->gdbStdout = gdbStdout;
  this->gdbStderr = gdbStderr;
  
  umlFd = open(UML_DEVICE,O_RDWR);
  assert(umlFd >= 0);

  rollbackCmdWaiting = 0;
  rollbackCalled = 0;
  checkpointCmd = CHECKPOINT_CMD_NONE;
  signal(SIGALRM, sigalrm);
  signal(SIGUSR1, sigusr1);
  memset(&rbs, 0, sizeof(rbs));  
  checkPeriod = DEFAULT_CHECK_PERIOD_SEC;
  startThread();
}
void introspect_gdb::startThread(void) {
  int ret;

  killThread = 0;
  threadRunning = 1;
  ret = pthread_create(&cpThread, NULL, checkpointThread, this);
  assert(ret == 0);
}

void introspect_gdb::stopThread(void) {
  if(threadRunning) {
    killThread = 1;
    pthread_join(cpThread, NULL);
    threadRunning = 0;
  }
}

introspect_gdb::~introspect_gdb(void) {
  stopThread();
}

void introspect_gdb::printCheck(struct uml_ioc_checkinfo_struct *cis) {
  cout << "checkId:      " << cis->checkId << endl; 
  cout << "guestSyscall: " << cis->guestSyscallCount << endl;
  cout << "scheduleId:   " << cis->revirtId.rhid << endl;
  cout << "branchCount:  " << cis->revirtId.bc << endl;
  cout << "eip:          " << (void *) cis->eip << endl;
  cout << "\n";
}

void introspect_gdb::listCheck(void) {
  unsigned long idx;
  struct uml_ioc_checkinfo_struct currInfo, *checkList;

  getCheckInfo(&currInfo, &checkList);

  sortCheckList(checkList);

  for(idx = 0; idx < MAX_UML_CHECK; idx++) {
    if(checkList[idx].checkId != INVALID_CHECK_ID) {
      printCheck(&(checkList[idx]));
    }
  }
  printCheck(&currInfo);

  delete [] checkList;
}

void introspect_gdb::readFdBuffer(int readFd, string *str) {
  int numBytes;
  char *buf;
  int ret;
  
  str->erase();
  numBytes = 0;
  ret = ioctl(readFd, FIONREAD, &numBytes);
  assert(ret == 0);
  assert(numBytes > 0);
  
  buf = new char[numBytes];
  
  ret = read(readFd, buf, numBytes);
  assert(ret == numBytes);
  str->append(buf, numBytes);
  
  delete [] buf;
}
int introspect_gdb::atOrPast(struct uml_ioc_checkinfo_struct *targetInfo) {
  struct uml_ioc_checkinfo_struct currInfo;

  getCheckInfo(&currInfo);

  if(currInfo.revirtId.rhid > targetInfo->revirtId.rhid) {
    return 1;
  } else if(currInfo.revirtId.rhid < targetInfo->revirtId.rhid) {
    return 0;
  }

  // the rhid's are equal, look at branch counts
  if(currInfo.revirtId.bc > targetInfo->revirtId.bc) {
    return 1;
  } else if(currInfo.revirtId.bc < targetInfo->revirtId.bc) {
    return 0;
  }

  // both the rhid's and bc are equal, check eip

  return currInfo.eip >= targetInfo->eip;

}

void introspect_gdb::clearTfAfterSyscall(void) {
  int fd;

  fd = open(UMLGDB_DEV,O_RDWR);
  assert(fd >= 0);

  ioctl(fd, UMLGDB_IOC_CLEAR_TF_AFTER_SYSCALL,0);
  
  close(fd);
}

void introspect_gdb::goBack(unsigned long checkId, 
			    struct uml_ioc_checkinfo_struct *finalPosInfo,
			    unsigned long long targetNumBreaks, 
			    unsigned long long *numBreaks,
			    int step,
			    string *strOut,
			    unsigned long breakNo) {
  struct uml_ioc_checkinfo_struct currInfo;
  string breakStr;
  char tmp[100];

  sprintf(tmp,"Breakpoint %lu,",breakNo);

  getCheckInfo(&currInfo, NULL);

  // roll back to the checkpoint and trap gdb until we are  at or
  // past the current position
  rollback(checkId, step, finalPosInfo->guestSyscallCount);

  // first break until we get rolled back
  *numBreaks = 0;
  pthread_mutex_lock(&cmdLock);
  do {
    pthread_mutex_unlock(&cmdLock);
    cont();
    waitForBreak(&breakStr);
    pthread_mutex_lock(&cmdLock);
  } while (rollbackCmdWaiting);
  pthread_mutex_unlock(&cmdLock);

  while(!atOrPast(finalPosInfo)) {
    // check to make sure we did not hit the breakpoint we inserted
    if((breakStr.find(tmp) == string::npos) || (breakNo==0)) {
      *numBreaks = *numBreaks + 1;
      if(strOut != NULL) *strOut = breakStr;
    }

    if(*numBreaks == targetNumBreaks) break;

    cont();
    waitForBreak(&breakStr);
  }

  if(step) clearTfAfterSyscall();    

}

void introspect_gdb::sortCheckList(struct uml_ioc_checkinfo_struct *checkList) {
  struct uml_ioc_checkinfo_struct *tmpList = 
    new struct uml_ioc_checkinfo_struct[MAX_UML_CHECK];
  unsigned long idx;
  unsigned long numCheck = 0;
  unsigned long maxIdx, count;
  unsigned long maxCount;

  for(idx = 0; idx < MAX_UML_CHECK; idx++) {
    if(checkList[idx].checkId != INVALID_CHECK_ID) {
      numCheck++;
    }
  }

  maxIdx = 0;
  for(count = 0; count < numCheck; count++) {
    maxCount = 0;
    for(idx = 0; idx < MAX_UML_CHECK; idx++) {
      if(checkList[idx].checkId != INVALID_CHECK_ID) {
	if(checkList[idx].guestSyscallCount > maxCount) {
	  maxCount = checkList[idx].guestSyscallCount;
	  maxIdx = idx;
	}
      }
    }
    assert(maxCount != 0);
    tmpList[count] = checkList[maxIdx];
    checkList[maxIdx].checkId = INVALID_CHECK_ID;
  }

  for(count = 0; count < numCheck; count++) {
    checkList[count] = tmpList[count];
  }


  delete [] tmpList;
}

int introspect_gdb::isInfoEqual(struct uml_ioc_checkinfo_struct *cis0,
				struct uml_ioc_checkinfo_struct *cis1) {

  return (cis0->checkId == cis1->checkId) &&
    (cis0->revirtId.rhid == cis1->revirtId.rhid) &&
    (cis0->revirtId.bc == cis1->revirtId.bc) &&
    (cis0->guestSyscallCount == cis1->guestSyscallCount) &&
    (cis0->eip == cis1->eip);
}

unsigned long introspect_gdb::getCurrEip(void) {
  // set the break at the innermost frame
  int ret;
  string out;
  string cmd = string("down\n");
  unsigned long count = 0;
  do {
    assert(count++ < 100000);
    ret = write(gdbStdin, cmd.c_str(), cmd.size());
    assert((unsigned) ret == cmd.size());
    waitForBreak(&out);
  } while(out.find("Bottom (i.e., innermost) frame selected") == string::npos);

  return readUnsigned("$pc");
}

void introspect_gdb::revBreak(int step, string arg, string *strOut) {
  // find the most recent checkpoint and roll back to it
  struct uml_ioc_checkinfo_struct currInfo, *checkList;
  unsigned long start;
  unsigned long idx;
  unsigned long long numBreaks;
  unsigned long breakNo = 0;
  unsigned long numBreaksBack;
  int ret;

  numBreaksBack = 0;
  ret = sscanf(arg.c_str(),"%lu",&numBreaksBack);
  if(ret != 1) {
    numBreaksBack = 1;
  } else {
    if(numBreaksBack == 0) {
      numBreaksBack = 1;
    }
  }

  cout << "backing up " << numBreaksBack;
  if(step) {
    cout << " instruction(s)" << endl;
  } else {
    cout << " breakpoint(s)" << endl;
  }

  setCpCmd(CHECKPOINT_CMD_STOP);

  // dont worry about setting a break if we are stepping, we will
  // have plenty of chances to find out we are beyond the point
  // of intrest
  if(!step) breakNo = setBreak();

  getCheckInfo(&currInfo, &checkList);
  
  sortCheckList(checkList);

  // find the starting index
  for(start = 0; start < MAX_UML_CHECK; start++) {
    if(checkList[start].guestSyscallCount <= currInfo.guestSyscallCount) {
      break;
    }
  }

  // fixup the real pc, let gdb handle the fact that it just executed
  // a breakpoint inst
  currInfo.eip = getCurrEip();
  for(idx = start; idx < MAX_UML_CHECK; idx++) {    
    if(checkList[idx].checkId != INVALID_CHECK_ID) {
      goBack(checkList[idx].checkId, &currInfo, 0, &numBreaks, step,strOut,breakNo);
      if(numBreaks > 0) {
	numBreaks++;
	if(numBreaksBack >= numBreaks) {
	  numBreaksBack = numBreaks-1;
	  cout << "warning, only going back " << numBreaksBack << endl;
	}
	//cerr << "found " << numBreaks << " going back" << endl;
	goBack(checkList[idx].checkId, &currInfo, numBreaks - numBreaksBack, 
	       &numBreaks, step, strOut,breakNo);
	break;
      } else {
	// if we are stepping back and we did not get a break point, we must
	// want to break before the sys call.  decrement the guest syscallcount
	// and the next checkpoint should work
	//currInfo.guestSyscallCount--;
	//cerr << "did find a break" << endl;
	if(step) {
	  if(currInfo.guestSyscallCount > 
	     (checkList[idx].guestSyscallCount + 2)) {
	    //cerr << "decrement syscall count and try the same check" << endl;
	    currInfo.guestSyscallCount = checkList[idx].guestSyscallCount+1;
	    idx--;
	  }
	}
      }
    } 
  }

  if(!step) deleteBreak(breakNo);

  setCpCmd(CHECKPOINT_CMD_RESTART);

  delete [] checkList;
}

void introspect_gdb::hdWatch(string arg) {
  unsigned long addr;
  int fd;
  struct umlgdb_ioc_watch_struct w;
  int nonWhiteFound = 0;
  unsigned long idx;

  for(idx = 0; idx < arg.size(); idx++) {
    if(!isspace(arg[idx])) {
      nonWhiteFound = 1;
      break;
    }
  }

  fd = open(UMLGDB_DEV,O_RDWR);
  if(fd < 0) {
    cerr << "could not open " << UMLGDB_DEV << endl;
    return;
  }

  if(nonWhiteFound) {
    addr = readUnsigned(arg.c_str());
    if(addr == 0) {
      cout << "could not parse hdwatch argument" << endl;
      close(fd);
      return;
    }
    cout << "setting hardware watchpoint at address " << (void *) addr << endl;
    w.set = 1;
    w.addr = addr;
    w.regNo = 1;
  } else {
    cout << "clearing all hardware watchpoints" << endl;
    w.set = 0;
    w.regNo = 1;
  }
  
  ioctl(fd, UMLGDB_IOC_WATCH, &w);
  close(fd);
}

void introspect_gdb::checkPerCmd(string arg) {
  unsigned long newPeriod;
  int ret;

  ret = sscanf(arg.c_str(),"%lu",&newPeriod);
  if(ret != 1) {
    cout << "invalid period " << arg << endl;
    return;
  }

  cout << "setting new checkpoint period to " << newPeriod << " seconds" << endl;

  checkPeriod = newPeriod;
}

#define NMATCH 128

void introspect_gdb::mysplit(string str, char *delim, vector< string > *vec) {
  unsigned long start, next, len;

  len = str.length();
  start = str.find_first_not_of(delim);
  

  while(start != string::npos) {
    next = str.find_first_of(delim, start);
    if(next != string::npos) {
      vec->push_back(str.substr(start, next-start));
    } else {
      vec->push_back(str.substr(start));
      break;
    }
    start = str.find_first_not_of(delim, next);
  }

}

void introspect_gdb::dbgUsrCmd(string arg) {
  unsigned long idx;
  int ret;
  vector<string> vec;
  struct umlgdb_ioc_debug_guest_user_struct dbs;  

  mysplit(arg, " \t\n", &vec);

  if(vec.size() == 0) {
    cerr << "invalid argument " << arg << endl;
    return;
  }

  dbs.numPid = vec.size();
  dbs.pidArray = new pid_t[vec.size()];
  assert(dbs.pidArray != NULL);
  for(idx = 0; idx < vec.size(); idx++) {
    cout << "adding pid " << vec[idx] << endl;
    ret = sscanf(vec[idx].c_str(),"%dx",&(dbs.pidArray[idx]));
    if(ret != 1) {
      cerr << "malformed pid " << vec[idx] << endl;
      delete [] dbs.pidArray;
      return;
    }
  }

  int fd;

  fd = open(UMLGDB_DEV,O_RDWR);
  assert(fd >= 0);

  ioctl(fd, UMLGDB_IOC_DEBUG_GUEST_USER,&dbs);
  
  close(fd);
  
  delete [] dbs.pidArray;
}

void introspect_gdb::gotoCmd(unsigned long long syscallNo, string *strOut) {
  struct uml_ioc_checkinfo_struct currInfo, *checkList;
  unsigned long checkId = 0, numCheck, idx;

  numCheck = getCheckInfo(&currInfo, &checkList);
  sortCheckList(checkList);

  assert(numCheck > 0);

  // find the checkpoint to roll back to
  for(idx = 0; idx < numCheck; idx++) {
    checkId = checkList[idx].checkId;
    if(checkList[idx].guestSyscallCount < syscallNo) {
      break;
    }
  }


  rollback(checkId, 0, 0);

  delete [] checkList;
 
  pthread_mutex_lock(&cmdLock);
  do {
    pthread_mutex_unlock(&cmdLock);
    cont();
    waitForBreak(strOut);
    getCheckInfo(&currInfo);
    pthread_mutex_lock(&cmdLock);
  } while((currInfo.guestSyscallCount <= syscallNo) || rollbackCmdWaiting);
  pthread_mutex_unlock(&cmdLock);
}

void introspect_gdb::gotoCmd(string arg, string *strOut) {
  unsigned long long syscallNo;
  int ret;

  ret = sscanf(arg.c_str(),"%llu",&syscallNo);
  if(ret != 1) {
    cout << "could not parse goto arg " << arg << endl;
    return;
  }

  setCpCmd(CHECKPOINT_CMD_STOP);

  gotoCmd(syscallNo, strOut);

  setCpCmd(CHECKPOINT_CMD_RESTART);
}

void introspect_gdb::runTest(void) {
  struct uml_ioc_checkinfo_struct currInfo, *checkList;
  long idx;
  struct timeval tvStart, tvStop;
  unsigned long long numBreaks;
  unsigned long tvLastSec;

  setCpCmd(CHECKPOINT_CMD_STOP);

  getCheckInfo(&currInfo, &checkList);
  sortCheckList(checkList);

  tvLastSec = readUnsigned("lastTv.tv_sec");
  printf("%lu\n",tvLastSec);

  for(idx = (MAX_UML_CHECK-1); idx > 0; idx--) {
    if(checkList[idx].checkId != INVALID_CHECK_ID) {
      // restore the checkpoint
      gettimeofday(&tvStart,NULL);
      // let it run foward a little
      currInfo.guestSyscallCount = checkList[idx].guestSyscallCount + 1;

      goBack(checkList[idx].checkId,
	     &currInfo,
	     1,
	     &numBreaks,
	     1,
	     NULL,
	     0);
      gettimeofday(&tvStop, NULL);
      tvLastSec = readUnsigned("lastTv.tv_sec");
      printf("%lu\t",tvLastSec);
      printTimeDiff(&tvStart, &tvStop);

      //get back to our place
      gotoCmd(checkList[0].guestSyscallCount+1, NULL);

      // flush the file cache
      //flushHostFileCache(1);
    }
  }

  delete [] checkList;
  setCpCmd(CHECKPOINT_CMD_RESTART);
}

void introspect_gdb::printTimeDiff(struct timeval *tvStart, 
				   struct timeval *tvStop) {
  long dsec, dusec;

  dsec = tvStop->tv_sec - tvStart->tv_sec;
  dusec = tvStop->tv_usec - tvStart->tv_usec;

  if(dusec < 0) {
    dsec--;
    dusec += USEC_PER_SEC;
  }

  assert(dusec < USEC_PER_SEC);
  assert(dusec >= 0);

  printf("%ld.%06ld\n",dsec, dusec);
}

#define FLUSH_BLOCK_SIZE (1024*1024*10)
#define FLUSH_NUM_ITER 200
#define FLUSH_FILE "/root/bigfile"

void introspect_gdb::flushHostFileCache(unsigned long numIter) {
  int fd;
  char *data;
  int ret;

  for(unsigned long i = 0; i < numIter; i++) {
    fd = open(FLUSH_FILE,O_RDWR | O_CREAT | O_TRUNC,0600);
    unlink(FLUSH_FILE);
    assert(fd >= 0);
    data = new char[FLUSH_BLOCK_SIZE];
    memset(data, 0, FLUSH_BLOCK_SIZE);
    for(unsigned long idx = 0; idx < FLUSH_NUM_ITER; idx++) {
      ret = write(fd, data, FLUSH_BLOCK_SIZE);
      assert(ret == FLUSH_BLOCK_SIZE);
    }
    close(fd);
  }
}

int introspect_gdb::processCmd(string *cmd) {
  int ret = 0;
  string strOut;

  if(cmd->find(REVBREAK_STR) != string::npos) {
    revBreak(0,cmd->substr(cmd->find(REVBREAK_STR)+strlen(REVBREAK_STR)), &strOut);
    ret = 1;
  } else if(cmd->find(REVSTEP_STR) != string::npos) {
    revBreak(1,cmd->substr(cmd->find(REVSTEP_STR)+strlen(REVSTEP_STR)), &strOut);
    ret = 1;
  } else if(cmd->find(LISTCHECK_STR) != string::npos) {
    listCheck();
    ret = 1;
  } else if(cmd->find(HDWATCH_STR) != string::npos) {
    hdWatch(cmd->substr(cmd->find(HDWATCH_STR)+strlen(HDWATCH_STR)));
    ret = 1;
  } else if(cmd->find(GOTO_STR) != string::npos) {
    gotoCmd(cmd->substr(cmd->find(GOTO_STR)+strlen(GOTO_STR)), &strOut);
    ret = 1;
  } else if(cmd->find(CHECKPER_STR) != string::npos) {
    checkPerCmd(cmd->substr(cmd->find(CHECKPER_STR)+strlen(CHECKPER_STR)));
    ret = 1;
  } else if(cmd->find(RUNTEST_STR) != string::npos) {
    runTest();
    ret = 1;
  } else if(cmd->find(DBGUSR_STR) != string::npos) {
    dbgUsrCmd(cmd->substr(cmd->find(DBGUSR_STR) + strlen(DBGUSR_STR)));
    ret = 1;
  }

  if(ret) {
    if(strOut.size() == 0) {
      cout << "(gdb) ";
      cout.flush();
    } else {
      cout << strOut;
      cout.flush();
    }
  }

  return ret;
}

void introspect_gdb::setCpCmd(int cmd) {
  struct timespec ts;

  pthread_mutex_lock(&cmdLock);
  assert(checkpointCmd == CHECKPOINT_CMD_NONE);
  assert(cmd != CHECKPOINT_CMD_NONE);
  checkpointCmd = cmd;

  while(cmd == checkpointCmd) {
    pthread_mutex_unlock(&cmdLock);
    ts.tv_sec = 0;
    ts.tv_nsec = 10 * MS_PER_NS;
    nanosleep(&ts, NULL);
    pthread_mutex_lock(&cmdLock);
  }

  pthread_mutex_unlock(&cmdLock);
}

// this function assumes that the vm is stopped when it is called and
// it will not restart the vm
void introspect_gdb::deleteCheckpoint(introspect_gdb *gdb, unsigned long checkId) {
  struct uml_ioc_rollback_struct rbs;
  int ret;
 
  memset(&rbs, 0, sizeof(rbs));
  rbs.checkId = checkId;
  rbs.restore = 0;
  ret = write(gdb->umlFd, &rbs, sizeof(rbs));
  assert(ret == sizeof(rbs));
}

//long long refCurve[] = {0,1,16,81,256,625,1296,2401,4096,6561};
long long refCurve[] = {0,1,2,4,8,16,32,64,128,256};
//long long refCurve[] = {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19};
//long long refCurve[] = {0,1,2,3,4,5,6,7,8,9};
//long long refCurve[] = {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48,49};

// transform the ref curve so that the current guest syscall count is
// at the 0 position and the last ref curve point is equal to the
// oldest checkpoint (checkList[numCheck-1]).  Also transform the 
// guest syscall counts so that curr is 0 and it increases as it gets older
long long introspect_gdb::abserr(unsigned long checkPos,
				 unsigned long refPos,
				 struct uml_ioc_checkinfo_struct *currInfo,
				 struct uml_ioc_checkinfo_struct *checkList,
				 unsigned long numCheck) {
  long long val = checkList[checkPos].guestSyscallCount;
  long long lastVal = checkList[numCheck-1].guestSyscallCount;
  long long refFactor;
  long long refVal;

  assert((sizeof(refCurve) / sizeof(refCurve[0])) == MAX_UML_CHECK);
  assert((sizeof(refCurve) / sizeof(refCurve[0])) == numCheck);
  assert(checkPos < numCheck);
  assert(refPos < numCheck);

  // first flip the checkpoint value and oldest checkpoint so that the newest
  // checkpoint guest syscall no is 0 and the oldest guest syscall no is positive
  val -= currInfo->guestSyscallCount;
  val *= -1;
  lastVal -= currInfo->guestSyscallCount;
  lastVal *= -1;

  if(lastVal < refCurve[numCheck-1]) {
    cerr << "WARNING: not enough guest syscalls, deleting random checkpoint" << endl;
    return 0;
  }

  // now normalize the reference curve so that the last point is equal to
  // the oldest checkpoint.  This ref value is what will be compared against
  // the checkpoint in question
  refFactor = lastVal / refCurve[numCheck-1];
  refVal = refFactor * refCurve[refPos];

  return (refVal > val) ? refVal - val : val - refVal; 
}

void introspect_gdb::removeOneCheckpoint(introspect_gdb *gdb, 
					 struct uml_ioc_checkinfo_struct *currInfo,
					 struct uml_ioc_checkinfo_struct *checkList,
					 unsigned long numCheck) {
  unsigned long idx, rem, add, pos, minCheckOut;
  long long error, minError;
  
  // find the initial error
  error = 0;
  for(idx = 1; idx < (numCheck-1); idx++) {
    pos = idx;
    add = idx;
    error += gdb->abserr(add, pos, currInfo, checkList, numCheck);
  }
  
  minError = error;
  minCheckOut = 0;
  
  for(idx = 1; idx < (numCheck-1); idx++) {
    rem = idx;
    add = idx-1;
    pos = idx;
    error -= gdb->abserr(rem, pos, currInfo, checkList, numCheck);
    error += gdb->abserr(add, pos, currInfo, checkList, numCheck);
    
    if(error <= minError) {
      minError = error;
      minCheckOut = rem;
    }
  }
  
  gdb->deleteCheckpoint(gdb, checkList[minCheckOut].checkId);
}

void introspect_gdb::takeCheckpoint(introspect_gdb *gdb) {
  int ret;
  unsigned long idx;
  struct uml_ioc_checkpoint_struct cps;  
  unsigned long numCheck;
  struct uml_ioc_checkinfo_struct currInfo, *checkList;

  ret = ioctl(gdb->umlFd, UML_IOC_PAUSE, 1);
  if(ret != 0) {
    // the request timed out and the vm is NOT stopped
    return;
  }

  // if we are here, the pause ioctl stopped the vm and it will
  // not be restarted until we call restart_uml
  numCheck = gdb->getCheckInfo(&currInfo, &checkList);
  gdb->sortCheckList(checkList);

  // first check if we are beyond the last checkpoint, if not do not take
  // a new checkpoint
  if(numCheck > 0) {
    for(idx = 0; idx < numCheck; idx++) {
      if(checkList[idx].guestSyscallCount == currInfo.guestSyscallCount) {
	// we already have a checkpoint here, no need for another
	ioctl(gdb->umlFd, UML_IOC_RESTART_UML, 0);
	delete [] checkList;
	return;
      }
    }
  }

  // now see if we have to remove a checkpoint
  if(numCheck >= MAX_UML_CHECK) {
    assert(numCheck == MAX_UML_CHECK);
    assert(numCheck > 3);
    /*
    cout << "pre" << endl;
    cout << "curr\t" << currInfo.guestSyscallCount << endl;
    for(idx = 0; idx < numCheck; idx++) {
      cout << idx << "\t" << checkList[idx].guestSyscallCount << endl;
    }
    */
    gdb->removeOneCheckpoint(gdb, &currInfo, checkList, numCheck);
  }

  delete [] checkList;  

  // add the new checkpoint
  ret = read(gdb->umlFd, &cps, sizeof(cps));
  assert(ret == sizeof(cps));

  // this can be removed start
  /*
  if(numCheck == MAX_UML_CHECK) {
    cout << "post" << endl;
    gdb->getCheckInfo(&currInfo, &checkList);
    gdb->sortCheckList(checkList);
    for(idx = 0; idx < numCheck; idx++) {
      cout << idx << "\t" << checkList[idx].guestSyscallCount << endl;
    }
    cout << "\n";
    delete [] checkList;
  }
  */
  // this can be removed stop

  ioctl(gdb->umlFd, UML_IOC_RESTART_UML, 0);
}

void *introspect_gdb::checkpointThread(void *arg) {
  introspect_gdb *gdb = (introspect_gdb *) arg;
  int ret;
  unsigned long count;
  int takeCp = 1;
  struct timespec ts;
  struct uml_ioc_rollback_struct rbs;

  gdb->checkpointState = CHECKPOINT_STATE_RUNNING;
  count = gdb->checkPeriod * SLEEP_PERIOD_PER_SEC;

  while(!gdb->killThread) {
    ts.tv_sec = 0;
    ts.tv_nsec = MS_SLEEP * MS_PER_NS; // 100 ms
    nanosleep(&ts,NULL);
    count++;
    
    pthread_mutex_lock(&gdb->cmdLock);
    if(gdb->rollbackCmdWaiting) {
      rbs = gdb->rbs;
      pthread_mutex_unlock(&gdb->cmdLock);
      // this is not perfect, but should be good enough for now
      gdb->rollbackCalled = 1;
      ret = write(gdb->umlFd, &rbs, sizeof(rbs));
      if(ret == -EINTR) return NULL;

      pthread_mutex_lock(&gdb->cmdLock);
      if(ret == sizeof(gdb->rbs)) {
	ioctl(gdb->umlFd, UML_IOC_RESTART_UML, 0);
      }
      gdb->rollbackCmdWaiting = 0;
    }
    if(gdb->checkpointCmd == CHECKPOINT_CMD_STOP) {
      gdb->checkpointCmd = CHECKPOINT_CMD_NONE;
      gdb->checkpointState = CHECKPOINT_STATE_STOPPED;
      takeCp = 0;
    } else if(gdb->checkpointCmd == CHECKPOINT_CMD_RESTART) {
      gdb->checkpointCmd = CHECKPOINT_CMD_NONE;
      gdb->checkpointState = CHECKPOINT_STATE_RUNNING;
      takeCp = 1;
    }
    pthread_mutex_unlock(&gdb->cmdLock);

    // take a checkpoint
    if((count >= (gdb->checkPeriod * SLEEP_PERIOD_PER_SEC)) && 
       (gdb->checkPeriod != 0)) {
      count = 0;
      if(takeCp) {
	gdb->takeCheckpoint(gdb);
      }
    }
  }

  return NULL;
}
unsigned long introspect_gdb::getCheckInfo(struct uml_ioc_checkinfo_struct *currInfo,
				  struct uml_ioc_checkinfo_struct **checkList) {
  struct uml_ioc_checkinfo_struct cis;
  unsigned long idx;
  unsigned long numCheck = 0;

  currInfo->checkId = INVALID_CHECK_ID;
  ioctl(umlFd, UML_IOC_GET_CHECKINFO, currInfo);
  
  if(checkList != NULL) {
    *checkList = new struct uml_ioc_checkinfo_struct[MAX_UML_CHECK];
    assert(*checkList != NULL);
    for(idx = 0; idx < MAX_UML_CHECK; idx++) {
      cis.checkId = idx;
      ioctl(umlFd, UML_IOC_GET_CHECKINFO, &cis);     
      (*checkList)[idx] = cis;
      if(cis.checkId != INVALID_CHECK_ID) {
	numCheck++;
      }
    }
  }
  
  return numCheck;
}

void introspect_gdb::rollback(unsigned long rollbackCheckId, 
			      int step,
			      unsigned long long guestSyscallCount) {
  struct timespec ts;

  pthread_mutex_lock(&cmdLock);
  rbs.checkId = rollbackCheckId; 
  rbs.restore = 1;
  rbs.gdb_step_after_syscall = step;
  rbs.gdb_revbreak_syscall_count = guestSyscallCount;

  rollbackCmdWaiting = 1;
  rollbackCalled = 0;

  pthread_mutex_unlock(&cmdLock);

  do {
    ts.tv_sec = 0;
    ts.tv_nsec = 100 * MS_PER_NS;
    nanosleep(&ts, NULL);
  } while(rollbackCalled == 0);
}

void introspect_gdb::cont(void) {
  int ret = write(gdbStdin,CONT_STR,strlen(CONT_STR));
  assert(ret = strlen(CONT_STR));
}

/**
 * these functions are wrappers around the "set" and "print" gdb
 * commands.  They simply pass the string to gdb and parse the 
 * results.  Var name can be any valid name that would be used
 * with the gdb command line.  This includes casting memory
 * or calling functions.  If you call a function make sure that
 * the arguments are set up properly
 */
unsigned long introspect_gdb::readUnsigned(const char *varName) {
  string buf;
  int ret;
  int ival=0;
  int dummy;

  buf = string("p/x ") + string(varName) + string("\n");
  ret = write(gdbStdin,buf.c_str(),buf.size()); 
  assert(((unsigned) ret) == buf.size());

  // now wait for gdb to return the data
  buf.erase();
  waitForBreak(&buf);

  ret = sscanf(buf.c_str(),"$%d = %x",&dummy, &ival);
  if(ret != 2) {
    printf("warning, malformed read int on %s:%s\n",varName,buf.c_str());
    return 0;
  }

  return ival;
}

#if 0

void introspect_gdb::setUnsigned(char *varName, unsigned long val) {
  char buf[128];

  sprintf(buf,"set %s = 0x%lx\n",varName,val);
  write(gdbStdin[1],buf,strlen(buf));
  waitForBreak(gdbStdout[0]);

}
  

/**
 * used to manipulate the current pointer in the guest.  GDB does not
 * do well with macros, so the functionallity has been reproduced here
 * and the current macro is something used often.
 */
void introspect_gdb::currentSetUnsigned(char *varName, unsigned long val) {
  currentAccess(varName, val, 0);
}  
unsigned long introspect_gdb::currentReadUnsigned(char *varName) {
  return currentAccess(varName,0,1);
}
unsigned long introspect_gdb::currentAccess(char *varName, unsigned long val,
					    int isRead) {
  char buf[128];
  unsigned long sp = readUnsigned("$sp");
  // this is what the current macro does
  sp &= ~0x1fff;
  
  sprintf(buf,"((struct task_struct *) 0x%08lx)->%s",sp,varName);  

  if(isRead) {
    return readUnsigned(buf);
  } else {
    setUnsigned(buf,val);
  }

  return 0;
}

/**
 * these functions are for dealing with the gdb process
 */
#endif

int introspect_gdb::strMatchEnd(string *str,const char *c_str) {
  unsigned idx = str->rfind(string(c_str));

  return idx == (str->size() - strlen(c_str));
}

unsigned long introspect_gdb::setBreak() {
  string cmd;
  int ret;
  string out;
  string::size_type pos;
  unsigned long breakNo;
  char tmp[100];
  unsigned long pc = getCurrEip();


  cmd = string("break ");
  sprintf(tmp,"*0x%08lx\n",pc);
  cmd.append(tmp);

  ret = write(gdbStdin, cmd.c_str(), cmd.size());
  assert((unsigned) ret == cmd.size());

  // now parse the results
  waitForBreak(&out);

  pos = out.find("Breakpoint");
  assert(pos != string::npos);

  out = out.substr(pos);
  ret = sscanf(out.c_str(),"Breakpoint %lu",&breakNo);
  assert(ret == 1);

  return breakNo;
}

void introspect_gdb::deleteBreak(unsigned long breakNo) {
  string cmd;
  char tmp[25];
  int ret;

  sprintf(tmp,"%lu",breakNo);

  cmd.append("delete ");
  cmd.append(tmp);
  cmd.append("\n");

  ret = write(gdbStdin, cmd.c_str(), cmd.size());
  assert((unsigned) ret == cmd.size());

  // just to clear the output buffer
  waitForBreak();
}

void introspect_gdb::waitForBreak(string *resStr) {
  int maxFd;
  fd_set readSet;
  string str;
  string outBuffer;
  int ret;

  do {
    maxFd = (gdbStdout > gdbStderr) ? gdbStdout : gdbStderr;

    maxFd = (STDIN_FILENO > maxFd) ? STDIN_FILENO : maxFd;

    FD_ZERO(&readSet);
    FD_SET(gdbStdout, &readSet);
    FD_SET(gdbStderr, &readSet);
    FD_SET(STDIN_FILENO, &readSet);
    str.erase();
    ret = select(maxFd+1, &readSet, NULL, NULL, NULL);
 
    if(ret > 0) {
      if(FD_ISSET(gdbStderr, &readSet)) {
	readFdBuffer(gdbStderr, &str);
	
	outBuffer += str;
      }
      if(FD_ISSET(gdbStdout, &readSet)) {
	readFdBuffer(gdbStdout, &str);
	outBuffer += str;
      }     

      if((str.size() > 0) && debugPrintOn) {
	cerr << str << endl;
      }

    }
  } while(!strMatchEnd(&outBuffer,"(gdb) "));

  if(resStr != NULL) *resStr = outBuffer;
}
