#ifndef __FAST_INTROSPECTION_H__
#define __FAST_INTROSPECTION_H__

#include <asm/ptrace.h>

typedef void (*breakCallback)(void);

#include <string>
#include <vector>
#include <map>
#include <list>

#include <unistd.h>

#include "guestStructs.h"

using namespace std;

typedef map<string, vector<unsigned long> *> StringMap_t;
typedef map<pid_t, vector<unsigned long> *> PidMap_t;

StringMap_t *copyStringMap(StringMap_t *m);
PidMap_t *copyPidMap(PidMap_t *m);
pid_t getCurrentPid(int fd);
unsigned long getCurrent(struct pt_regs *kernRegs);


typedef enum {
  PLATO_MEM_MODE_USER,
  PLATO_MEM_MODE_KERN
} MemMode_t;

#include "localPlato.h"

class Plato {
 public:
  Plato();
  ~Plato();
  
  unsigned long getReplayState();
  pid_t getVmHostPid();

  unsigned long guestReadMemFunc(unsigned long addr);

  // use this if you want to explicitly set the memory mode
  // for accessing the guest (i.e. read/write kernel or user mode memory)
  std::string guest_strcpy(void *guestAddr, MemMode_t memMode) {
    return guest_strcpy((unsigned long) guestAddr, memMode);
  }
  std::string guest_strcpy(unsigned long guestAddr, MemMode_t memMode);
  unsigned long guest_read_ulong(void *guestAddr, MemMode_t memMode);
  void guest_write_ulong(void *guestAddr, unsigned long value, MemMode_t memMode);
  void guest_memset(void *guestAddr, unsigned char val,
		    unsigned long len, MemMode_t memMode);
  void copy_to_guest(unsigned long guestAddr, void *buf,
		     unsigned long len, MemMode_t memMode);
  void copy_from_guest(void *buf, unsigned long guestAddr, 
		       unsigned long len, MemMode_t memMode);

  // using these functions relies on the breakpoint context.  If you are
  // in a kernel mode breakpoint, it assumes kernel mode.  if you are in
  // a user mode breakpoint, it assumes user mode.
  std::string guest_strcpy(void *guestAddr) {
    return guest_strcpy((unsigned long) guestAddr);
  }
  std::string guest_strcpy(unsigned long guestAddr);
  unsigned long guest_read_ulong(void *guestAddr);
  void guest_write_ulong(void *guestAddr, unsigned long value);
  void guest_memset(void *guestAddr, unsigned char val,
		    unsigned long len);
  void copy_to_guest(unsigned long guestAddr, void *buf,
		     unsigned long len);
  void copy_from_guest(void *buf, unsigned long guestAddr, 
		       unsigned long len);

  void copy_from_guest(void *buf, void  *guestAddr, unsigned long len) {
    copy_from_guest(buf, (unsigned long) guestAddr, len);
  }


  unsigned long registerBreak(char *breakStr, breakCallback breakCb,
			      bool internalBreak = false);
  void enableBreak(unsigned long id);
  void disableBreak(unsigned long id);
  void init(char *configFileName);
  void startEventLoop(breakCallback initCb);
  unsigned long readVar(char *varStr);
  void writeVar(char *varStr, unsigned long val);
  unsigned long callFunction(char *funcName);
  unsigned long callFunction(char *funcName, unsigned long arg0);
  unsigned long callFunction(char *funcName, unsigned long arg0,
			     unsigned long arg1);
  unsigned long callFunction(char *funcName, unsigned long arg0,
			     unsigned long arg1, unsigned long arg2);
  unsigned long callFunction(unsigned long funcAddr, unsigned long arg0);

  unsigned long checkpoint(bool pauseRevirt = true);
  void rollback(unsigned long checkId);
  void deleteCheckpoint(unsigned long checkId);

  unsigned long getCurrent();
  pid_t getCurrentPid();
  unsigned long readStackStructPtr(char *varStr);
  string getCurrentComm();

  unsigned long readFuncArg(unsigned long argNo);
  void writeFuncArg(unsigned long argNo, unsigned long val);

  unsigned long readFuncArgUsingFb(unsigned long argNo);

  void setRegVal(long regno, unsigned long val);
  unsigned long getRegVal(long regno);

  unsigned long callFunctionNoStack(char *funcName, unsigned long arg0);
  void umSchedCb();
  void umSwapCb();
  void umForkCb();
  void umExecCb();
  void umMapCb();
  void umExitCb();
  void slMmapPgoffCb();
  void killCurrProc();
  bool hasUserModeCb;

 private:
  void queueSetBreaks(pid_t pid, unsigned long pageAddr);
  void tryToSetBreaks(pid_t pid, vector<unsigned long> *vec);
  void searchForSharedLibs(pid_t pid, struct linux_task_struct *tsk);
  void setSharedLibBreaks(pid_t pid, struct linux_file *file, unsigned long pgoff,
			  unsigned long startAddr, unsigned long len);
  unsigned long lookupInode(struct linux_file *fPtr);
  unsigned long guestStat(string fileName);
  void delBreakMap(StringMap_t *m);
  void delSetBreakMap(PidMap_t *m);
  bool atFunctionReturn(struct pt_regs *regs, unsigned long numArgs, pid_t pid);
  void umExecForkCb(bool isExec);
  void cleanupBreakMap(pid_t pid, PidMap_t *bmap);
  void queueBreakPoint(pid_t pid, unsigned long breakIdx);
  void insertIntoSetBreakMap(pid_t pid, unsigned long breakIdx);
  bool handleTrap(vector<breakpoint>::iterator iter, struct pt_regs *regs,
		  bool internalOnly);
  void registerUmBreak(unsigned long idx);
  bool isUserModeTrap(void);
  void handleUserModeTrap(struct pt_regs *regs, bool internalOnly = false);
  vector<breakpoint>::iterator handleKernelModeTrap(struct pt_regs *regs, 
						    bool internalOnly = false);
  void set_guest_regs(vector< breakpoint >::iterator iter, 
		      struct pt_regs *regs);
  unsigned long callFunction(char *funcName, vector< unsigned long > *args, 
			     unsigned long funcAddr = 0);
  unsigned long callFunction(char *funcName, vector< unsigned long > *args, 
			     bool useStack, unsigned long funcAddr = 0);

  vector< breakpoint > breakPointVec;
  list< unsigned long > kmBreakPointCache;
  list< unsigned long > umBreakPointCache;
  vector< breakpoint >::iterator currBreak;
  vector< breakpoint >::iterator globalVars;
  vector< breakpoint >::iterator forkBreak;
  vector< breakpoint >::iterator execBreak;
  vector< breakpoint >::iterator mapBreak;
  vector< breakpoint >::iterator exitBreak;
  vector< breakpoint >::iterator schedBreak;
  map<unsigned long, PlatoCheckpoint *> checkMap;
  StringMap_t *umBreakMap;
  StringMap_t *slNameBreakMap;
  PidMap_t *slInodeBreakMap;
  PidMap_t *umSetBreakMap;
  PidMap_t *umQueuedBreakMap;
  bool hasSharedLibBp;
  bool killProc;

  struct pt_regs currRegs;

  int umlFd;
};  


#define FORK_BREAK_ID "fork_handler:1:begin"
#define EXEC_BREAK_ID "exec.c:929:begin"
#define MAP_BREAK_ID  "mem_user.c:37:begin"
#define EXIT_BREAK_ID "do_exit:1:end"
#define MMAP_PGOFF_BREAK_ID "mmap.c:587:begin"
#define SWAP_BREAK_ID "vmscan.c:76:end"
#define SCHED_BREAK_ID "sched.c:695:begin"

#endif
