#include <stdio.h>
#include <stdlib.h>

#include <unistd.h>
#include <iostream>
#include <sys/ioctl.h>
#include <sys/time.h>

#include "plato.h"
#include "../introspect/guestStructs.h"
#include "../../../revirt-skas-2.4.18/uml/uml.h"

#ifndef PAGE_SIZE
#define PAGE_SIZE 4096
#endif

#define COMM_SIZE 16

static unsigned char getvarOpcode[] = "\x55\x89\xe5\x8b\x45\x08\x8b\x00\xc9\xc3";

unsigned long Plato::guestReadMemFunc(unsigned long addr) {
  unsigned long val;
  unsigned long opSize = sizeof(getvarOpcode)-1; // remove null char at end
  unsigned long target = currRegs.esp;
  if((target & (PAGE_SIZE-1)) >= (PAGE_SIZE - opSize)) {
    // XXX FIXME, move it back a little, adjust target
    cerr << "warning, getvarOpcode is going across page boundaries" << endl;
  }

  unsigned long checkId = checkpoint();
  copy_to_guest(target, getvarOpcode, opSize);
  val = callFunction(target, addr);
  rollback(checkId);
  deleteCheckpoint(checkId);

  return val;
}
void Plato::cleanupBreakMap(pid_t pid, PidMap_t *bmap) {
  PidMap_t::iterator iter;
  vector<unsigned long> *vec;

  iter = bmap->find(pid);
  if(iter != bmap->end()) {
    vec = iter->second;
    bmap->erase(iter);
    delete vec;
  }
}

void Plato::umExitCb() {
  vector<breakpoint>::iterator iter;

  pid_t pid = getCurrentPid();

  cleanupBreakMap(pid, umSetBreakMap);
  cleanupBreakMap(pid, umQueuedBreakMap);
  for(iter = breakPointVec.begin(); iter != breakPointVec.end(); iter++) {
    if(iter->getBreakType() == SHARED_LIB_BP) {
      iter->clearOffset(pid);
    }
  }
}

void Plato::umMapCb() {
  PidMap_t::iterator iter;
  pid_t pid = getCurrentPid();

  vector<unsigned long> *vec, *newVec;
  iter = umQueuedBreakMap->find(pid);
  if(iter != umQueuedBreakMap->end()) {
    newVec = new vector<unsigned long>;
    vec = iter->second;
    for(unsigned long idx = 0; idx < vec->size(); idx++) {
      try {
	breakPointVec.at(vec->at(idx)).setBreak();
	insertIntoSetBreakMap(pid, vec->at(idx));
      } catch (...) {
	newVec->insert(newVec->end(), vec->at(idx));
      }
    }
    umQueuedBreakMap->erase(iter);
    delete vec;
    
    if(newVec->size() > 0) {
      (*umQueuedBreakMap)[pid] = newVec;
    } else {
      delete newVec;
    }
  }
  
  if(umQueuedBreakMap->size() == 0) {
    mapBreak->disable(true);
    schedBreak->disable(true);
  }
}

void Plato::umExecCb() {
  umExecForkCb(true);
}

void Plato::umForkCb() {
  umExecForkCb(false);
}

void Plato::tryToSetBreaks(pid_t pid, vector<unsigned long> *vec) {
  for(unsigned long idx = 0; idx < vec->size(); idx++) {
    try {
      breakPointVec.at(vec->at(idx)).setBreak();
      insertIntoSetBreakMap(pid, vec->at(idx));
    } catch (...) {
      // could not set breakpoint, queue it up and set it later
      queueBreakPoint(pid, vec->at(idx));
    }
  }
}

void Plato::searchForSharedLibs(pid_t pid, struct linux_task_struct *tsk) {
  struct linux_mm_struct *mm;
  struct linux_vm_area_struct *vma;
  struct linux_file *file;

  copy_from_guest(&mm, &tsk->mm, sizeof(mm));
  copy_from_guest(&vma, &mm->mmap, sizeof(vma));
  while(vma != NULL) {
    copy_from_guest(&file, &vma->vm_file, sizeof(file));
    if(file != NULL) {
      setSharedLibBreaks(pid, file, guest_read_ulong(&vma->vm_pgoff),
			 guest_read_ulong(&vma->vm_start),
			 guest_read_ulong(&vma->vm_end) - 
			 guest_read_ulong(&vma->vm_start));
    }
    copy_from_guest(&vma, &vma->vm_next, sizeof(vma));
  }
}
    

void Plato::umExecForkCb(bool isExec) {
  string comm;
  char commBuf[COMM_SIZE+1];
  struct linux_task_struct *tsk;
  pid_t pid;
  tsk = (struct linux_task_struct *) this->getCurrent();
  
  memset(commBuf,0, sizeof(commBuf));
  this->copy_from_guest(commBuf, &tsk->comm, COMM_SIZE);
  comm = string(commBuf);
  this->copy_from_guest(&pid, &tsk->pid, sizeof(pid));

  // try to set the breakpoints for this process
  vector<unsigned long> *vec;
  StringMap_t::iterator iter = umBreakMap->find(comm);
  if(iter != umBreakMap->end()) {
    vec = iter->second;
    tryToSetBreaks(pid, vec);
  } else if(isExec) {
    // exec on a forked process that we used to track will clear breakpoints
    umExitCb();
  }
  
  // try to setup any shrared libs
  if(!isExec && hasSharedLibBp) {
    searchForSharedLibs(pid, tsk);      
  }
}

void Plato::queueBreakPoint(pid_t pid, unsigned long breakIdx) {
  if(umQueuedBreakMap->size() == 0) {
    mapBreak->enable();
    schedBreak->enable();
  }

  PidMap_t::iterator iter = umQueuedBreakMap->find(pid);
  vector<unsigned long> *vec;
  
  if(iter == umQueuedBreakMap->end()) {
    vec = new vector<unsigned long>;
    (*umQueuedBreakMap)[pid] = vec;
    mapBreak->enable();
  } else {
    vec = iter->second;
  }

  vec->insert(vec->end(), breakIdx);
}

void Plato::insertIntoSetBreakMap(pid_t pid, unsigned long breakIdx) {
  PidMap_t::iterator iter = umSetBreakMap->find(pid);
  vector<unsigned long> *vec;

  if(iter == umSetBreakMap->end()) {
    vec = new vector<unsigned long>;
    (*umSetBreakMap)[pid] = vec;
  } else {
    vec = iter->second;
  }

  vec->insert(vec->end(), breakIdx);
}

void Plato::registerUmBreak(unsigned long idx) {
  string desc;
  StringMap_t *m;

  if(breakPointVec.at(idx).getBreakType() == USER_MODE_BP) {
    desc = breakPointVec.at(idx).getUserModeComm();
    m = umBreakMap;
  } else if(breakPointVec.at(idx).getBreakType() == SHARED_LIB_BP) {
    desc = breakPointVec.at(idx).getUserModeSharedLib();
    m = slNameBreakMap;
    hasSharedLibBp = true;
  } else {
    assert(0);
  }

  hasUserModeCb = true;

  assert(breakPointVec.at(idx).isUserMode());
  StringMap_t::iterator iter = m->find(desc);
  if(iter == m->end()) {
    (*m)[desc] = new vector<unsigned long>;
    iter = m->find(desc);
  }
    
  assert(iter != m->end());
  iter->second->insert(iter->second->end(), idx);
}

pid_t Plato::getCurrentPid() {
  return ::getCurrentPid(umlFd);
}

pid_t getCurrentPid(int fd) {
  struct pt_regs guestKernRegs;
  struct linux_task_struct *curr;
  pid_t pid;

  ioctl(fd,PLATO_IOC_QUERY_GUEST_KERN_REGS,&guestKernRegs);
  curr = (struct linux_task_struct *) getCurrent(&guestKernRegs);

  // since there is no breakpoint yet, call the low level ioctl to get
  // the memory
  struct plato_mem_struct pMem;
  
  pMem.bufferAddr = (unsigned long) &pid;
  pMem.guestAddr = (unsigned long) &curr->pid;
  pMem.len = sizeof(pid);
  pMem.isKern = 1;

  ioctl(fd, PLATO_IOC_READ_MEM, &pMem);
  assert(pMem.ret == 0);

  return pid;
}
  

void Plato::handleUserModeTrap(struct pt_regs *regs, bool internalOnly) {
  vector<unsigned long> *vec;
  PidMap_t::iterator bmIter;
  unsigned long breakIdx, idx;
  pid_t pid = getCurrentPid();
  vector< breakpoint >::iterator iter;

  bmIter = umSetBreakMap->find(pid);
  if(bmIter == umSetBreakMap->end()) {
    cerr << "guest user mode breakpoint at " << (void *) regs->eip 
	 << "for process " << pid << " not found" << endl;
    return;
  }

  vec = bmIter->second;

  for(idx = 0; idx < vec->size(); idx++) {
    breakIdx = vec->at(idx);
    iter = breakPointVec.begin() + breakIdx;
    assert(iter->isUserMode());
    if(handleTrap(iter, regs, internalOnly)) break;
  }             

  if(idx >= vec->size()) {
    cerr << "could not find usermode breakpoint " 
	 << "eip = " << (void *) regs->eip << endl;
    unsigned long tmpOp;
    for(idx = regs->eip - 4; idx < (unsigned long) (regs->eip + 20); idx+=4) {
      copy_from_guest(&tmpOp, idx, sizeof(tmpOp));
      cerr << "opcode " << (void *) idx << " = " << (void *) tmpOp << endl;
    }
  }
}

using namespace std;
/* Action modifiers - doesn't change the zoning */
#define __GFP_WAIT      0x10    /* Can wait and reschedule? */
#define __GFP_HIGH      0x20    /* Should access emergency pools? */
#define __GFP_IO        0x40    /* Can start low memory physical IO? */
#define __GFP_HIGHIO    0x80    /* Can start high mem physical IO? */
#define __GFP_FS        0x100   /* Can call down to low-level FS? */

#define GFP_KERNEL      (__GFP_HIGH | __GFP_WAIT | __GFP_IO | __GFP_HIGHIO | __GFP_FS)

unsigned long Plato::lookupInode(struct linux_file *f) {
  try {
    if(f == NULL) return 0;
    struct linux_dentry *d = (struct linux_dentry *) guest_read_ulong(&f->f_dentry);
    if(d == NULL) return 0;
    struct linux_inode *i = (struct linux_inode *) guest_read_ulong(&d->d_inode);
    if(i == NULL) return 0;
    return guest_read_ulong(&i->i_ino);
  } catch (...) {
    //cout << "invalid file pointer" << endl;
  }
  
  return 0;
}

unsigned long Plato::guestStat(string fileName) {
  struct linux_stat *st;
  struct linux_task_struct *curr;
  unsigned long addrLimit;
  int ret;
  unsigned long fileNameAddr;
  unsigned long inode;

  unsigned long checkId = checkpoint();
  curr = (struct linux_task_struct *) getCurrent();  
  
  // allocate memory for the stat struct
  st = (struct linux_stat *) callFunction("kmalloc",
					  sizeof(struct linux_stat),
					  GFP_KERNEL);
  guest_memset(st,0,sizeof(struct linux_stat));
  
  fileNameAddr = callFunction("kmalloc",fileName.size()+1,GFP_KERNEL);
  copy_to_guest(fileNameAddr, (void *) fileName.c_str(),fileName.size() + 1);
  
  // set the addr_limit so we can call sys_newstat
  addrLimit = guest_read_ulong(&curr->addr_limit);
  guest_write_ulong(&curr->addr_limit, 0xffffffff);
  
  // now call sys_newstat
  ret = callFunction("sys_newstat",fileNameAddr,(unsigned long) st);
  
  if(ret == 0) {
    inode = guest_read_ulong(&st->st_ino);
  } else {
    inode = 0;
    cerr << "failed stat " << fileName << ", ret = " << ret << endl;
  }

  rollback(checkId);
  deleteCheckpoint(checkId);

  return inode;
}

void Plato::setSharedLibBreaks(pid_t pid, struct linux_file *file, 
			       unsigned long pgoff, unsigned long startAddr,
			       unsigned long len) {
  unsigned long inode = lookupInode(file);
  unsigned long base, breakAddr, idx;
  if(slInodeBreakMap->find((pid_t) inode) != slInodeBreakMap->end()) {
    vector<unsigned long> *vec = slInodeBreakMap->find((pid_t) inode)->second;
    base = pgoff * PAGE_SIZE;
    
    for(idx = 0; idx < vec->size(); idx++) {
      breakAddr = breakPointVec.at(vec->at(idx)).getRelBreakAddr();
      if((breakAddr >= base) && (breakAddr < (base+len))) {
	breakPointVec.at(vec->at(idx)).setOffset(pid, startAddr - base);
	tryToSetBreaks(pid, vec);
      }
    }
  } 
}

void Plato::slMmapPgoffCb() {
  unsigned long inode;

  if(slNameBreakMap->size() > 0) {
    StringMap_t::iterator iter;
    unsigned long idx;
    vector<string> freeVec;

    for(iter = slNameBreakMap->begin(); iter != slNameBreakMap->end(); iter++) {
      inode = guestStat(iter->first);
      if(inode != 0) {
	(*slInodeBreakMap)[(pid_t) inode] = iter->second;
	freeVec.insert(freeVec.end(), iter->first);
      }
    }
    for(idx = 0; idx < freeVec.size(); idx++) {
      slNameBreakMap->erase(slNameBreakMap->find(freeVec[idx]));
    }
  }

  setSharedLibBreaks(getCurrentPid(), (struct linux_file *) readVar("file"),
		     readVar("pgoff"), readVar("addr"), readVar("len"));
}

void Plato::queueSetBreaks(pid_t pid, unsigned long pageAddr) {
  unsigned long idx, breakId, breakAddr;
  assert(!(pageAddr & (PAGE_SIZE-1)));
  if(umSetBreakMap->find(pid) != umSetBreakMap->end()) {
    vector<unsigned long> *setVec = umSetBreakMap->find(pid)->second;
    vector<unsigned long> *newSetVec = new vector<unsigned long>;
    
    // check if the breakpoints are about to get unmapped
    for(idx = 0; idx < setVec->size(); idx++) {
      breakId = setVec->at(idx);
      assert(breakPointVec[breakId].isUserMode());
      breakAddr = breakPointVec[breakId].absBreakAddr();
      if((breakAddr >= pageAddr) && (breakAddr < (pageAddr+PAGE_SIZE))) {
	cerr << "queuing swapped out breakpoint" << endl;
	queueBreakPoint(pid, breakId);
      } else {
	newSetVec->insert(newSetVec->end(), breakId);
      }
    }
    
    delete setVec;
    (*umSetBreakMap)[pid] = newSetVec;
  }
}

void Plato::umSchedCb() {
  struct linux_task_struct *next = (struct linux_task_struct *) readVar("next");
  pid_t pid = (pid_t) guest_read_ulong(&next->pid);

  if(umQueuedBreakMap->find(pid) != umQueuedBreakMap->end()) {
    mapBreak->enable();
  } else {
    mapBreak->disable(true);
  }

  if(umQueuedBreakMap->size() == 0) {
    mapBreak->disable(true);
    schedBreak->disable(true);
  }
}

void Plato::umSwapCb() {
  struct linux_task_struct *tsk, *cur;
  struct linux_mm_struct *mm = (struct linux_mm_struct *) readVar("mm");
  struct linux_mm_struct *tskMm;
  unsigned long address = readVar("address");
  pid_t pid;
  bool foundMm;

  // now look for any of our procs
  cur = (struct linux_task_struct *) getCurrent();
  tsk = cur;
  foundMm = false;
  do {
    copy_from_guest(&tskMm, &tsk->mm, sizeof(tskMm));
    if(mm == tskMm) {
      copy_from_guest(&pid, &tsk->pid, sizeof(pid));
      queueSetBreaks(pid, address);
      foundMm = true;
      break;
    }
    copy_from_guest(&tsk, &tsk->next_task, sizeof(tsk));
  } while(tsk != cur);
  if(!foundMm) {
    cerr << "could not find mm" << endl;
  }
}

static Plato *llPlatoPtr = NULL;
void llSwapCb() {
  llPlatoPtr->umSwapCb();
}

void llForkCb() {
  llPlatoPtr->umForkCb();
}

void llExitCb() {
  llPlatoPtr->umExitCb();
}

void llExecCb() {
  llPlatoPtr->umExecCb();
}

void llMapCb() {
  llPlatoPtr->umMapCb();
}

void llMmapPgoffCb() {
  llPlatoPtr->slMmapPgoffCb();
}

void llSchedCb() {
  llPlatoPtr->umSchedCb();
}

void userModeInit(Plato *ptr) {
  if(ptr->hasUserModeCb) {
    ptr->registerBreak(FORK_BREAK_ID, llForkCb, true);
    ptr->registerBreak(EXEC_BREAK_ID, llExecCb, true);
    ptr->registerBreak(MAP_BREAK_ID,  llMapCb, true);
    ptr->registerBreak(EXIT_BREAK_ID, llExitCb, true);
    ptr->registerBreak(MMAP_PGOFF_BREAK_ID, llMmapPgoffCb, true);
    ptr->registerBreak(SWAP_BREAK_ID, llSwapCb, true);
    ptr->registerBreak(SCHED_BREAK_ID, llSchedCb, true);
    assert(llPlatoPtr == NULL);
    llPlatoPtr = ptr;
  }
}
