/*
 * chkpt_restore.c
 * v3
 * 
 * Take a bunch of checkpoints between markers, stopping at last marker.
 * for each checkpoint {
 *   trash cache  { write two 2GB files and delete them }
 *   time { move to checkpoint }
 *   move to end
 * }
 */
#include <stdlib.h>
#include <stdio.h>
#include <asm/unistd.h>
#include <errno.h>
#include <time.h>
#include <string.h>

#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>

#include <revirt/rlog.h>
#include <revirt/ztrace.h>
#include <revirt/syscalls.h>
#include <revirt/flags.h>

#include <sys/vfs.h>

#include "library.h"
#include "lib-checkpoint.h"

#include "uml.h"

#define MAX_DATAPOINTS 50
#define MAX_REDUNDANCY 5

int interval, max_chkpts, redundancy;

struct datapoint_struct {
    int delta;
    struct {
	struct timeval cold_start;
	struct timeval cold_end;
	struct timeval warm_start;
	struct timeval warm_end;
    } data[MAX_REDUNDANCY];
} dp[MAX_DATAPOINTS];

#define ts_to_double(x)	\
	((x).tv_sec + (((double)(x).tv_usec) / 1000000.0))

#define dprintf(x...)

void print_datapoint(struct datapoint_struct *d) {
	int i;
	double cold[MAX_REDUNDANCY], warm[MAX_REDUNDANCY], c_total, w_total,
		c_avg, w_avg;
	c_total=w_total=0;
	for(i=0; i<redundancy; i++) {
	    cold[i] = ts_to_double(d->data[i].cold_end) 
	    		- ts_to_double(d->data[i].cold_start);
	    c_total +=cold[i];
	    dprintf("  cold[%d]= %3.2lf (%d.%6d-%d.%6d) sum %3.2lf\n",
		i, cold[i], 
	    	d->data[i].cold_end.tv_sec, d->data[i].cold_end.tv_usec,
	    	d->data[i].cold_start.tv_sec, d->data[i].cold_start.tv_usec,
		c_total);
	    warm[i] = ts_to_double(d->data[i].warm_end) 
	    		- ts_to_double(d->data[i].warm_start);
	    w_total += warm[i];
	    dprintf("  warm[%d]= %3.2lf (%d.%6d-%d.%6d) sum %3.2lf\n",
		i, warm[i], 
	    	d->data[i].warm_end.tv_sec, d->data[i].warm_end.tv_usec,
	    	d->data[i].warm_start.tv_sec, d->data[i].warm_start.tv_usec,
		w_total);
	}
	c_avg = c_total / redundancy;
	w_avg = w_total / redundancy;
    fprintf(stderr, "Interval %d delta %d cold_avg %3.2lf warm_avg %3.2lf\n",
    	interval * d->delta,
	d->delta,
	c_avg,
	w_avg);
    fprintf(stdout, "Interval %d delta %d cold_avg %3.2lf warm_avg %3.2lf\n",
    	interval * d->delta,
	d->delta,
	c_avg,
	w_avg);
}

void print_intermediate_datapoint(struct datapoint_struct * d, int k) {
    double cold1, cold2, warm1, warm2;
    cold1=ts_to_double(d->data[k].cold_start);
    cold2=ts_to_double(d->data[k].cold_end);
    warm1=ts_to_double(d->data[k].warm_start);
    warm2=ts_to_double(d->data[k].warm_end);

    fprintf(stderr, "%d Interval %d delta %d cold_avg %3.2lf warm_avg %3.2lf\n",
	k,
    	interval * d->delta,
	d->delta,
	cold2-cold1,
	warm2-warm1);
    fprintf(stdout, "%d Interval %d delta %d cold %3.2lf warm %3.2lf\n",
	k,
    	interval * d->delta,
	d->delta,
	cold2-cold1,
	warm2-warm1);
}

//int first_cid;
//int max_time=0;

#define TBLEN (1024*1024)
char tbuf[TBLEN];

int trash_cache(void) {
    int fd1, fd2, left=0x0fffffff, len;

    if((fd1=creat("/tmp/trash1", 0666))<0) {
    	perror("creating /tmp/trash1");
    }
    if((fd2=creat("/tmp/trash2", 0666))<0) {
    	perror("creating /tmp/trash1");
    }
    fprintf(stderr, "Writing 2GB each to two files\n");
    for(left=0x7fffffff; left>0; ) {
    	len=(left < TBLEN) ? left : TBLEN;
    	write(fd1, tbuf, len);
    	write(fd2, tbuf, len);
	left -= len;
    }
    fprintf(stderr, "Unlinking files\n");
    unlink("/tmp/trash1");
    unlink("/tmp/trash2");

    return 0;
}

int checkpoint_periodic(int interval, int max_chkpts) {
    int i, ret;
    struct timespec ts;

    ts.tv_sec=interval;
    ts.tv_nsec=0;

    printf("checkpoint_periodic, interval %d\n", interval);

    for(i=0; i<max_chkpts ; i++) {
	ioctl(umlfd, UML_IOC_RESTART_UML, 0);
	nanosleep(&ts, NULL);

	/* This is a mild race condition (between checking for the event and
	 * calling PAUSE) but it's better than nothing */
	if(waitevent_is_waiting()) {
	    fprintf(stderr, "waitevent_is_waiting returns true! setting max_chkpts == %d\n", i);
	    max_chkpts=i+1;
	    waitevent_cont();
	}
	ret = ioctl(umlfd, UML_IOC_PAUSE, 10);
	if(ret) {
    	    fprintf(stderr, "uml_pause failed: %d\n", ret);
    	    exit(1);
       	}
	take_checkpoint_d(); // this gets clobbered each time... 
    }

    return 0;
}

/* FIXME start here tomorrow... */
void do_run(void) {
    int i, j, k;
    struct uml_ioc_checkinfo_struct cur;
    int last_cid;

    get_checkpoint_list();
    get_current_info(&cur);

    if(cur.guestSyscallCount != 
    	checkpoint_list[nr_checkpoints-1].guestSyscallCount) {
	fprintf(stderr, "At syscall %lld, wanted %lld!\n", 
		cur.guestSyscallCount,
		checkpoint_list[nr_checkpoints-1].guestSyscallCount);
	return;
    }

    last_cid = checkpoint_list[nr_checkpoints-1].checkId;

    for(k=0; k<redundancy; k++) {
     j=0;
     for(i=nr_checkpoints-2; i>0 ; i--, j++) {
	//trash_cache();

	dp[j].delta = nr_checkpoints-i-1;
    	fprintf(stderr, "Moving to checkpoint %d from %d [delta %d]\n",
		i, last_cid, dp[j].delta);

	gettimeofday(&dp[j].data[k].cold_start, NULL);
	restore_checkpoint(checkpoint_list[i].checkId);
	gettimeofday(&dp[j].data[k].cold_end, NULL);

	gettimeofday(&dp[j].data[k].warm_start, NULL);
	restore_checkpoint(last_cid);
	gettimeofday(&dp[j].data[k].warm_end, NULL);

	print_intermediate_datapoint(dp+j, k);
     }
    }
    for(i=0; i<j ; i++) {
	print_datapoint(dp+i);
    }
}

int main(int argc, char * argv[]) {
    if(argc<3 || argc>4) {
    	fprintf(stderr, "Usage: %s interval max_chkpts [redundancy]\n", 
			argv[0]);
	exit(1);
    }

    if(argc == 4) {
    	redundancy = atoi(argv[3]);
	if(redundancy > MAX_REDUNDANCY) {
	    fprintf(stderr, "redundancy wanted %d, resetting to max %d\n",
	    	redundancy, MAX_REDUNDANCY);
	    redundancy=MAX_REDUNDANCY;
	}
    } else {
    	redundancy=1;
    }


    interval=atoi(argv[1]);
    max_chkpts=atoi(argv[2]);

    setup_stuff();

    fprintf(stderr, "Taking %d checkpoints, interval %d, redunancy %d\n", 
    	max_chkpts, interval, redundancy);

    dbraw_reclaim();

    checkpoint_periodic(interval, max_chkpts);

    do_run();

    cleanup_stuff();

    return 0;
}
